/*
 * ieeetest.c --- test IDCT code against the IEEE Std 1180-1990 spec
 *
 * Note that this does only one pass of the test.
 * Six invocations of ieeetest are needed to complete the entire spec.
 * The shell script "doieee" performs the complete test.
 *
 * Written by Tom Lane (tgl@cs.cmu.edu).
 * Released to public domain 11/22/93.
 *
 * Modified by Royce Liao (liaor@iname.com) 09/2000
 *  v1.01
 *  -----
 *  moved DCTELEM variable declarations to top of main(), to force 8-byte
 *  aligntment in VC++ 6.  (otherwise arrays are only DWORD aligned, causing
 *  massive stalls in the MMX DCT functions.)
 *
 *  v1.02
 *  -----
 *  Added new iDCT functions, AP922float and AP922hybrid,
 *  dynamic memory allocation using calloc(), for improved memory alignment
 *    For Intel-SSE code, memory alignment is *critical*
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>

#include "dct.h"      // variable and function declarations

#define I_LOW  -256   // iDCT output floor
#define I_HIGH  255   // iDCT output ceiling
#define F_LOW  -2048  // fDCT output floor
#define F_HIGH  2047  // fDCT output ceiling

///////////////////////////////////////////////////////////////////////////
//  If you do *NOT* have the Visual C++ 6 Processor Pack Beta (for Intel-SSE),
//  then you will need to compile the project without idct_ap922sse.
//  If compiling withOUT idct_ap922sse.c, include the declaration below:
const float tab_i_01234567ssef[32*8]={0.0};

//////////////
// If compiling with idct_ap922sse.c, include the extern declaration below:
//extern const float tab_i_01234567ssef[]; // from idct_ap922sse

/* prototypes */ 

void usage (char *msg);
__inline long ieeerand (long L, long H);

void ref_dct_init(void);           // initialize reference forward DCT
void fdct_reference(short *block); // x87-double IEEE-1180/1990 fDCT
void idct_reference(short *block); // x87-double IEEE-1180/1990 iDCT
void fdct_int(short *block);       //  integer32 IEEE-1180/1190 iDCT

void initialize_dcts( void ); 
//////////////////////////////////////////////////////////////////////////
//
// iDCT function prototypes are located in dct.h
//
//////////////////////////////////////////////////////////////////////////

/* error stat accumulators -- assume initialized to 0 */

long sumerrs[DCTSIZE2];
long sumsqerrs[DCTSIZE2];
int maxerr[DCTSIZE2];

void *fTempArray; // global pointer to temp floating point array
void *p_fTempArray; // global pointer to temp floating point array

char * meets (double val, double limit)
{
  return ((fabs(val) <= limit) ? "meets" : "FAILS");
}

// for the MPEG-codec's fast int32 idct, idct.c 
void  idct_int32_init( void );


main(int argc, char **argv)
{
  void (*dctfunction_pointer)( short *) = NULL;
  //typedef (*dctfunction_pointer) dctfunction_ptr_type ;

  DCTELEM   *p_block;	/* random source data */
  DCTELEM   *p_refcoefs; /* coefs from reference FDCT */
  DCTELEM   *p_refout;	/* output from reference IDCT */
  DCTELEM   *p_testout; /* output from test IDCT */

  /* the following are pointers to aligned memory! */
  DCTELEM   *block;	/* random source data */
  DCTELEM   *refcoefs; /* coefs from reference FDCT */
  DCTELEM   *refout;	/* output from reference IDCT */
  DCTELEM   *testout; /* output from test IDCT */

  long minpix, maxpix, sign;
  long curiter, niters;
  int i, j;
  double max, total;
  int test_idct_out_of_bound = 0; /* IDCT output outside of {-256,+255} */

  /* initialize memory */
     p_block = (DCTELEM *) calloc( sizeof(DCTELEM)*((DCTSIZE2>>4)+1), 16 );
     p_refcoefs = (DCTELEM *) calloc( sizeof(DCTELEM)*((DCTSIZE2>>4)+1), 16 );
     p_refout = (DCTELEM *) calloc( sizeof(DCTELEM)*((DCTSIZE2>>4)+1), 16 );
     p_testout= (DCTELEM *) calloc( sizeof(DCTELEM)*((DCTSIZE2>>4)+1), 16 );
     p_fTempArray =( void *) calloc( sizeof(float)*((DCTSIZE2>>4)+1), 32 );

  /* fix addresses of allocated memory */
#define MM( x ) ((long)(x))

     block   = (short *) ( (MM( p_block    ) + 15) & 0xFFFFFFF0 );
     refcoefs= (short *) ( (MM( p_refcoefs ) + 15) & 0xFFFFFFF0 );
     refout  = (short *) ( (MM( p_refout   ) + 15) & 0xFFFFFFF0 );
     testout = (short *) ( (MM( p_testout  ) + 31) & 0xFFFFFFE0 );
     //fTempArray = (void *) ( (MM( p_fTempArray ) + 15) & 0xFFFFFFF0 );
     fTempArray = p_fTempArray;

  //  printf( "\nblock=0x%08X, refcoefs=0x%08X\nrefout=0x%08X, testout=0x%08X",
  //    block, refcoefs, refout, testout );
     if ( block == NULL || refcoefs == NULL || refout == NULL || 
       testout == NULL || fTempArray == NULL )
     {
       fprintf( stderr, "\nCould not allocate memory.  Cannot continue.");
       exit( -1 );
     }

     if ( (MM(block) & 0x7) || (MM(refcoefs) & 0x7) || (MM(refout) & 0x7) 
       || (MM(testout) & 0x7) || (MM(fTempArray) & 0x7) )
     {
       fprintf( stderr, "\nWARNING: Allocated RAM(s) are not 8-byte aligned.");
       fprintf( stderr, "\n         All MMX-accelerated iDCTs in this package perform better with alignment.");
       fprintf( stderr, "\n         Unaligned structures may reduce performance!");
     }

     if ( (MM(fTempArray) & 0xF) )
     {
       fprintf( stderr, "\nWARNING: Allocated RAM is not 16-byte aligned.");
       fprintf( stderr, "\n         The Intel-SSE iDCT in this package requires alignment.");
       fprintf( stderr, "\n         Unaligned structures may cause program crash!");
     }
  /* Argument parsing --- not very bulletproof at all */

  initialize_dcts(); // perform any necesary initialization

/*
  dctfunction_pointer = fdctdaan;
  printf("\n---------------------------------------------------");
  printf("\nIEEE-1180 test: [fdctdaan.c] 'x87d AAN FDCT' (BBMPEG)" );
  printf("\n---------------------------------------------------\n");
*/
/*
  dctfunction_pointer = fdct_3dnow;
  printf("\n---------------------------------------------------");
  printf("\nIEEE-1180 test: [fdct3dn.c]  '3DNow AAN FDCT'" );
  printf("\n---------------------------------------------------\n");
*/
/*
  dctfunction_pointer = fdct_int;
  printf("\n-------------------------------------------------");
  printf("\nIEEE-1180 test: [fdct_int.c]  'int32 AAN FDCT'" );
  printf("\n-------------------------------------------------\n");
*/
/*
  dctfunction_pointer = idct_int;
  printf("\n-------------------------------------------------");
  printf("\nIEEE-1180 test: [IEEE-1180]  'int32 AAN iDCT'" );
  printf("\n-------------------------------------------------\n");
*/

  dctfunction_pointer = idct_ap922float_x87;
  printf("\n-------------------------------------------------------");
  printf("\nIEEE-1180 test: [idct_ap922x87.c] 'AP922float iDCT'" );
  printf("\n-------------------------------------------------------\n");

/*
  dctfunction_pointer = idct_ap922float_amdx87;
  printf("\n-------------------------------------------------------");
  printf("\nIEEE-1180 test: [idct_ap922tdn.c] 'AP922float iDCT' X87" );
  printf("\n-------------------------------------------------------\n");
*/
/*
  dctfunction_pointer = idct_ap922float_3dn;
  printf("\n-------------------------------------------------------");
  printf("\nIEEE-1180 test: [idct_ap922tdn.c] 'AP922float iDCT' 3DNow" );
  printf("\n-------------------------------------------------------\n");
*/
/*
  dctfunction_pointer = idct_ap922float_sse;
  printf("\n-------------------------------------------------------");
  printf("\nIEEE-1180 test: [idct_ap922sse.c] 'AP922float iDCT' SSE" );
  printf("\n-------------------------------------------------------\n");
*/
/*
  // *** NOT QUITE WORKING ***
  dctfunction_pointer = idct_ap922int86;
  printf("\n-----------------------------------------------------");
  printf("\nIEEE-1180 test: [idctmm32.c] 'AP922 integer iDCT' (x86)" );
  printf("\n-----------------------------------------------------\n");
*/
/*
  dctfunction_pointer = idct_ap922hybr_3dn;
  printf("\n--------------------------------------------------------");
  printf("\nIEEE-1180 test: [idct_ap922hybr.c] 'AP922hybrid iDCT 3DNow'" );
  printf("\n--------------------------------------------------------\n");
*/
/*
  dctfunction_pointer = idct_ap922hybr_x87;
  printf("\n--------------------------------------------------------");
  printf("\nIEEE-1180 test: [idct_ap922hybr.c] 'AP922hybrid iDCT x87'" );
  printf("\n--------------------------------------------------------\n");
*/
  if (argc != 5) usage(NULL);

  if ( dctfunction_pointer == NULL )  // did user forget to select a DCT?
  {
      fprintf( stderr, "\nieeetest.c - main() : dctfunction_pointer == NULL!");
      exit(-1);
  }

  printf("\n");

  minpix = atoi(argv[1]);
  maxpix = atoi(argv[2]);
  sign   = atoi(argv[3]);
  niters = atol(argv[4]);

  /* Loop once per generated random-data block */

  for (curiter = 0; curiter < niters; curiter++) {

    /* generate a pseudo-random block of data */
    for (i = 0; i < DCTSIZE2; i++)
      block[i] = (DCTELEM) (ieeerand(-minpix,maxpix) * sign);

    /* perform reference FDCT */
    memcpy(refcoefs, block, sizeof(DCTELEM)*DCTSIZE2);
    fdct_reference(refcoefs);
    /* clip the ref-fDCT output */
    for (i = 0; i < DCTSIZE2; i++) {
      if (refcoefs[i] < F_LOW) refcoefs[i] = F_LOW;
      else if (refcoefs[i] > F_HIGH) refcoefs[i] = F_HIGH;
    }

    /* perform reference IDCT */
    memcpy(refout, refcoefs, sizeof(DCTELEM)*DCTSIZE2);
    idct_reference(refout);
    /* clip the ref-iDCT output  */
    for (i = 0; i < DCTSIZE2; i++) {
      if (refout[i] < I_LOW) refout[i] = I_LOW;
      else if (refout[i] > I_HIGH) refout[i] = I_HIGH;
    }

//    memcpy(testout, block, sizeof(DCTELEM)*DCTSIZE2);
//    fdct_reference(testout);

    /* perform test DCT */
    memcpy(testout, refcoefs, sizeof(DCTELEM)*DCTSIZE2);
/*------------------------------------------------------------*/

	(*dctfunction_pointer)(testout);

    // The DCT function is responsible for calling EMMS, if necessary.
/*------------------------------------------------------------*/

    /* clip the test output */

    for (i = 0; i < DCTSIZE2; i++)
    {
      if (testout[i] < I_LOW) 
      { 
        testout[i] = I_LOW;
        test_idct_out_of_bound = 1; // flag error
      }
      else if (testout[i] > I_HIGH) 
      {
        testout[i] = I_HIGH;
        test_idct_out_of_bound = 1; // flag error
      }
    }

    /* accumulate error stats */
    for (i = 0; i < DCTSIZE2; i++) {
      register int err = testout[i] - refout[i];
      sumerrs[i] += err;
      sumsqerrs[i] += err * err;
      if (err < 0) err = -err;
      if (maxerr[i] < err) maxerr[i] = err;
    }

    if (curiter % 100 == 99) {
      fprintf(stderr, ".");
      fflush(stderr);
    }
  }
  fprintf(stderr, "\n");

  /* print results */
  printf("IEEE test conditions: -L = %ld, +H = %ld, sign = %ld, #iters = %ld\n",
	 minpix, maxpix, sign, niters);

  printf("Peak absolute values of errors:\n");
  for (i = 0, j = 0; i < DCTSIZE2; i++) {
    if (j < maxerr[i]) j = maxerr[i];
    printf("%4d", maxerr[i]);
    if ((i%DCTSIZE) == DCTSIZE-1) printf("\n");
  }
  printf("Worst peak error = %d  (%s spec limit 1)\n\n", j,
	 meets((double) j, 1.0));

  printf("Mean square errors:\n");
  max = total = 0.0;
  for (i = 0; i < DCTSIZE2; i++) {
    double err = (double) sumsqerrs[i]  / ((double) niters);
    total += (double) sumsqerrs[i];
    if (max < err) max = err;
    printf(" %8.5f", err);
    if ((i%DCTSIZE) == DCTSIZE-1) printf("\n");
  }
  printf("Worst pmse  = %12.8f  (%s spec limit 0.06)\n", 
    max, meets(max, 0.06));
  total /= (double) (64*niters);
  printf("Overall mse = %12.8f  (%s spec limit 0.02)\n\n", total,
	 meets(total, 0.02));

  printf("Mean errors:\n");
  max = total = 0.0;
  for (i = 0; i < DCTSIZE2; i++) {
    double err = (double) sumerrs[i]  / ((double) niters);
    total += (double) sumerrs[i];
    printf(" %8.5f", err);
    if (err < 0.0) err = -err;
    if (max < err) max = err;
    if ((i%DCTSIZE) == DCTSIZE-1) printf("\n");
  }
  printf("Worst mean error   = %12.8f  (%s spec limit 0.015)\n", max,
	 meets(max, 0.015));
  total /= (double) (64*niters);
  printf("Overall mean error = %12.8f  (%s spec limit 0.0015)\n\n", total,
	 meets(total, 0.0015));

  printf("iDCT output range check {-256,+255} : " );
  if ( test_idct_out_of_bound )
      printf("FAIL\n\n");
  else
      printf("pass\n\n");

  /* test for 0 input giving 0 output */
  memset(testout, 0, sizeof(DCTELEM)*DCTSIZE2);
  fdct_int(testout);
  for (i = 0, j=0; i < DCTSIZE2; i++) {
    if (testout[i]) {
      printf("Position %d of IDCT(0) = %d (FAILS)\n", i, testout[i]);
      j++;
    }
  }
  printf("%d elements of IDCT(0) were not zero\n\n\n", j);

  /* free the allocated memory (not the aliases) */
  free( p_block );
  free( p_refcoefs );
  free( p_refout );
  free( p_testout );
  free( p_fTempArray );

  exit(0);
  return 0;
}


void usage (char *msg)
{
  if (msg != NULL)
    fprintf(stderr, "\nerror: %s\n", msg);

  fprintf(stderr, "\n");
  fprintf(stderr, "usage: ieeetest minpix maxpix sign niters\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "  minpix = -L value per IEEE spec\n");
  fprintf(stderr, "  maxpix =  H value per IEEE spec\n");
  fprintf(stderr, "  sign = +1 for normal, -1 to run negated test\n");
  fprintf(stderr, "  niters = # iterations (10000 for full test)\n");
  fprintf(stderr, "\n");

  exit(1);
}

 
/* Pseudo-random generator specified by IEEE 1180 */
 
__inline long ieeerand (long L, long H)
{
  static long randx = 1;
  static long dummy_for_padding; // not used
  static double z = (double) 0x7fffffff;

  long i,j;
  double x;

  randx = (randx * 1103515245) + 12345;
  i = randx & 0x7ffffffe;
  x = ((double) i) / z;
  x *= (L+H+1);
  j = x;
  return j-L;
}



/* The cosine lookup tables for idct_reference and fdct_reference */
/* coslu[a][b] = C(b)/2 * cos[(2a+1)b*pi/16] */
double coslu[8][8];
float cosluf[8][8];
double res[8][8];

/* Routine to initialise the reference-DCT cosine lookup table */
void ref_dct_init(void)
{
  int a,b;
  double tmp;

  for(a=0;a<8;a++)
    for(b=0;b<8;b++) {
      tmp = cos((double)((a+a+1)*b) * (3.14159265358979323846 / 16.0));
      if(b==0)
	tmp /= sqrt(2.0);
      coslu[a][b] = tmp * 0.5;
      cosluf[a][b] = tmp * 0.5;
    }
}


void fdct_reference (short *block)
{
  int x,y,u,v;
  double tmp, tmp2;

  for (v=0; v<8; v++) {
    for (u=0; u<8; u++) {
      tmp = 0.0;
      for (y=0; y<8; y++) {
	tmp2 = 0.0;
	for (x=0; x<8; x++) {
	  tmp2 += (double) block[(y*DCTSIZE) + x] * coslu[x][u];
	}
	tmp += coslu[y][v] * tmp2;
      }
      res[v][u] = tmp;
    } 
  }

  for (v=0; v<8; v++) {
    for (u=0; u<8; u++) {
      tmp = res[v][u];
      if (tmp < 0.0) {
	x = - ((int) (0.5 - tmp));
      } else {
	x = (int) (tmp + 0.5);
      }
      block[(v*DCTSIZE)+u] = (DCTELEM) x;
    }
  }
}


void idct_reference (short *block)
{
  int x,y,u,v;
  double tmp, tmp2;

  for (y=0; y<8; y++) {
    for (x=0; x<8; x++) {
      tmp = 0.0;
      for (v=0; v<8; v++) {
	tmp2 = 0.0;
	for (u=0; u<8; u++) {
	  tmp2 += (double) block[(v*DCTSIZE)+u] * coslu[x][u];
	}
	tmp += coslu[y][v] * tmp2;
      }
      res[y][x] = tmp;
    }
  }

  for (v=0; v<8; v++) {
    for (u=0; u<8; u++) {
      tmp = res[v][u];
      if (tmp < 0.0) {
	x = - ((int) (0.5 - tmp));
      } else {
	x = (int) (tmp + 0.5);
      }
      block[(v*DCTSIZE)+u] = (DCTELEM) x;
    }
  }
}

void
initialize_dcts( void )
{
  ref_dct_init(); // initialize ieee1180 test package's reference iDCT/fDCT

  /*
  init_fdct_3dnow(); // initialize 3dNow fDCT
  init_fdctdaan();   // initialize BBMPEG AAN fDCT x87d
*/
  init_idct_int();

  //////////////////////////////////////////////////////////////////////
  // check alignment of idct_ap922sse.c
  // If the table is not aligned, uh oh, the code will not function properly!
  fprintf( stderr, "\n&tab_i_01234567ssef[] = 0x%08X", MM(tab_i_01234567ssef) );
  if ( MM(tab_i_01234567ssef) & 0xF )
  {
    fprintf( stderr, "\nWARNING: The address of tab_i_01234567ssef[] is not 16-byte aligned.");
    fprintf( stderr, "\n         The Intel-SSE iDCT WILL crash!");
  }

}
