/*

	8x8	forward DCT using MMX
	This implementation of forward DCT using a slightly modification
	of Chen's algorithm. It's well-known that Chen's is the
	slowest algorithm in all fast algorithms for DCT in
	literature (that means there's still space for improving
	using faster algorithms like Lee's, LLM's or AAN's).
	The modification makes Chen's even slower!
	People in signal processing prefer to use the number of
	additions and multiplications for evaluating the 
	complexity of an algorithm. The corresponding numbers
	for familiar methods are:

    Methods		Chen	Lee		LLM		AANscale	ThisMethod
	Mul.		256		192		176		80			256
	Add.		416		464		464		464			672 (it's right!)

    and some overhead in computations.
    Calling syntax and data format

    fdct_chen_mmx1(short int *data)

    where data is a pointer to input block (16 bit integers).
	The DCT output is put  back into the original block.
	(Don't ask me where is fdct_chen_mmx2(), the faster one)

    Comments and/or recommendations should be directed to
	knguyen@ece.ubc.ca

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>


static __int64 fdct_chen_mmx1_c00c  =0x0002000200020002;	// 2
static __int64 fdct_chen_mmx1_c00	=0x2D412D412D412D41;
static __int64 fdct_chen_mmx1_c10c	=0x0002000200020002;	// 2
static __int64 fdct_chen_mmx1_c10	=0x3EC53EC53EC53EC5;
static __int64 fdct_chen_mmx1_c20c	=0x0002000200020002;	// 2
static __int64 fdct_chen_mmx1_c20	=0x3536353635363536;
static __int64 fdct_chen_mmx1_c30c	=0x0003000300030003;	// 4,3,2
static __int64 fdct_chen_mmx1_c30	=0x238E238E238E238E;
static __int64 fdct_chen_mmx1_c40c	=0x000A000A000A000A;	// A
static __int64 fdct_chen_mmx1_c40	=0x0C7C0C7C0C7C0C7C;
static __int64 fdct_chen_mmx1_c50c	=0x0002000200020002;	// 2
static __int64 fdct_chen_mmx1_c50	=0x3B203B203B203B20;
static __int64 fdct_chen_mmx1_c60c	=0x0005000500050005;	//	5 
static __int64 fdct_chen_mmx1_c60	=0x187D187D187D187D;

void fdct_chen_mmx1(short int *x)		// x should be (32)align
{ 
	__asm{

	mov	ebx, x

#define r0	mm0
#define r2	mm1
#define r1	mm2	
#define r3	mm3	
#define r5	mm4	
#define r4	mm5	
#define r6	mm6	
#define r7	mm7	

    movq	r0, [ebx+0*2];		// r0=x0		r0		 						

	movq	r2, [ebx+56*2];		// r2=x7		r2
    movq	r1, r0;				// r1=x0		r1							
	psubw	r1, r2;				// r1=c3								

	paddw	r0,r2;				// r0=a0				r2
    movq	r3, [ebx+8*2];		// r3=x1		r3		 						
	psllw	r1, 2;				// c3<<=2

    movq	r5, [ebx+48*2];		// r5=x6		r5
	psllw	r0, 2;				// a0<<=2
    movq	r4, r3;				// r4=x1		r4							

	paddw	r3, r5;				// r3=a1
	movq	r6, [ebx+16*2];		// r6=x2		r6		 						
	psubw	r4, r5;				// r4=c2				r5
#define r8	r2	
	movq	[ebx+56*2], r1;		// x7=c3				r1	
    movq	r7, r6;				// r7=x2		r7							
    movq	r8, [ebx+40*2];		// r8=x5		r8

	psllw	r4, 2;				// c2<<=2
	psubw	r7, r8;				// r7=c1				
	psllw	r3, 2;				// a1<<=2
#define r9	r5	
	movq	[ebx+48*2], r4;		// x6=c2				r4
	psllw	r7, 2;				// c1<<=2
	movq	r9, [ebx+24*2];		// r9=x3		r9		 						
#define r11	r1
#define r10	r4		
	paddw	r6, r8;				// r6=a2				r8
    movq	r11, [ebx+32*2];	// r11=x4		r11
    movq	r10, r9;			// r10=x3		r10							

	movq	[ebx+40*2],r7;		// x5=c1				r7
	psubw	r10, r11;			// r10=c0				r11
	paddw	r9, r11;			// r9=a3
#define r13	r8	
	psllw	r10, 2;				// c3<<=2
   	movq	r13, r3;			// r13=a1		r13
	psllw	r6, 2;				// a2<<=2

	movq	[ebx+32*2], r10;	// x4=c0				r10
	psllw	r9, 2;				// a3<<=2
	paddw	r3, r6;				// r3=b1

						// even part, a(0,1,2,3) in r(0,3,6,9)
#define r12	r7	
	psubw	r13, r6;			// r13=b2				r6
	movq	r12, r0;				// r12=a0		r12
	paddw	r0, r9;				// r0=b0
						
						// b(0,1,2,3) in r(0,3,13,12)
						// y2 and y6 first, then y0 and y4
#define r16	r11
#define r18	r10		
	movq	r16, fdct_chen_mmx1_c50c;	//		r16
	psubw	r12, r9;			// r12=b3				r9
 	movq	r18, fdct_chen_mmx1_c60c;	//		r18
#define r14	r6
#define r15	r9	
	movq	r14, r12;			// r14=r12=b3	r14
	paddw	r12, r16;
    movq	r15, r13;			// r15=r13=b2	r15
#define	r19	r16
	pmulhw	r12, fdct_chen_mmx1_c50;
	paddw	r13, r16;			//						r16
 	paddw	r15, r18;			//						
#define	r21	r18 
	paddw	r14, r18;			//						r18
	movq	r19, fdct_chen_mmx1_c60;	//		r19
	movq	r21, r0;			// r21=r0=b0	r21

	pmulhw	r13, fdct_chen_mmx1_c50;	//					
	paddw	r0, r3;				// b0+b1
	pmulhw	r15, r19;			//						
#define r22	r3
	psubw	r21, r3;			// b0-b1				r3
	movq	r22, fdct_chen_mmx1_c00c;	//		r22
	pmulhw	r14, r19;			//						r19
#define r23	r19	
	paddw	r0, r22;
	paddw	r21, r22;
	movq	r23, fdct_chen_mmx1_c00;	//		r23
#define	r20	r15
	paddw	r12, r15;			// r12=y2				r15
	movq	r20, [ebx+48*2];	// r20=c2		r20
	psubw	r14, r13;			// r14=y6				r13
#define	r25	r13
	movq	[ebx+16*2], r12;	// x2=y2				r12
	pmulhw	r0, r23;			// r0=y0
    movq	r25, [ebx+40*2];	// r25=c1		r25
#define	r26	r12
	movq	r26, r20;			// r26=c2;		r26
	paddw	r20, r25;			// c2+c1
	psubw	r26, r25;			// c2-c1				r25

	movq	[ebx+48*2], r14;	// x6=y6				r14
	psllw	r20, 2;
	psllw	r26, 2;

	paddw	r20, r22;
	paddw	r26, r22;			//						r22
	pmulhw	r21, r23;			// r4=y4
#define	r24	r25
	movq	[ebx+0*2], r0;		// x0=y0				r0
	pmulhw	r20, r23;			// r20=b1
	movq	r24, [ebx+32*2];	// r24=c0		r24
#define	r27	r14
#define	r28	r22
	pmulhw	r26, r23;			// r26=b0				r23
    movq	r27, [ebx+56*2];	// r27=c3		r27
	movq	r28, r24;			// r28=r24=c0	r28
#define r29 r0
	movq	[ebx+32*2], r21;	// x4=y4				r21
	movq	r29, r27;			// r29=r27=c3	r29
    paddw	r24, r26;			// r24=a0

	psubw	r28, r26;			// r28=a1				r26
	paddw	r27, r20;			// r27=a3
	psubw	r29, r20;			// r29=a2				r20
	
						// odd part, a(0,1,2,3)=r(24,28,29,27)
			
						// y1 and y7 from r24 and r27, r28 and r29 idle
#define	r32	r23
#define	r30	r21
#define	r33	r26
	movq	r32, fdct_chen_mmx1_c40c;	//		r32
 	movq	r30, r24;			// r30=r24=a0	r30
	movq	r33, fdct_chen_mmx1_c10c;	//		r33
#define	r31	r20
	paddw	r24, r32;
	movq	r31, r27;			// r31=r27=a3	r31
	paddw	r27, r32;			//						r32
#define	r35	r32
    pmulhw	r24, fdct_chen_mmx1_c40;	
	paddw	r31, r33;			//						 
	movq	r35, fdct_chen_mmx1_c10;	//		r35
#define r38 r33	
	paddw	r30, r33;			//						r33
	movq	r38, fdct_chen_mmx1_c20c;	//		r38
	pmulhw	r31, r35;			//						
#define	r36 r35
	pmulhw	r27, fdct_chen_mmx1_c40;	//	
    pmulhw	r30, r35;			//						r35
    movq	r36, r29;			// r36=r29=a2	r36
#define r37 r31
	paddw	r29, r38;
	paddw	r24, r31;			// r24=y1				r31
  	movq	r37, r28;			// r37=r28=a1	r37
#define r39 r38
	paddw	r28, r38;			//						r38
	movq	r39, fdct_chen_mmx1_c30c;	//		r39
	psubw	r27, r30;			// r27=y7				r30
#define r40 r30
	movq	[ebx+8*2], r24;		// x1=y1				r24
	paddw	r36, r39;
	movq	r40, fdct_chen_mmx1_c20;	//		r40  
#define r41 r24
	paddw	r37, r39;			//						r39
    movq	r41, fdct_chen_mmx1_c30;	//		r41
	pmulhw	r29, r40;
#define v0 r39
	movq	[ebx+56*2], r27;		// x7=y7			r27
	pmulhw	r28, r40;			//						r40
 	movq	v0, [ebx+8+0*2];	// v0=x0		v0		 						
#define v2 r27
	pmulhw	r36, r41;
	movq	v2, [ebx+8+56*2];	// v2=x7		v2
	pmulhw	r37, r41;			//						r41
#define v1 r40	
	movq	v1, v0;				// v1=x0		v1							
	paddw	v0, v2;				// v0=a0				v2
 	paddw	r36, r28;			// r36=y5				r28
#define v3 r41
	psubw	v1, v2;				// v1=c3					
    movq	v3, [ebx+8+8*2];	// v3=x1		v3		 						
	psubw	r29, r37;			// r29=y3				r37
#define v5 r28	
	movq	[ebx+40*2], r36;	// x5=y5				r36
	psllw	v1, 2;				// c3<<=2
	movq	v5, [ebx+8+48*2];	// v5=x6		v5
#define v4 r37
#define v6	r36	
	psllw	v0, 2;				// a0<<=2
	movq	v6, [ebx+8+16*2];	// v6=x2		v6		 						
    movq	v4, v3;				// v4=x1		v4							

								// finish first column half

#define v8 r29
	movq	[ebx+24*2], r29;	//						r29
	paddw	v3, v5;				// v3=a1
    movq	v8, [ebx+8+40*2];		// v8=x5	v8
#define v7	v2	
	psubw	v4, v5;				// v4=c2				v5
    movq	v7, v6;				// v7=x2		v7							
	psllw	v4, 2;				// c2<<=2
#define v9	v5		
	movq	[ebx+8+56*2], v1;	// x7=c3				v1	
	psubw	v7, v8;				// v7=c1				
	movq	v9, [ebx+8+24*2];	// v9=x3		v9		 						
#define v11	v1
	psllw	v3, 2;				// a1<<=2
    movq	v11, [ebx+8+32*2];	// v11=x4		v11
	psllw	v7, 2;				// c1<<=2

#define v10	v4		
	movq	[ebx+8+48*2], v4;	// x6=c2				v4
	paddw	v6, v8;				// v6=a2				v8
 	movq	v10, v9;			// v10=x3		v10							
#define v13	v8	
	paddw	v9, v11;			// v9=a3				
   	movq	v13, v3;			// v13=a1		v13
	psubw	v10, v11;			// v10=c0				v11
#define v12	v7	
	movq	[ebx+8+40*2],v7;	// x5=c1				v7
	psllw	v10, 2;				// c3<<=2
	movq	v12, v0;			// v12=a0		v12

	psllw	v6, 2;				// a2<<=2
	psllw	v9, 2;				// a3<<=2
	paddw	v3, v6;				// v3=b1
#define v16	v11
						// even pavt, a(0,1,2,3) in v(0,3,6,9)
	movq	[ebx+8+32*2], v10;	// x4=c0				v10
	psubw	v13, v6;			// v13=b2				v6
	movq	v16, fdct_chen_mmx1_c50c;	//		v16
						
						// b(0,1,2,3) in v(0,3,13,12)
						// y2 and y6 fivst, then y0 and y4

#define v18	v10		
	paddw	v0, v9;				// v0=b0
 	movq	v18, fdct_chen_mmx1_c60c;	//		v18
	psubw	v12, v9;			// v12=b3				v9
#define v14	v6
#define v15	v9	
	movq	v14, v12;			// v14=v12=b3	v14
	paddw	v12, v16;
    movq	v15, v13;			// v15=v13=b2	v15
#define	v19	v16
	pmulhw	v12, fdct_chen_mmx1_c50;
	paddw	v13, v16;			//						v16
 	paddw	v15, v18;			//						
#define	v21	v18 
	paddw	v14, v18;			//						v18
	movq	v19, fdct_chen_mmx1_c60;	//		v19
	movq	v21, v0;			// v21=v0=b0	v21

	pmulhw	v13, fdct_chen_mmx1_c50;	//					
	paddw	v0, v3;				// b0+b1
	pmulhw	v15, v19;			//						
#define v22	v3
	psubw	v21, v3;			// b0-b1				v3
	movq	v22, fdct_chen_mmx1_c00c;	//		v22
	pmulhw	v14, v19;			//						v19
#define v23	v19	
	paddw	v0, v22;
	paddw	v21, v22;
	movq	v23, fdct_chen_mmx1_c00;	//		v23
#define	v20	v15
	paddw	v12, v15;			// v12=y2				v15
	movq	v20, [ebx+8+48*2];	// v20=c2		v20
	psubw	v14, v13;			// v14=y6				v13
#define	v25	v13
	movq	[ebx+8+16*2], v12;	// x2=y2				v12
	pmulhw	v0, v23;			// v0=y0
    movq	v25, [ebx+8+40*2];	// v25=c1		v25
#define	v26	v12
	movq	v26, v20;			// v26=c2;		v26
	paddw	v20, v25;			// c2+c1
	psubw	v26, v25;			// c2-c1				v25

	movq	[ebx+8+48*2], v14;	// x6=y6				v14
	psllw	v20, 2;
	psllw	v26, 2;

	paddw	v20, v22;
	paddw	v26, v22;			//						v22
	pmulhw	v21, v23;			// v4=y4
#define	v24	v25
	movq	[ebx+8+0*2], v0;		// x0=y0				v0
	pmulhw	v20, v23;			// v20=b1
	movq	v24, [ebx+8+32*2];	// v24=c0		v24
#define	v27	v14
#define	v28	v22
	pmulhw	v26, v23;			// v26=b0				v23
    movq	v27, [ebx+8+56*2];	// v27=c3		v27
	movq	v28, v24;			// v28=v24=c0	v28
#define v29 v0
	movq	[ebx+8+32*2], v21;	// x4=y4				v21
	movq	v29, v27;			// v29=v27=c3	v29
    paddw	v24, v26;			// v24=a0

	psubw	v28, v26;			// v28=a1				v26
	paddw	v27, v20;			// v27=a3
	psubw	v29, v20;			// v29=a2				v20
	
						// odd pavt, a(0,1,2,3)=v(24,28,29,27)
			
						// y1 and y7 fvom v24 and v27, v28 and v29 idle
#define	v32	v23
#define	v30	v21
#define	v33	v26
	movq	v32, fdct_chen_mmx1_c40c;	//		v32
 	movq	v30, v24;			// v30=v24=a0	v30
	movq	v33, fdct_chen_mmx1_c10c;	//		v33
#define	v31	v20
	paddw	v24, v32;
	movq	v31, v27;			// v31=v27=a3	v31
	paddw	v27, v32;			//						v32
#define	v35	v32
    pmulhw	v24, fdct_chen_mmx1_c40;	
	paddw	v31, v33;			//						 
	movq	v35, fdct_chen_mmx1_c10;	//		v35
#define v38 v33	
	paddw	v30, v33;			//						v33
	movq	v38, fdct_chen_mmx1_c20c;	//		v38
	pmulhw	v31, v35;			//						
#define	v36 v35
	pmulhw	v27, fdct_chen_mmx1_c40;	//	
    pmulhw	v30, v35;			//						v35
    movq	v36, v29;			// v36=v29=a2	v36
#define v37 v31
	paddw	v29, v38;
	paddw	v24, v31;			// v24=y1				v31
  	movq	v37, v28;			// v37=v28=a1	v37
#define v39 v38
	paddw	v28, v38;			//						v38
	movq	v39, fdct_chen_mmx1_c30c;	//		v39
	psubw	v27, v30;			// v27=y7				v30
#define v40 v30
	movq	[ebx+8+8*2], v24;	// x1=y1				v24
	paddw	v36, v39;
	movq	v40, fdct_chen_mmx1_c20;	//		v40  
#define v41 v24
	pmulhw	v29, v40;
    movq	v41, fdct_chen_mmx1_c30;	//		v41
	paddw	v37, v39;			//						v39
#define z0 v39
	movq	[ebx+8+56*2], v27;	// x7=y7				v27
	pmulhw	v28, v40;			//						v40 
	movq	z0, [ebx+0*2];		//				z0
#define z1 v27
	pmulhw	v36, v41;
	movq	z1, [ebx+8*2];		//				z1
	pmulhw	v37, v41;			//						v41
#define z2 v40
#define z4 v41	
	movq	z2, [ebx+16*2];		//				z2
	movq	z4, z0				//				z4
	paddw	v36, v28;			// v36=y5				v28
#define z3 v28
	punpcklwd	z0, z1
	movq	z3, [ebx+24*2];		//				z3
	psubw	v29, v37;			// v29=y3				v37
#define z5 v37
	movq	[ebx+8+40*2], v36;	// x5=y5				v36
	punpckhwd	z4, z1			//						z1
    movq	z5, z2				//				z5

								// rows calculation

#define z6 v36
	punpcklwd	z2, z3
	movq	z6, z0				//				z6
	punpckhwd	z5, z3			//						z3

	movq	[ebx+8+24*2], v29;	//						v29
	punpckldq	z0, z2			// z0=(00,10,20,30)
	movq	[ebx+0*2], z0		//						z0
#define z7 z1
	punpckhdq	z6, z2			// z6=(01,11,21,31)		z2
	movq	[ebx+8*2], z6		//						z6
	movq	z7, z4				//				z7
#define z8 z3
	movq	z8, [ebx+8+0*2];	//				z8
	punpckldq	z4, z5			// z4=(02,12,22,32)
#define z9 v29
	movq	z9, [ebx+8+8*2];	//				z9
	punpckhdq	z7, z5			// z7=(03,13,23,33)		z5
#define z12 z0
	movq	[ebx+16*2], z4		//						z4
	movq	z12, z8				//				z12
#define z10 z2
	movq	[ebx+24*2], z7		//						z7
	punpcklwd	z8, z9
	movq	z10, [ebx+8+16*2];	//				z10
#define z11 z6
#define z13 z5
	punpckhwd	z12, z9			//						z9
	movq	z11, [ebx+8+24*2];	//				z11
    movq	z13, z10			//				z13		
#define z14 z4
#define z15 z7
	movq	z14, z8				//				z14		
	punpcklwd	z10, z11
	movq	z15, z12				//			z15
#define rr0 z9
	punpckhwd	z13, z11		//						z11
	movq	rr0, [ebx+0*2];		// rr0=x0		rr0		 						
	punpckhdq	z15, z13		// z15=(07,17,27,37)	
#define rr1	z11
#define rr2 z15
	movq	rr1, rr0;			// rr1=x0		rr1							
	punpckldq	z12, z13		// z12=(06,16,26,36)	z13

	psubw	rr1, rr2;			// rr1=c3								
	punpckhdq	z14, z10		// z14=(05,15,25,35)	

#define rr3	z13	
#define rr5 z12
	movq	rr3, [ebx+8*2];		// rr3=x1		rr3		 						
	paddw	rr0,rr2;			// rr0=a0				rr2
#define rr6	rr2	
	movq	rr6, [ebx+16*2];	// rr6=x2		rr6		 						
	punpckldq	z8, z10			// z8=(04,14,24,34)		z10
#define rr4	z10	
   	movq	rr4, rr3;			// rr4=x1		rr4							
	paddw	rr3, rr5;			// rr3=a1

	movq	[ebx+24*2+8], rr1;	// x7=c3				rr1	
	psubw	rr4, rr5;			// rr4=c2				rr5
#define rr7 rr1	
#define rr8 z14
	movq	rr7, rr6;			// rr7=x2		rr7							
	psubw	rr7, rr8;			// rr7=c1				
	movq	[ebx+16*2+8], rr4;	// x6=c2				rr4
	
#define rr9 rr5	
#define rr11	z8
#define rr10	rr4		
	paddw	rr6, rr8;			// rr6=a2				rr8
	movq	rr9, [ebx+24*2];	// rr9=x3		rr9		 						
    movq	rr10, rr9;			// rr10=x3		rr10							

	movq	[ebx+8*2+8],rr7;		// x5=c1				rr7
	psubw	rr10, rr11;			// rr10=c0				rr11
	paddw	rr9, rr11;			// rr9=a3
#define rr13	rr8	
	movq	[ebx+0*2+8], rr10;	// x4=c0				rr10	
   	movq	rr13, rr3;			// rr13=a1		rr13
	paddw	rr3, rr6;				// rr3=b1

						// even parrt, a(0,1,2,3) in rr(0,3,6,9)
#define rr12	rr7	
	psubw	rr13, rr6;			// rr13=b2				rr6
	movq	rr12, rr0;				// rr12=a0		rr12
	paddw	rr0, rr9;				// rr0=b0

						
						// b(0,1,2,3) in rr(0,3,13,12)
						// y2 and y6 firrst, then y0 and y4
#define rr16	rr11
#define rr18	rr10		
	movq	rr16, fdct_chen_mmx1_c50c;	//		rr16
	psubw	rr12, rr9;			// rr12=b3				rr9
 	movq	rr18, fdct_chen_mmx1_c60c;	//		rr18
#define rr14	rr6
#define rr15	rr9	
	movq	rr14, rr12;			// rr14=rr12=b3	rr14
	paddw	rr12, rr16;
    movq	rr15, rr13;			// rr15=rr13=b2	rr15
#define	rr19	rr16
	pmulhw	rr12, fdct_chen_mmx1_c50;
	paddw	rr13, rr16;			//						rr16
 	paddw	rr15, rr18;			//						
#define	rr21	rr18 
	paddw	rr14, rr18;			//						rr18
	movq	rr19, fdct_chen_mmx1_c60;	//		rr19
	movq	rr21, rr0;			// rr21=rr0=b0	rr21

	pmulhw	rr13, fdct_chen_mmx1_c50;	//					
	paddw	rr0, rr3;				// b0+b1
	pmulhw	rr15, rr19;			//						
#define rr22	rr3
	psubw	rr21, rr3;			// b0-b1				rr3
	movq	rr22, fdct_chen_mmx1_c00c;	//		rr22
	pmulhw	rr14, rr19;			//						rr19
#define rr23	rr19	
	paddw	rr0, rr22;
	paddw	rr21, rr22;
	movq	rr23, fdct_chen_mmx1_c00;	//		rr23
#define	rr20	rr15
	paddw	rr12, rr15;			// rr12=y2				rr15
	movq	rr20, [ebx+16*2+8];	// rr20=c2		rr20
	psubw	rr14, rr13;			// rr14=y6				rr13
#define	rr25	rr13
	movq	[ebx+16*2], rr12;	// x2=y2				rr12
	pmulhw	rr0, rr23;			// rr0=y0
    movq	rr25, [ebx+8*2+8];	// rr25=c1		rr25
#define	rr26	rr12
	movq	rr26, rr20;			// rr26=c2;		rr26
	paddw	rr20, rr25;			// c2+c1
	psubw	rr26, rr25;			// c2-c1				rr25

	movq	[ebx+16*2+8], rr14;	// x6=y6				rr14
	psllw	rr20, 2;
	psllw	rr26, 2;

	paddw	rr20, rr22;
	paddw	rr26, rr22;			//						rr22
	pmulhw	rr21, rr23;			// rr4=y4
#define	rr24	rr25
	movq	[ebx+0*2], rr0;		// x0=y0				rr0
	pmulhw	rr20, rr23;			// rr20=b1
	movq	rr24, [ebx+0*2+8];	// rr24=c0		rr24
#define	rr27	rr14
#define	rr28	rr22
	pmulhw	rr26, rr23;			// rr26=b0				rr23
    movq	rr27, [ebx+24*2+8];	// rr27=c3		rr27
	movq	rr28, rr24;			// rr28=rr24=c0	rr28
#define rr29 rr0
	movq	[ebx+0*2+8], rr21;	// x4=y4				rr21
	movq	rr29, rr27;			// rr29=rr27=c3	rr29
    paddw	rr24, rr26;			// rr24=a0

	psubw	rr28, rr26;			// rr28=a1				rr26
	paddw	rr27, rr20;			// rr27=a3
	psubw	rr29, rr20;			// rr29=a2				rr20
	
						// odd parrt, a(0,1,2,3)=rr(24,28,29,27)
			
						// y1 and y7 frrom rr24 and rr27, rr28 and rr29 idle
#define	rr32	rr23
#define	rr30	rr21
#define	rr33	rr26
	movq	rr32, fdct_chen_mmx1_c40c;	//		rr32
 	movq	rr30, rr24;			// rr30=rr24=a0	rr30
	movq	rr33, fdct_chen_mmx1_c10c;	//		rr33
#define	rr31	rr20
	paddw	rr24, rr32;
	movq	rr31, rr27;			// rr31=rr27=a3	rr31
	paddw	rr27, rr32;			//						rr32
#define	rr35	rr32
    pmulhw	rr24, fdct_chen_mmx1_c40;	
	paddw	rr31, rr33;			//						 
	movq	rr35, fdct_chen_mmx1_c10;	//		rr35
#define rr38 rr33	
	paddw	rr30, rr33;			//						rr33
	movq	rr38, fdct_chen_mmx1_c20c;	//		rr38
	pmulhw	rr31, rr35;			//						
#define	rr36 rr35
	pmulhw	rr27, fdct_chen_mmx1_c40;	//	
    pmulhw	rr30, rr35;			//						rr35
    movq	rr36, rr29;			// rr36=rr29=a2	rr36
#define rr37 rr31
	paddw	rr29, rr38;
	paddw	rr24, rr31;			// rr24=y1				rr31
  	movq	rr37, rr28;			// rr37=rr28=a1	rr37
#define rr39 rr38
	paddw	rr28, rr38;			//						rr38
	movq	rr39, fdct_chen_mmx1_c30c;	//		rr39
	psubw	rr27, rr30;			// rr27=y7				rr30
#define rr40 rr30
	movq	[ebx+8*2], rr24;		// x1=y1				rr24
	paddw	rr36, rr39;
	movq	rr40, fdct_chen_mmx1_c20;	//		rr40  
#define rr41 rr24
	paddw	rr37, rr39;			//						rr39
    movq	rr41, fdct_chen_mmx1_c30;	//		rr41
	pmulhw	rr29, rr40;

	movq	[ebx+24*2+8], rr27;		// x7=y7			rr27
	pmulhw	rr28, rr40;			//						rr40
#define zz0 rr39
	movq	zz0, [ebx+0*2];		//				zz0
	pmulhw	rr36, rr41;
#define zz1 rr27
	movq	zz1, [ebx+8*2];		//				zz1
	pmulhw	rr37, rr41;			//						rr41
 
	paddw	rr36, rr28;			// rr36=y5				rr28
	psubw	rr29, rr37;			// rr29=y3				rr37
#define zz2 rr40
#define zz4 rr41
	movq	zz2, [ebx+16*2];		//				zz2
	movq	zz4, zz0				//				zz4

	movq	[ebx+8*2+8], rr36;	// x5=y5				rr36
	punpcklwd	zz0, zz1

								// finish firrst column half
	movq	[ebx+24*2], rr29;	//						rr29
	punpckhwd	zz4, zz1			//					zz1

#define zz3	rr28	
#define zz5	rr37
#define zz6	rr36	
#define zz7	rr29	

	movq	zz3, [ebx+24*2];		//				zz3
    movq	zz5, zz2				//				zz5
	punpcklwd	zz2, zz3
	movq	zz6, zz0				//				zz6
	punpckhwd	zz5, zz3			//						zz3
	punpckldq	zz0, zz2			// zz0=(00,10,20,30)
	movq	[ebx+0*2], zz0		//						zz0
	punpckhdq	zz6, zz2			// zz6=(01,11,21,31)		zz2
	movq	[ebx+8*2], zz6		//						zz6
	movq	zz7, zz4				//				zz7
	punpckldq	zz4, zz5			// zz4=(02,12,22,32)
	punpckhdq	zz7, zz5			// zz7=(03,13,23,33)		zz5
	movq	[ebx+16*2], zz4		//						zz4
	movq	[ebx+24*2], zz7		//						zz7

#define zz8	mm0
#define zz9	mm1
#define zz10	mm2	
#define zz11	mm3	
#define zz12	mm4	
#define zz13	mm5	
#define zz14	mm6	
#define zz15	mm7	


	movq	zz8, [ebx+0*2+8];		//				zz0
	movq	zz9, [ebx+8*2+8];		//				zz1
	movq	zz10, [ebx+16*2+8];		//				zz2
	movq	zz12, zz8				//				zz4
	punpcklwd	zz8, zz9
	movq	zz11, [ebx+24*2+8];		//				zz3
	punpckhwd	zz12, zz9			//						zz1
    	movq	zz13, zz10				//				zz5
	punpcklwd	zz10, zz11
	movq	zz14, zz8				//				zz6
	punpckhwd	zz13, zz11			//						zz3
	punpckldq	zz8, zz10			// zz0=(00,10,20,30)
	movq	[ebx+0*2+8], zz8		//						zz0
	punpckhdq	zz14, zz10			// zz6=(01,11,21,31)	zz2
	movq	[ebx+8*2+8], zz14		//						zz6
	movq	zz15, zz12				//				zz7
	punpckldq	zz12, zz13			// zz4=(02,12,22,32)
	punpckhdq	zz15, zz13			// zz7=(03,13,23,33)	zz5
	movq	[ebx+16*2+8], zz12		//						zz4
	movq	[ebx+24*2+8], zz15		//						

									// lower half

#define ww0	mm0
#define ww1	mm1
#define ww2	mm2	
#define ww3	mm3	
#define ww4	mm4	
#define ww5	mm5	
#define ww6	mm6	
#define ww7	mm7	


	movq	ww0, [ebx+32*2];		//				ww0
	movq	ww1, [ebx+40*2];		//				ww1
	movq	ww2, [ebx+48*2];		//				ww2
	movq	ww4, ww0				//				ww4
	punpcklwd	ww0, ww1
	movq	ww3, [ebx+56*2];		//				ww3
	punpckhwd	ww4, ww1			//						ww1
    	movq	ww5, ww2				//				ww5
	punpcklwd	ww2, ww3
	movq	ww6, ww0				//				ww6
	punpckhwd	ww5, ww3			//						ww3
	punpckldq	ww0, ww2			// ww0=(00,10,20,30)
	movq	[ebx+32*2], ww0		//						ww0
	punpckhdq	ww6, ww2			// ww6=(01,11,21,31)		ww2
	movq	[ebx+40*2], ww6		//						ww6
	movq	ww7, ww4				//				ww7
	punpckldq	ww4, ww5			// ww4=(02,12,22,32)
	punpckhdq	ww7, ww5			// ww7=(03,13,23,33)		ww5
	movq	[ebx+48*2], ww4		//						ww4
	movq	[ebx+56*2], ww7		//						ww7

#define ww8	mm0
#define ww9	mm1
#define ww10	mm2	
#define ww11	mm3	
#define ww12	mm4	
#define ww13	mm5	
#define ww14	mm6	
#define ww15	mm7	

	movq	ww8, [ebx+32*2+8];		//				ww0
	movq	ww9, [ebx+40*2+8];		//				ww1
	movq	ww10, [ebx+48*2+8];		//				ww2
	movq	ww12, ww8				//				ww4
	punpcklwd	ww8, ww9
	movq	ww11, [ebx+56*2+8];		//				ww3
	punpckhwd	ww12, ww9			//						ww1
    	movq	ww13, ww10				//				ww5
	punpcklwd	ww10, ww11
	movq	ww14, ww8				//				ww6
	punpckhwd	ww13, ww11			//						ww3
	punpckldq	ww8, ww10			// ww0=(00,10,20,30)
	movq	[ebx+32*2+8], ww8		//						ww0
	punpckhdq	ww14, ww10			// ww6=(01,11,21,31)		ww2
	movq	[ebx+40*2+8], ww14		//						ww6
	movq	ww15, ww12				//				ww7
	punpckldq	ww12, ww13			// ww4=(02,12,22,32)
	punpckhdq	ww15, ww13			// ww7=(03,13,23,33)		ww5
	movq	[ebx+48*2+8], ww12		//						ww4
	movq	[ebx+56*2+8], ww15		//						


#define vv0 mm0
#define vv1 mm1	
#define vv2 mm2
#define vv3 mm3	
#define vv4 mm4	
#define vv5 mm5	
#define vv6 mm6	
#define vv7 mm7	

	movq	vv0, [ebx+32*2];		// vv0=x0		vv0		 						
	movq	vv1, vv0;			// vv1=x0		vv1							

	movq	vv2, [ebx+56*2+8];	// vv2=x7		vv2
	psubw	vv1, vv2;			// vv1=c3								

	paddw	vv0,vv2;			// vv0=a0				vv2
    movq	vv3, [ebx+40*2];		// vv3=x1		vv3		 						
	
    movq	vv5, [ebx+48*2+8];	// vv5=x6		vv5
	
    movq	vv4, vv3;			// vv4=x1		vv4							

	paddw	vv3, vv5;			// vv3=a1
	movq	vv6, [ebx+48*2];	// vv6=x2		vv6		 						
	psubw	vv4, vv5;			// vv4=c2				vv5
	movq	[ebx+56*2+8], vv1;	// x7=c3				vv1	
    movq	vv7, vv6;			// vv7=x2		vv7							
 #define vv8 vv2
    movq	vv8, [ebx+40*2+8];	// vv8=x5		vv8

	
	psubw	vv7, vv8;			// vv7=c1				
	
	movq	[ebx+48*2+8], vv4;	// x6=c2				vv4
 #define vv9 vv5	
	movq	vv9, [ebx+56*2];	// vv9=x3		vv9		 						
#define vv11	vv1
#define vv10	vv4		
	paddw	vv6, vv8;			// vv6=a2				vv8
    movq	vv11, [ebx+32*2+8];	// vv11=x4		vv11
    movq	vv10, vv9;			// vv10=x3		vv10							

	movq	[ebx+40*2+8],vv7;		// x5=c1				vv7
	psubw	vv10, vv11;			// vv10=c0				vv11
	paddw	vv9, vv11;			// vv9=a3
#define vv13	vv8	
	
   	movq	vv13, vv3;			// vv13=a1		vv13
	

	movq	[ebx+32*2+8], vv10;	// x4=c0				vv10
	
	paddw	vv3, vv6;				// vv3=b1

						// even pavvt, a(0,1,2,3) in vv(0,3,6,9)
#define vv12	vv7	
	psubw	vv13, vv6;			// vv13=b2				vv6
	movq	vv12, vv0;				// vv12=a0		vv12
	paddw	vv0, vv9;				// vv0=b0

						
						// b(0,1,2,3) in vv(0,3,13,12)
						// y2 and y6 fivvst, then y0 and y4
#define vv16	vv11
#define vv18	vv10		
	movq	vv16, fdct_chen_mmx1_c50c;	//		vv16
	psubw	vv12, vv9;			// vv12=b3				vv9
 	movq	vv18, fdct_chen_mmx1_c60c;	//		vv18
#define vv14	vv6
#define vv15	vv9	
	movq	vv14, vv12;			// vv14=vv12=b3	vv14
	paddw	vv12, vv16;
    movq	vv15, vv13;			// vv15=vv13=b2	vv15
#define	vv19	vv16
	pmulhw	vv12, fdct_chen_mmx1_c50;
	paddw	vv13, vv16;			//						vv16
 	paddw	vv15, vv18;			//						
#define	vv21	vv18 
	paddw	vv14, vv18;			//						vv18
	movq	vv19, fdct_chen_mmx1_c60;	//		vv19
	movq	vv21, vv0;			// vv21=vv0=b0	vv21

	pmulhw	vv13, fdct_chen_mmx1_c50;	//					
	paddw	vv0, vv3;				// b0+b1
	pmulhw	vv15, vv19;			//						
#define vv22	vv3
	psubw	vv21, vv3;			// b0-b1				vv3
	movq	vv22, fdct_chen_mmx1_c00c;	//		vv22
	pmulhw	vv14, vv19;			//						vv19
#define vv23	vv19	
	paddw	vv0, vv22;
	paddw	vv21, vv22;
	movq	vv23, fdct_chen_mmx1_c00;	//		vv23
#define	vv20	vv15
	paddw	vv12, vv15;			// vv12=y2				vv15
	movq	vv20, [ebx+48*2+8];	// vv20=c2		vv20
	psubw	vv14, vv13;			// vv14=y6				vv13
#define	vv25	vv13
	movq	[ebx+48*2], vv12;	// x2=y2				vv12
	pmulhw	vv0, vv23;			// vv0=y0
    movq	vv25, [ebx+40*2+8];	// vv25=c1		vv25
#define	vv26	vv12
	movq	vv26, vv20;			// vv26=c2;		vv26
	paddw	vv20, vv25;			// c2+c1
	psubw	vv26, vv25;			// c2-c1				vv25

	movq	[ebx+48*2+8], vv14;	// x6=y6				vv14
	psllw	vv20, 2;
	psllw	vv26, 2;

	paddw	vv20, vv22;
	paddw	vv26, vv22;			//						vv22
	pmulhw	vv21, vv23;			// vv4=y4
#define	vv24	vv25
	movq	[ebx+32*2], vv0;		// x0=y0				vv0
	pmulhw	vv20, vv23;			// vv20=b1
	movq	vv24, [ebx+32*2+8];	// vv24=c0		vv24
#define	vv27	vv14
#define	vv28	vv22
	pmulhw	vv26, vv23;			// vv26=b0				vv23
    movq	vv27, [ebx+56*2+8];	// vv27=c3		vv27
	movq	vv28, vv24;			// vv28=vv24=c0	vv28
#define vv29 vv0
	movq	[ebx+32*2+8], vv21;	// x4=y4				vv21
	movq	vv29, vv27;			// vv29=vv27=c3	vv29
    paddw	vv24, vv26;			// vv24=a0

	psubw	vv28, vv26;			// vv28=a1				vv26
	paddw	vv27, vv20;			// vv27=a3
	psubw	vv29, vv20;			// vv29=a2				vv20
	
						// odd pavvt, a(0,1,2,3)=vv(24,28,29,27)
			
						// y1 and y7 fvvom vv24 and vv27, vv28 and vv29 idle
#define	vv32	vv23
#define	vv30	vv21
#define	vv33	vv26
	movq	vv32, fdct_chen_mmx1_c40c;	//		vv32
 	movq	vv30, vv24;			// vv30=vv24=a0	vv30
	movq	vv33, fdct_chen_mmx1_c10c;	//		vv33
#define	vv31	vv20
	paddw	vv24, vv32;
	movq	vv31, vv27;			// vv31=vv27=a3	vv31
	paddw	vv27, vv32;			//						vv32
#define	vv35	vv32
    pmulhw	vv24, fdct_chen_mmx1_c40;	
	paddw	vv31, vv33;			//						 
	movq	vv35, fdct_chen_mmx1_c10;	//		vv35
#define vv38 vv33	
	paddw	vv30, vv33;			//						vv33
	movq	vv38, fdct_chen_mmx1_c20c;	//		vv38
	pmulhw	vv31, vv35;			//						
#define	vv36 vv35
	pmulhw	vv27, fdct_chen_mmx1_c40;	//	
    pmulhw	vv30, vv35;			//						vv35
    movq	vv36, vv29;			// vv36=vv29=a2	vv36
#define vv37 vv31
	paddw	vv29, vv38;
	paddw	vv24, vv31;			// vv24=y1				vv31
  	movq	vv37, vv28;			// vv37=vv28=a1	vv37
#define vv39 vv38
	paddw	vv28, vv38;			//						vv38
	movq	vv39, fdct_chen_mmx1_c30c;	//		vv39
	psubw	vv27, vv30;			// vv27=y7				vv30
#define vv40 vv30
	movq	[ebx+40*2], vv24;		// x1=y1			vv24
	paddw	vv36, vv39;
	movq	vv40, fdct_chen_mmx1_c20;	//		vv40  
#define vv41 vv24
	paddw	vv37, vv39;			//						vv39
    movq	vv41, fdct_chen_mmx1_c30;	//		vv41
	pmulhw	vv29, vv40;

	movq	[ebx+56*2+8], vv27;		// x7=y7			vv27
	pmulhw	vv28, vv40;			//						vv40

	pmulhw	vv36, vv41;

	pmulhw	vv37, vv41;			//						vv41

	paddw	vv36, vv28;			// vv36=y5				vv28
	psubw	vv29, vv37;			// vv29=y3				vv37

	movq	[ebx+40*2+8], vv36;	// x5=y5				vv36

								// finish fivvst column half


	movq	[ebx+56*2], vv29;	//						vv29






#define uu0	mm0
#define uu1	mm1
#define uu2	mm2	
#define uu3	mm3	
#define uu4	mm4	
#define uu5	mm5	
#define uu6	mm6	
#define uu7	mm7	


	movq	uu0, [ebx+32*2];		//				uu0
	movq	uu1, [ebx+40*2];		//				uu1
	movq	uu2, [ebx+48*2];		//				uu2
	movq	uu4, uu0				//				uu4
	punpcklwd	uu0, uu1
	movq	uu3, [ebx+56*2];		//				uu3
	punpckhwd	uu4, uu1			//						uu1
    	movq	uu5, uu2				//				uu5
	punpcklwd	uu2, uu3
	movq	uu6, uu0				//				uu6
	punpckhwd	uu5, uu3			//						uu3
	punpckldq	uu0, uu2			// uu0=(00,10,20,30)
	movq	[ebx+32*2], uu0		//						uu0
	punpckhdq	uu6, uu2			// uu6=(01,11,21,31)		uu2
	movq	[ebx+40*2], uu6		//						uu6
	movq	uu7, uu4				//				uu7
	punpckldq	uu4, uu5			// uu4=(02,12,22,32)
	punpckhdq	uu7, uu5			// uu7=(03,13,23,33)		uu5
	movq	[ebx+48*2], uu4		//						uu4
	movq	[ebx+56*2], uu7		//						uu7

#define uu8	mm0
#define uu9	mm1
#define uu10	mm2	
#define uu11	mm3	
#define uu12	mm4	
#define uu13	mm5	
#define uu14	mm6	
#define uu15	mm7	

	movq	uu8, [ebx+32*2+8];		//				uu0
	movq	uu9, [ebx+40*2+8];		//				uu1
	movq	uu10, [ebx+48*2+8];		//				uu2
	movq	uu12, uu8				//				uu4
	punpcklwd	uu8, uu9
	movq	uu11, [ebx+56*2+8];		//				uu3
	punpckhwd	uu12, uu9			//						uu1
    	movq	uu13, uu10				//				uu5
	punpcklwd	uu10, uu11
	movq	uu14, uu8				//				uu6
	punpckhwd	uu13, uu11			//						uu3
	punpckldq	uu8, uu10			// uu0=(00,10,20,30)
	movq	[ebx+32*2+8], uu8		//						uu0
	punpckhdq	uu14, uu10			// uu6=(01,11,21,31)		uu2
	movq	[ebx+40*2+8], uu14		//						uu6
	movq	uu15, uu12				//				uu7
	punpckldq	uu12, uu13			// uu4=(02,12,22,32)
	punpckhdq	uu15, uu13			// uu7=(03,13,23,33)		uu5
	movq	[ebx+48*2+8], uu12		//						uu4
	movq	[ebx+56*2+8], uu15		//						




	emms
  }



  









  
 

}


#define NUM_BLOCK_TEST 10000000
void fdct_chen_mmx1_bench()
{ int i,j;
  short int block1[64+16],block2[64+16];
  short int *b1=block1, *b2=block2;
  clock_t start,stop;
  
  __asm									// alignment
  { mov esi, b1
    add esi, 31
	and	esi,-32
	mov b1, esi
	mov esi, b2
    add esi, 31
	and	esi,-32
	mov b2, esi
  }

  for (j=0;j<64;j++)  block1[j]=rand()>>7;
  start=clock();
  for (i=0;i<NUM_BLOCK_TEST;i++) 
    fdct_chen_mmx1(block1);
  stop=clock();
  printf("%d ms\n",(stop-start)*1000/CLOCKS_PER_SEC);
}

