/*
 ---------------------------------------------------------
 idct_ap922hybr.c - AP922hybrid iDCT, C-code implementation
 ---------------------------------------------------------

 AP-922 "hybrid" float/int32 iDCT (x87 FPU)
 ----------------------------------------
 
    Intel Application Note AP-922 discusses an MMX implementation of 
  the MPEG fDCT/iDCT operations.  Numerical analysis of the AP922 iDCT
  reveals that the column_idct operator accounts for most of the
  output error-magnitude.  (Disclaimer, I did not perform any formal 
  numerical analysis...)
 
    "AP922hybrid" is an adaptation of the original AP922 algorithm.
  The 'hybrid' modification combines the original AP922 scaled_integer 
  row_idct operator with a new floating-point column_idct operator.  
  In terms of both speed and accuracy, the 3D-Now implementation of 
  AP922hybrid falls between AP922float(3DNow) and the original AP922mmx.
 
  This file contains two implementations of AP922hybrid :
    1) 3D-Now accelerated
    2) standard C (X87)
   
    The X87 implementation (in this file) should compile and run on any
  i486DX or higher CPU.  The standard-C version was coded to emulate the 
  3D-Now version as much as possible.

    The 3DNow implementation requires a 3D-Now capable CPU, such as
  the AMD K6-2 or Athlon.  The provided 3DNow code was developed under
  Visual C++ 6.0 Pro.  To use the 3DNow code 'as is', and will need
  VC++ 5/6 (and 'amd3dx.h', which is included.)  The Microsoft 
  Visual C++ 6.0 Processor Pack Beta is *not* needed.

    Since both AP922float and AP922hybrid use floating point math, 
  AP922hybrid shares the same set of numerical issues:

 Numerical Issue #1 : float->int conversion
 ------------------------------------------

    Currently, the 3D-Now instruction set only supports truncation
  for float->int conversion.  Therefore, accurate_rounding is
  disabled for the X87 implementation.  

  To minimize the truncation error, this version of AP922 utilizes the
  same strategy discussed in idct_ap922x87.c  Please refer to that file.


 Numerical precision of 3DNow versus X87(FPU)
 ---------------------------------------
    On 80x86 CPUs, FPU add/multiply operations are always 
  calculated with 80-bit internal precision.  Therefore, all other things 
  being equal, the X87 DCT will be more accurate than AMD 3D-Now, even
  when the X87 DCT stores intermediate results in 32-bit float.

    The statement 'typedef float tdnfloat;' fixes the x87's computational
  accuracy to 32-bit float.  AP922hybrid uses a combination of 
  scaled-integer and floating-point math, with the scaled-integer math
  responsible for nearly all output error. Therefore, switching to 
  tdnfloat to a more accurate type (like "double") does not appreciably
  improve accuracy.

 
 ------------------------------------------
 LEGAL MUMBO-JUMBO : Disclaimer of Warranty
 ------------------------------------------
 This software is available to the user without any license fee or
 royalty on an "as is" basis.  The author disclaims any and all warranties,
 whether express, implied, or statuary, including any implied warranties or 
 merchantability or of fitness for a particular purpose.  In no event shall 
 the copyright-holder be liable for any incidental, punitive, or consequential
 damages of any kind whatsoever arising from the use of these programs.

 This disclaimer of warranty extends to the user of these programs and user's
 customers, employees, agents, transferees, successors, and assigns.

 Revision history
 ----------------
 09/03/2000 initial release of 'AP922hybrid' iDCT
            Based on AP922, this iDCT uses the original scaled-int (MMX) row
            iDCT operator, paired with an all new floating-point column iDCT.
            To the best of my knowledge, AP922hybr is fully IEEE 1180/1990 
            compliant iDCT, including output-range clipping to {-256,+255}

            This file contains inline-assembly code developed under 
            Visual C++ 6.0 Pro.
            
 09/03/2000 liaor@iname.com   http://members.tripod.com/~liaor
 liaor@iname.com   http://members.tripod.com/~liaor
*/

#include "amd3dx.h" // inline-assembly macros for AMD 3D-Now

typedef float tdnfloat;  // 3D-Now float = IEEE-754 32-bit float

// comment out the following line to use accurate (slower) rounding
#define USE_FAST_FLOAT2INT 1

  #define DESCALE_SHIFT 17   // Total prescale-shift : 16+1 bit right-shift
  #define DESCALE_SHIFT2 7 // 2nd stage descale-shift : 9-bit output range
  #define DESCALE_SHIFT1 (DESCALE_SHIFT-DESCALE_SHIFT2) // 1st stage descale
  #define DESCALE_ROUND_COMPENSATION (1<<(DESCALE_SHIFT-1))

#ifdef USE_FAST_FLOAT2INT

  // approximate rounding
  #define SHIFT_ROUND_COLF( x )  ( ( ( (int)(x) ) + DESCALE_ROUND_COMPENSATION) >> DESCALE_SHIFT)

#else

  "DELIBERATE_SYNTAX_ERROR______ACCURATE_ROUNDING_IS_NOT_SUPPORTED";
  // accurate rounding is NOT supported
  #define SHIFT_ROUND_COLF( x )  (  ( ((int)floor( (x) )) + DESCALE_ROUND_COMPENSATION )>>DESCALE_SHIFT  ) 

#endif // USE_FAST_FLOAT2INT


//
// CONCATENATED iDCT ROW COEFF TABLE, rows 0,1,2,3,4,5,6,7 (in order )
// copied directly from Intel Application note AP-922
const static short tab_i_01234567hybr[] = {
  //row0, this row is required
  16384, 16384, 16384, -16384,  //; movq-> w06 w04 w02 w00
  21407, 8867, 8867, -21407,    //; w07 w05 w03 w01
  16384, -16384, 16384, 16384,  //; w14 w12 w10 w08
  -8867, 21407, -21407, -8867,  //; w15 w13 w11 w09
  22725, 12873, 19266, -22725,  //; w22 w20 w18 w16
  19266, 4520, -4520, -12873,   //; w23 w21 w19 w17
  12873, 4520, 4520, 19266,     //; w30 w28 w26 w24
  -22725, 19266, -12873, -22725,//; w31 w29 w27 w25

  //row1
  22725, 22725, 22725, -22725,  //; movq-> w06 w04 w02 w00
  29692, 12299, 12299, -29692,  //; w07 w05 w03 w01
  22725, -22725, 22725, 22725,  //; w14 w12 w10 w08
  -12299, 29692, -29692, -12299,//; w15 w13 w11 w09
  31521, 17855, 26722, -31521,  //; w22 w20 w18 w16
  26722, 6270, -6270, -17855,   //; w23 w21 w19 w17
  17855, 6270, 6270, 26722,     //; w30 w28 w26 w24
  -31521, 26722, -17855, -31521,//  w31 w29 w27 w25

  //row2
  21407, 21407, 21407, -21407,  // ; movq-> w06 w04 w02 w00
  27969, 11585, 11585, -27969,  // ;w07 w05 w03 w01
  21407, -21407, 21407, 21407,  // ;w14 w12 w10 w08
  -11585, 27969, -27969, -11585,// ;w15 w13 w11 w09
  29692, 16819, 25172, -29692,  // ;w22 w20 w18 w16
  25172, 5906, -5906, -16819,   // ;w23 w21 w19 w17
  16819, 5906, 5906, 25172,     // ;w30 w28 w26 w24
  -29692, 25172, -16819, -29692,// ;w31 w29 w27 w25

  //row3
  19266, 19266, 19266, -19266,  //; movq-> w06 w04 w02 w00
  25172, 10426, 10426, -25172,  //; w07 w05 w03 w01
  19266, -19266, 19266, 19266,  //; w14 w12 w10 w08
  -10426, 25172, -25172, -10426,//; w15 w13 w11 w09
  26722, 15137, 22654, -26722,  //; w22 w20 w18 w16
  22654, 5315, -5315, -15137,   //; w23 w21 w19 w17
  15137, 5315, 5315, 22654,     //; w30 w28 w26 w24
  -26722, 22654, -15137, -26722,//; w31 w29 w27 w25

  //row4
  16384, 16384, 16384, -16384,  //; movq-> w06 w04 w02 w00
  21407, 8867, 8867, -21407,    //  w07 w05 w03 w01
  16384, -16384, 16384, 16384,  //; w14 w12 w10 w08
  -8867, 21407, -21407, -8867,  //; w15 w13 w11 w09
  22725, 12873, 19266, -22725,  //; w22 w20 w18 w16
  19266, 4520, -4520, -12873,   //; w23 w21 w19 w17
  12873, 4520, 4520, 19266,     //; w30 w28 w26 w24
  -22725, 19266, -12873, -22725,//  w31 w29 w27 w25

  //row5
  19266, 19266, 19266, -19266,  //; movq-> w06 w04 w02 w00
  25172, 10426, 10426, -25172,  //; w07 w05 w03 w01
  19266, -19266, 19266, 19266,  //; w14 w12 w10 w08
  -10426, 25172, -25172, -10426,//; w15 w13 w11 w09
  26722, 15137, 22654, -26722,  //; w22 w20 w18 w16
  22654, 5315, -5315, -15137,   //; w23 w21 w19 w17
  15137, 5315, 5315, 22654,     //; w30 w28 w26 w24
  -26722, 22654, -15137, -26722,//; w31 w29 w27 w25

  //row6
  21407, 21407, 21407, -21407,  // ;movq-> w06 w04 w02 w00
  27969, 11585, 11585, -27969,  // ;w07 w05 w03 w01
  21407, -21407, 21407, 21407,  // ;w14 w12 w10 w08
  -11585, 27969, -27969, -11585,// ;w15 w13 w11 w09
  29692, 16819, 25172, -29692,  // ;w22 w20 w18 w16
  25172, 5906, -5906, -16819,   // ;w23 w21 w19 w17
  16819, 5906, 5906, 25172,     // ;w30 w28 w26 w24
  -29692, 25172, -16819, -29692,// ;w31 w29 w27 w25

  //row7
  22725, 22725, 22725, -22725,  //; movq-> w06 w04 w02 w00
  29692, 12299, 12299, -29692,  //; w07 w05 w03 w01
  22725, -22725, 22725, 22725,  //; w14 w12 w10 w08
  -12299, 29692, -29692, -12299,//; w15 w13 w11 w09
  31521, 17855, 26722, -31521,  //; w22 w20 w18 w16
  26722, 6270, -6270, -17855,   //; w23 w21 w19 w17
  17855, 6270, 6270, 26722,     //; w30 w28 w26 w24
  -31521, 26722, -17855, -31521 //; w31 w29 w27 w25
};

#define NCOS1_16 (0.980785280403230449126182236134239) // cosine(  Pi/16 )
#define NCOS2_16 (0.923879532511286756128183189396788) // cosine( 2Pi/16 )
#define NCOS3_16 (0.831469612302545237078788377617906) // cosine( 3Pi/16 )
#define NCOS4_16 (0.707106781186547524400844362104849) // cosine( 4Pi/16 )
#define NCOS5_16 (0.555570233019602224742830813948533) // cosine( 5Pi/16 )
#define NCOS6_16 (0.382683432365089771728459984030399) // cosine( 6Pi/16 )
#define NCOS7_16 (0.195090322016128267848284868477022) // cosine( 7Pi/16 )

#define TANG1_16 ( NCOS7_16 / NCOS1_16) // tangent( Pi/16)
#define TANG2_16 ( NCOS6_16 / NCOS2_16) // tangent(2Pi/16)
#define TANG3_16 ( NCOS5_16 / NCOS3_16) // tangent(3Pi/16)

// externally allocated array of 64 floats, for temporary storage
extern void *fTempArray;

const static tdnfloat tab_i_col[]=
{ 
  NCOS4_16,  NCOS4_16,   // cosine  (4Pi/16)
  TANG1_16,  TANG1_16,   // tangent ( Pi/16)
  TANG2_16,  TANG2_16,   // tangent (2Pi/16)
  TANG3_16,  TANG3_16    // tangent (3Pi/16)
};

const static tdnfloat tg_1_16f = TANG1_16; // scalar versions of tab_i_col[]
const static tdnfloat tg_2_16f = TANG2_16;
const static tdnfloat tg_3_16f = TANG3_16;
const static tdnfloat cos_4_16f= NCOS4_16;

// Rounding compensation for final output (in column_iDCT operator)
const static int rnd_compensation[2] = { DESCALE_ROUND_COMPENSATION,DESCALE_ROUND_COMPENSATION};



////////////////////////////////////////////////////////////////////////
//
//
// AP922hybr_3dn(), AMD 3D-Now implementation of AP922hybrid
//
//
////////////////////////////////////////////////////////////////////////
void
idct_ap922hybr_3dn( short *data )
{
  __asm {

#define INPR  eax
#define OUTR  edx
#define TABLE ecx
  mov INPR, dword ptr [data];   ;// row 0
   mov edi, 0x00; //x = 0

  lea TABLE, dword ptr [tab_i_01234567hybr]; // row 0
//   mov OUT, INP;  // algorithm writes data in-place  -> row 0

  mov OUTR, dword ptr [fTempArray];

  // for ( x = 0; x < 8; ++x )  // transform one row per iteration
ALIGN 16
acc_idct_rowloop1:
  movq mm0, qword ptr [INPR] ;   // 0 ; x3 x2 x1 x0

  movq mm1, qword ptr [INPR+8] ; // 1 ; x7 x6 x5 x4
   movq mm2, mm0 ;        // 2 ; x3 x2 x1 x0

  movq mm3, qword ptr [TABLE] ; // 3 ; w06 w04 w02 w00
   punpcklwd mm0, mm1 ;     // x5 x1 x4 x0

  movq mm5, mm0 ;         // 5 ; x5 x1 x4 x0
   punpckldq mm0, mm0 ;     // x4 x0 x4 x0

  movq mm4, qword ptr [TABLE+8] ; // 4 ; w07 w05 w03 w01
   punpckhwd mm2, mm1 ;     // 1 ; x7 x3 x6 x2

  pmaddwd mm3, mm0 ;        // x4*w06+x0*w04 x4*w02+x0*w00
   movq mm6, mm2 ;        // 6 ; x7 x3 x6 x2

  movq mm1, qword ptr [TABLE+32] ;// 1 ; w22 w20 w18 w16
   punpckldq mm2, mm2 ;     // x6 x2 x6 x2

  pmaddwd mm4, mm2 ;        // x6*w07+x2*w05 x6*w03+x2*w01
   punpckhdq mm5, mm5 ;     // x5 x1 x5 x1

  pmaddwd mm0, qword ptr [TABLE+16] ;// x4*w14+x0*w12 x4*w10+x0*w08
   punpckhdq mm6, mm6 ;     // x7 x3 x7 x3

  movq mm7, qword ptr [TABLE+40] ;// 7 ; w23 w21 w19 w17
   pmaddwd mm1, mm5 ;       // x5*w22+x1*w20 x5*w18+x1*w16

  pmaddwd mm7, mm6 ;       // x7*w23+x3*w21 x7*w19+x3*w17
   add INPR, 16;          // increment INPUT pointer -> row 1

  pmaddwd mm2, qword ptr [TABLE+24] ;// x6*w15+x2*w13 x6*w11+x2*w09
   paddd mm3, mm4 ;       // 4 ; a1=sum(even1) a0=sum(even0)

  pmaddwd mm5, qword ptr [TABLE+48] ;// x5*w30+x1*w28 x5*w26+x1*w24
   movq mm4, mm3 ;        // 4 ; a1 a0

  pmaddwd mm6, qword ptr [TABLE+56] ;// x7*w31+x3*w29 x7*w27+x3*w25
   paddd mm1, mm7 ;       // 7 ; b1=sum(odd1) b0=sum(odd0)

  psubd mm3, mm1 ;        //     a1-b1 a0-b0  [y6 y7] <int32>
   paddd mm0, mm2 ;        // 2 ; a3=sum(even3) a2=sum(even2)

  paddd mm1, mm4 ;       // 4 ; a1+b1 a0+b0  [y1 y0] <int32>
   punpckldq mm7, mm3;    //    mm7 <= [y7 XX] <int32>

  PI2FD( mm1, mm1);       // convert [y1 y0] to <float>
   punpckhdq mm3, mm7;    //    mm3 <=[y7 y6] <int32>

  PI2FD( mm3, mm3);       // convert [y7 y6] to <float>
   paddd mm5, mm6 ;        // 6 ; b3=sum(odd3) b2=sum(odd2)

  movq mm4, mm0 ;        // 4 ; a3 a2
   psubd mm0, mm5 ;       // 5 ; a3-b3 a2-b2  [y4 y5] <int32>

  paddd mm4, mm5 ;        // a3+b3 a2+b2      [y3 y2] <int32>
   punpckldq mm6, mm0;    //   mm6 <= [y5 XX] <int32>

  PI2FD( mm4, mm4);       // convert [y3 y2] to <float>
   punpckhdq mm0, mm6;    //   mm0 <= [y5 y4] <int32>

  movq qword ptr [OUTR], mm1 ; // 1 ; save y1 y0
   PI2FD( mm0, mm0);       // convert [y5 y4] to <float>

  movq qword ptr [OUTR+24], mm3 ; // 7 ; save y7 y6
   add TABLE, 64;         // ptr TABLE += 64 -> row 1

  movq qword ptr [OUTR+8], mm4 ; // 7 ; save y3 y2
   add edi, 0x01;

  movq qword ptr [OUTR+16], mm0 ; // 7 ; save y5 y4
   add OUTR, 32;          // increment OUTPUT pointer -> row 1

  cmp edi, 0x08;
   jl acc_idct_rowloop1;   // end for ( x = 0; x < 8; ++x )  

//  FEMMS;
//  }
  // done with the iDCT row-transformation
  ////////////////////////////////////////////////////////////////////////
  //
  // AP922float iDCT column transform
  // Base code originally from VirtualDub (Avery Lee),
  //   http://www.concentric.net/~psilon



;/////////////////////////////////////////////////////////////////////
;//
;//
;// iDCT column transform
;//
;//
;/////////////////////////////////////////////////////////////////////


#define SCRATCH esi
#define INPC edx
#define OUTC eax
    mov OUTC, dword ptr [data];   ;// row 0
     mov edi, 0x00; // i = 0

    mov INPC, dword ptr [fTempArray];
     sub OUTC, 4;   // precompensate out

#define _PRND_INV_ROW TABLE

#define _OCOS4        0
#define _OTAN1        8
#define _OTAN2        16
#define _OTAN3        24
#define _ORND_INV_ROW 32

    lea TABLE, [tab_i_col];
     mov SCRATCH, INPC; // after we read the first input column, use as scratch

     // for ( x = 0; x < 8; x=x+2 )   // process two columns per pass
ALIGN 16
acc_idct_colloop1:
    movq mm0, [INPC + 1*32] ; // (1) xmm0 <= dp[4]
     add edi, 2;              // x = x +2 (two columns per pass)

    movq mm2, [INPC + 3*32];  // (1) xmm2 <= dp[6]

    movq mm1, [INPC + 7*32];  // (1) xmm1 <= dp[5]
     movq mm4, mm0;          // (1) xmm4 <= dp[4]

    movq mm3, [INPC + 5*32] ; // (1) xmm3 <= dp[7]
     movq mm6, mm2;         ;// (2a) xmm6 <= dp[6]
    ;// 1) Apply (D_T * P_T) - the cos() coefficients of D_T are implicit
    ;//    in the idct_row operation.  But we still need to apply the
    ;//    shuffling operation of D_T.
    ;//
    ;// dp[0] = xc[ 0 *8];
    ;// dp[1] = xc[ 4 *8];
    ;// dp[2] = xc[ 2 *8];
    ;// dp[3] = xc[ 6 *8];

    ;// dp[4] = xc[ 1 *8];
    ;// dp[5] = xc[ 7 *8];
    ;// dp[6] = xc[ 3 *8];
    ;// dp[7] = xc[ 5 *8];

    // 2) Apply B_T
    //
    //    b[4] =   dp[4] + ( dp[5]*tg_1_16f );
    //    b[5] = - dp[5] + ( dp[4]*tg_1_16f );
    //    b[6] =   dp[6] + ( dp[7]*tg_3_16f );
    //    b[7] = - dp[7] + ( dp[6]*tg_3_16f );

    PFMULM( mm4, TABLE, _OTAN1);    ;// (2a) xmm4 <= (dp[4]*tg_1_16f)
     movq mm5, mm1       ;// (2a) xmm5 <= dp[5] 

    PFMULM( mm5, TABLE, _OTAN1);    ;// (2a) xmm5 <= (dp[5]*tg_1_16f)
     movq mm7, mm3;      ;// (2a) xmm7 <= dp[7]

    PFMULM( mm6, TABLE, _OTAN3);    ;// (2a) xmm6 <= (dp[6]*tg_3_16f)
     PFSUB( mm4, mm1);        ;// (2b) xmm4 <= b[5] = (dp[4]*tg_1_16f) - dp[5]

// xmm1 free

    PFMULM( mm7, TABLE, _OTAN3);    ;// (2a) xmm7 <= (dp[7]*tg_3_16f)
     PFADD( mm5, mm0);        ;// (2b) xmm5 <= b[4] = (dp[5]*tg_1_16f) + dp[4]
// xmm0 free

    PFSUB( mm6, mm3);        ;// (2b) xmm6 <= b[7] = (dp[6]*tg_1_16f) - dp[7]
// xmm3 free
// xmm1 used
     movq mm1, mm4;      ;// (2b) xmm1 <= b[5]

// xmm3 used
    movq mm3, [INPC + 6*32]   ;// (2a) xmm3 <= dp[3]
     PFADD( mm7, mm2);         // (2b) xmm7 <= b[6] = (dp[7]*tg_1_16f) + dp[6]
// xmm2 free

// xmm0 used
    movq mm0, mm5;      ;// (2b) xmm0 <= b[4]

    // 3) Apply E_T
    //
    //     e[4] = b[4] + b[6];
    //     e[5] = b[4] - b[6];
    //     e[6] = b[5] + b[7];
    //     e[7] = b[5] - b[7];
     PFADD( mm1, mm6 );          ;// (3) xmm1 <= e[6] = b[5] + b[7]

// xmm2 used
    movq mm2, mm3;          ;// (2a) xmm2 <= dp[3]
     PFSUB( mm5, mm7);           ;// (3) xmm5 <= e[5] = b[4] - b[6]

    PFMULM( mm3, TABLE, _OTAN2);       ;// (2a) xmm3 <= (dp[3]*tg_2_16f)
     PFSUB( mm4, mm6);           ;// (3) xmm4 <= e[7] = b[5] - b[7]
// xmm6 free

// xmm6 used
    movq mm6, [INPC + 2*32] ;// (2a) xmm6 <= dp[2]
     PFADD( mm0, mm7);           ;// (3) xmm0 <= e[4] = b[4] + b[6]
// xmm7 free

// xmm7 used
    movq mm7, mm5;         ;// (3) xmm7 <= e[5]

    // 4) Apply F_T
    //
    //    f[4] = e[4]
    //    f[5] = (e[5] + e[6]) * cos_4_16f;
    //    f[6] = (e[5] - e[6]) * cos_4_16f;
    //    f[7] = e[7]

// xmm0 <= f[4] = e[4]
// xmm4 <= f[7] = e[7]
     PFADD( mm5, mm1 );           ;// (4a) xmm5 <= (e[5]+e[6])

    PFSUB( mm7, mm1 );           ;// (4a) xmm7 <= (e[5]-e[6])
// xmm1 free
// xmm1 used
     movq mm1, [INPC + 0*32] ;// (1) xmm1 <= dp[0]

movq [SCRATCH], mm0;         ;// store f[4] in scratch space
// xmm0 free (in scratch)

    PFMULM( mm5, TABLE, _OCOS4);     ;// (4b) xmm5 <= f[5] = (e[5] + e[6]) * cos_4_16f;
     PFADD( mm3, mm6);       ;// (2b) xmm3 <= b[2] = (dp[3]*tg_2_16f) + dp[2]

// xmm4 f[7] free (in scratch)

    PFMULM( mm6, TABLE, _OTAN2);       ;// (2a) xmm6 <= (dp[2]*tg_2_16f)
     ;

    PFMULM( mm7, TABLE, _OCOS4);        ;// (4b) xmm7 <= f[6] = (e[5] - e[6]) * cos_4_16f;
// xmm0 used
     movq mm0, mm1;          ;// (1) xmm0 <= dp[0]
    
    PFSUB( mm6, mm2 );           ;// (2b) xmm6 <= b[3] = (dp[2]*tg_2_16f) - dp[3]
// xmm2 free
// xmm2 used
     movq mm2, [INPC+4*32];  ;// (1) xmm2 <= dp[1]

    PFADD( mm1, mm2 );           ;// (2b) xmm1 <= b[0] = dp[0] + dp[1]
     add INPC, 8;

    PFSUB( mm0, mm2 );           ;// (2b) xmm0 <= b[1] = dp[0] - dp[1]
// xmm2 free
     add OUTC, 4;

//    e[0] = b[0] + b[2];
//    e[1] = b[1] + b[3];
//    e[2] = b[1] - b[3];
//    e[3] = b[0] - b[2];


// xmm2 used 
    movq mm2, mm1;          ;// (2b) xmm1 <= b[0]
     PFADD( mm1, mm3);          ;// (3) xmm1 <= e[0] = b[0] + b[2];

    PFSUB( mm2, mm3);           ;// (3) xmm3 <= e[3] = b[0] - b[2];
// xmm3 free
// xmm3 used
     movq mm3, mm0;         ;// (2b) xmm3 <= b[1]

    PFADD( mm0, mm6);           ;// (3) xmm0 <= e[1] = b[1] + b[3];
     ;

    PFSUB( mm3, mm6);           ;// (3) xmm3 <= e[2] = b[1] - b[3];
// xmm6 free
// xmm6 used
     movq mm6, mm2;         ;// (4) xmm6 <= f[3]

// xmm1 <= f[0]
// xmm0 <= f[1]
// xmm3 <= f[2]
// xmm2 <= f[3]
/*

    y[0*8] = SHIFT_ROUND_COLF( _F0 + _F4 );
    y[1*8] = SHIFT_ROUND_COLF( _F1 + _F5 );
    y[2*8] = SHIFT_ROUND_COLF( _F2 + _F6 );
    y[3*8] = SHIFT_ROUND_COLF( _F3 + _F7 );

    y[4*8] = SHIFT_ROUND_COLF( _F3 - _F7 );
    y[5*8] = SHIFT_ROUND_COLF( _F2 - _F6 );
    y[6*8] = SHIFT_ROUND_COLF( _F1 - _F5 );
    y[7*8] = SHIFT_ROUND_COLF( _F0 - _F4 );

    6a) float -> int32 conversion (cvtps2pi)
       pf2id mm0, mm0;  // mm0 <= int32 versions of y[1 0]

    6b) round compensation, adjust int32 value to correct round alignment
        paddd mm0, [rnd_compensation]

    6c) stage1 descaling shift
        psrad mm0, DESCALE_SHIFT1

    6d) int32 -> int16 conversion (packssdw)
        packssdw mm0, mm0; // mm0 <= int16 versions of y[1 0 1 0] (duplicates)

    6e) stage2 descaling shift (psraw)
       psraw mm0, DESCALE_SHIFT2; // mm0 <= final descaled output, {-256,+255}

    6f) store results
    ** The descale1/descale2 shifts are chosen to clip final result to 
       {-256,+255} (9-bit range)

*/ 

    PFADD( mm2, mm4);           ;// (5) xmm2 <= y[3] = f[3] + f[7]
     ;

    PFSUB( mm6, mm4);           ;// (5) xmm6 <= y[4] = f[3] - f[7]
// xmm4 free
// xmm4 used
     movq mm4, mm0;         ;// xmm0 <= f[1]


    PF2ID( mm2, mm2);           ;// (6a) [y3] <float> -> <int32>
     ;

//    addps xmm2, [_PRND_INV_ROW]; // (6b) ROUND compensation

    PFADD( mm0, mm5);           ;// (5) xmm0 <= y[1] = f[1] + f[5]
     ;

    paddd mm2, [ rnd_compensation ] ;// (6b) ROUND compensation
     PF2ID( mm0, mm0 )              ;// (6a) [y1] <float> -> <int32>

    psrad mm2, DESCALE_SHIFT1;
     PFSUB( mm4, mm5);            ;// (5) xmm4 <= y[6] = f[1] - f[5]
// xmm5 free

    paddd mm0, [ rnd_compensation ] ;// (6b) ROUND compensation
// xmm5 used
     movq mm5, mm3;          ;// xmm5 <= f[2]

    psrad mm0, DESCALE_SHIFT1;
     PF2ID( mm6, mm6 );      ;// (6a) [y4] <float> -> <int32>

    packssdw mm0, mm2;           ;// (6d) mm0 <= [ y[3] | y[1]]
     PFADD( mm3, mm7 );            ;// (5) xmm3 <= y[2] = f[2] + f[6]
// xmm2 free
    
    psraw mm0, DESCALE_SHIFT2;    // (6e) clip y[3,1] to {-256,+255}
     PFSUB( mm5, mm7);            ;// (5) xmm5 <= y[5] = f[2] - f[6]

// xmm7 free
    paddd mm6, [ rnd_compensation ] ;// (6b) ROUND compensation
     PF2ID( mm3, mm3 );      ;// (6a) [y2] <float> -> <int32>

    movd [OUTC + 1*16], mm0;    l// (6f) store y[1]  (store y[3] next)
     PF2ID( mm5, mm5 );      ;// (6a) [y5] <float> -> <int32>

    movq mm2, [SCRATCH];     ;// xmm2 <= f[4]
     psrad mm6, DESCALE_SHIFT1;

    punpckhdq mm0, mm0;         // (6f) mm0 = [y3c1 y3c0 y3c1 y3c0] <int16>
     movq mm7, mm1;           ;// xmm7 <= f[0]

    paddd mm3, [rnd_compensation];  ;// (6b) ROUND compensation
     PFADD( mm1, mm2);            ;// (5) xmm2 <= y[0] = f[0] + f[4]

    movd [OUTC + 3*16], mm0;    // (6f) store y[1]  (store y[3] next)
// mm0 free
     PFSUB( mm7, mm2);            ;// (5) xmm7 <= y[7] = f[0] - f[4]

    psrad mm3, DESCALE_SHIFT1;
     PF2ID( mm1, mm1)            ;// (6a) [y0] <float> -> <int32>

    packssdw mm3, mm6;           ;// (6d) mm3 <= [ y[4] y[2] ]
     PF2ID( mm7, mm7)            ;// (6a) [y7] <float> -> <int32>

    paddd mm5, [rnd_compensation];  ;// (6b) ROUND compensation
     PF2ID( mm4, mm4)            ;// (6a) [y6] <float> -> <int32>

    paddd mm1, [rnd_compensation];  ;// (6b) ROUND compensation
     psraw mm3, DESCALE_SHIFT2;    // (6e) clip y[4,2] to {-256,+255}

    paddd mm7, [rnd_compensation];  ;// (6b) ROUND compensation
     psrad mm5, DESCALE_SHIFT1;        

    paddd mm4, [rnd_compensation];  ;// (6b) ROUND compensation
     psrad mm1, DESCALE_SHIFT1;        

    movd [ OUTC + 2*16], mm3;    // (6f) store y[2]  (store y[4] next)
     psrad mm4, DESCALE_SHIFT1;

// mm1 free
    punpckhdq mm3, mm3;            ;//(6f) mm3 = [y4c1 y4c0 y4c1 y4c0] <int16>
     psrad mm7, DESCALE_SHIFT1;        

    packssdw mm5, mm1;            ;// (6d) mm5 = [y0 y5] <int16>
     packssdw mm7, mm4;            ;// (6d) mm7 = [y6 y7] <int16>

    movd [ OUTC + 4*16], mm3;    // (6f) store y[4]
     psraw mm5, DESCALE_SHIFT2;    // (6e) clip y[5,0] to {-256,+255}

    cmp edi, 8;
     psraw mm7, DESCALE_SHIFT2;    // (6e) clip y[6,7] to {-256,+255}

    movd [ OUTC + 5*16], mm5;    // (6f) store y[5]
     punpckhdq mm5, mm5;            ;// (6f) mm3 = [y0c1 y0c0 y0c1 y0c0] <int16>

    movd [ OUTC + 7*16], mm7;    // (6f) store y[7]
     punpckhdq mm7, mm7;            ;// (6f) mm3 = [y6c1 y6c0 y6c1 y6c0] <int16>

    movd [ OUTC + 0*16], mm5;    // (6f) store y[0]

    movd [ OUTC + 6*16], mm7;    // (6f) store y[6]

    jl acc_idct_colloop1;
  // end for ( x=0; x < 8; x=x+2 )

   FEMMS;
  }
}


////////////////////////////////////////////////////////////////////////
//
//
// AP922hybr_x87(), conventional x87 implementation of hybrid AP922
//
//
////////////////////////////////////////////////////////////////////////

void
idct_ap922hybr_x87( short *data )
{
//  static tdnfloat fTempArray[64]; // intermediate (row-iDCT output) matrix
  static int ar[8], br[8];
  static tdnfloat b[8], e[8], dp[8]; // intermediate results
  static tdnfloat tmp[2];  // temp calculation, scratch pad

  static short xi[8];  // temp 16-bit ints for initial P_T operator

//  static short const _one_corr=0x0001; 
  short *ti;    // pointer to IDCT row coefficient table
 
  short *x;     // pointer to input
  short *y;     // pointer to final output
  tdnfloat *yr, *xc; // pointer to output row

  int i;

  ////////////////////////////////////////////////////////////////////////
  //
  // AP922 iDCT row transform (int32 version)
  //   Core transform code is very similar to the original AP922mmx iDCT_row
  //   except full 32-bit output is preserved (instead of shift/rnd to 16-bit)
  //    
  //
  // Row IDCT operator :A_T*M_T*P_T
  // Let Y=[output column data, 8 elements] 32-bit IEEE-754 float
  //     X=[input column data, 8 elements] 16-bit short integer
  //
  //     Y= [ A_T*M_T*P_T ] * X
  //
  //   (Y and X are both column vectors)


  for ( i = 0; i < 8; ++i ) // row iDCT
  {
    ti = &tab_i_01234567hybr[ i * 32 ];
    x = &data[ 8*i ];
    yr = ((tdnfloat *) fTempArray) + 8*i; // intermediate output, select row#i

    //
    // 1) Apply M_T * P_T operator 
    //    

    // The operation "output(32-bit) = x(16-bit) * y(16-bit)" produces a
    // 32-bit result.  Later, the 32-bit result is converted to float,
    // which means the prior integer-mult effectively scales up the
    // result by 65536 (left shift 16.)

    // ar[3..0], br[3..0] = 32-bit integers
    //
    // The AP922hybrid iDCT_row computes ar[3..0] & br[3..0] without
    // any post-shift/round.  As a result, the pmaddwd/paddd sequence
    // has the side-effect of upscaling by a factor of 2 (left shift by 1.)
    //
    // This effect is reflected in the constant PSCALE_SHIFT, which has 
    // been adjusted upward from 16 to 17.

    ar[0] = x[0] * ti[0]  + x[4] * ti[1]  + x[2] * ti[4]  + x[6] * ti[5];
    ar[1] = x[0] * ti[2]  + x[4] * ti[3]  + x[2] * ti[6]  + x[6] * ti[7];
    ar[2] = x[0] * ti[8]  + x[4] * ti[9]  + x[2] * ti[12] + x[6] * ti[13];
    ar[3] = x[0] * ti[10] + x[4] * ti[11] + x[2] * ti[14] + x[6] * ti[15];

    br[0] = x[1] * ti[16] + x[5] * ti[17] + x[3] * ti[20] + x[7] * ti[21];
    br[1] = x[1] * ti[18] + x[5] * ti[19] + x[3] * ti[22] + x[7] * ti[23];
    br[2] = x[1] * ti[24] + x[5] * ti[25] + x[3] * ti[28] + x[7] * ti[29];
    br[3] = x[1] * ti[26] + x[5] * ti[27] + x[3] * ti[30] + x[7] * ti[31];    


    // 2) Apply A_T operator, store outputs
    //
    //    1  0  0  0   1  0  0  0
    //    0  1  0  0   0  1  0  0
    //    0  0  1  0   0  0  1  0
    //    0  0  0  1   0  0  0  1
    //    0  0  0  1   0  0  0 -1
    //    0  0  1  0   0  0 -1  0
    //    0  1  0  0   0 -1  0  0
    //    1  0  0  0  -1  0  0  0

    //    The output is 32-bit integer.  The 32-bit int is converted to
    //    float and stored in a temporary array, 

    yr[0] = ( ar[0] + br[0] ); // store as 32-bit int
    yr[1] = ( ar[1] + br[1] ); // store as 32-bit int
    yr[2] = ( ar[2] + br[2] ); // store as 32-bit int
    yr[3] = ( ar[3] + br[3] ); // store as 32-bit int

    yr[4] = ( ar[3] - br[3] ); // store as 32-bit int
    yr[5] = ( ar[2] - br[2] ); // store as 32-bit int
    yr[6] = ( ar[1] - br[1] ); // store as 32-bit int
    yr[7] = ( ar[0] - br[0] ); // store as 32-bit int
  } // end for( i = 0; i < 8; ++i ) // end of row iDCT

  //     AP922hybr iDCT row transform done 
  //
  ////////////////////////////////////////////////////////////////////////


  //////////////////////////////////////////////////////////////////////
  //
  // Column IDCT operator :A_T*(F_T*E_T*B_T*D_T)*P_T
  //  AP922hybrid and AP922float share *identical* column_idct operators
  //
  //  (minor difference in the final descale_shift : 
  //   17bits instead of 16bits.)
  //
  // Let Y=[output column data, 8 elements], 16-bit short integer
  //     X=[input column data, 8 elements], 32-bit IEEE-754 float
  //
  //     Y= [ A_T*(F_T*E_T*B_T*D_T)*P_T ] * X
  //
  //   (Y and X are both column vectors)

  for ( i = 0; i < 8; ++i ) // column iDCT
  {
    xc = ((tdnfloat *) fTempArray) + i; // select column #i
    y = &data[ i ];

    // 1) Apply (D_T * P_T) - the cos() coefficients of D_T are implicit
    //    in the idct_row operation.  But we still need to apply the
    //    shuffling operation of D_T.
    //
    //    1  0  0  0   0  0  0  0
    //    0  0  0  0   1  0  0  0
    //    0  0  1  0   0  0  0  0
    //    0  0  0  0   0  0  1  0
    //    0  1  0  0   0  0  0  0
    //    0  0  0  0   0  0  0  1
    //    0  0  0  1   0  0  0  0
    //    0  0  0  0   0  1  0  0

    dp[0] = xc[ 0 *8];
    dp[1] = xc[ 4 *8];
    dp[2] = xc[ 2 *8];
    dp[3] = xc[ 6 *8];

    dp[4] = xc[ 1 *8];
    dp[5] = xc[ 7 *8];
    dp[6] = xc[ 3 *8];
    dp[7] = xc[ 5 *8];
 
    // 2) Apply B_T
    //
    //    1  1  0  0
    //    1 -1  0  0
    //    0  0  1 t2
    //    0  0 t2 -1
    //                1 t1  0  0
    //               t1 -1  0  0
    //                0  0  1 t3
    //                0  0 t3 -1
 
 
    b[0] =   dp[1]            + dp[0];
    b[1] =   dp[0]            - dp[1];

    b[2] = ( dp[3]*tg_2_16f ) + dp[2];
    b[3] = ( dp[2]*tg_2_16f ) - dp[3];

    b[4] = ( dp[5]*tg_1_16f ) + dp[4];
    b[5] = ( dp[4]*tg_1_16f ) - dp[5];

    b[6] = ( dp[7]*tg_3_16f ) + dp[6];
    b[7] = ( dp[6]*tg_3_16f ) - dp[7];
  
    // 3) Apply E_T
    //
    //    1  0  1  0
    //    0  1  0  1
    //    0  1  0 -1
    //    1  0 -1  0
    //                1  0  1  0
    //                1  0 -1  0
    //                0  1  0  1
    //                0  1  0 -1
 
    e[0] = b[0] + b[2];
    e[1] = b[1] + b[3];
    e[2] = b[1] - b[3];
    e[3] = b[0] - b[2];
    e[4] = b[4] + b[6];
    e[5] = b[4] - b[6];
    e[6] = b[5] + b[7];
    e[7] = b[5] - b[7];

    // 4) Apply F_T
    //
    //    1  0  0  0
    //    0  1  0  0
    //    0  0  1  0
    //    0  0  0  1
    //                1  0  0  0
    //                0  1  0  0
    //                0  0  1  0
    //                0  0  0  1

#define _F0 e[0] 
#define _F1 e[1] 
#define _F2 e[2] 
#define _F3 e[3] 
#define _F4 e[4] 
#define _F5 e[5] 
#define _F6 e[6] 
#define _F7 e[7] 

    tmp[0] = (e[5] + e[6]) * cos_4_16f;
    tmp[1] = (e[5] - e[6]) * cos_4_16f;
    _F5 = tmp[0];
    _F6 = tmp[1];

    // 5) Apply A_T
    //
    //    1  0  0  0   1  0  0  0
    //    0  1  0  0   0  1  0  0
    //    0  0  1  0   0  0  1  0
    //    0  0  0  1   0  0  0  1
    //    0  0  0  1   0  0  0 -1
    //    0  0  1  0   0  0 -1  0
    //    0  1  0  0   0 -1  0  0
    //    1  0  0  0  -1  0  0  0
    //
    //    yfloat[0]= F0 + F4
    //    yfloat[1]= F1 + F5
    //    yfloat[2]= F2 + F6
    //    yfloat[3]= F3 + F7
    //           
    //    yfloat[4]= F3 - F7
    //    yfloat[5]= F2 - F6
    //    yfloat[6]= F1 - F5
    //    yfloat[7]= F0 - F4
    //
    //
    // 6) float -> int, shift/round to final output y[]
    //    The final shift&round operation reverses the row-input prescaling.
    //    It also applies the chosen rounding-mode (accurate or fast.)
    //
    //    Note, the C-code below differs *substantially* from the AMD_3DNOW
    //    implementation.  The 3D_Now code applies this basic sequence:
    //
    //       ;// mm0 = [y1 y0] <float32>
    //       ;// mm1 = [y3 y2] <float32>
    //
    //      PF2ID mm0, mm0;  // mm0 <= [y1 y0] <int32>
    //      PF2ID mm1, mm1;  // mm1 <= [y3 y2] <int32>
    //
    //       ;// "0.5" is a 32-bit integer constant scaled up by some bitshift
    //      paddd mm0, [rnd_compensation]; // mm0 <=[y1+"0.5" y0+"0.5"]
    //      paddd mm1, [rnd_compensation]; // mm0 <=[y3+"0.5" y2+"0.5"]
    //
    //      psrad mm0, DESCALE_SHIFT1;     // stage1 shift
    //      psrad mm1, DESCALE_SHIFT1;     // stage1 shift
    //
    //      packssdw mm0, mm1;  // mm0 <= [y3 y2 y1 y0] <int16>
    //
    //      psraw mm0, DESCALE_SHIFT2;  // clip y[] to the range {-256,+255}
    //
    //       ;// DESCALE_SHIFT1 + DESCALE_SHIFT2 = PRESCALE_SHIFT 
    //       ;//      10         +        7       =      16+1
    //
    //      movq [OUTC+...*8], mm0;


    y[0*8] = SHIFT_ROUND_COLF( _F0 + _F4 );
    y[1*8] = SHIFT_ROUND_COLF( _F1 + _F5 );
    y[2*8] = SHIFT_ROUND_COLF( _F2 + _F6 );
    y[3*8] = SHIFT_ROUND_COLF( _F3 + _F7 );

    y[4*8] = SHIFT_ROUND_COLF( _F3 - _F7 );
    y[5*8] = SHIFT_ROUND_COLF( _F2 - _F6 );
    y[6*8] = SHIFT_ROUND_COLF( _F1 - _F5 );
    y[7*8] = SHIFT_ROUND_COLF( _F0 - _F4 );

  } // end for ( i = 0; i < 8; ++i ) // end of SSEfloat column iDCT

  //    AP922hybr iDCT column transform done 
  //
  ////////////////////////////////////////////////////////////////////////

  ////////////////////////////////////////////////////////////////////////
  //    Post transform clipping
  //
  //    In standard-C, the output clip code adds significantly to 
  //    execution time.  

#define IDCT_CLIP_LOW -256  // IDCT output range is 9-bits
#define IDCT_CLIP_HIGH 255  // IDCT output range is 9-bits

  for ( i = 0; i < 64; ++i )
  { // clip output to {-256,+255}
    if ( data[ i ] < IDCT_CLIP_LOW )
      data[ i ] = IDCT_CLIP_LOW;

    if ( data[ i ] > IDCT_CLIP_HIGH )
      data[ i ] = IDCT_CLIP_HIGH;
  }

}