/*
 ------------------------------------------------------------------------
 idct_ap922sse_rawcode.c - AP922float iDCT, Intel SSE implementation

 Requirements : 
  Visual C++ 6.0 Pro + Visual Studio Service Pack 4 + Processor Pack Beta
  SSE capable CPU (Pentium-III, Celeron-II)
 ------------------------------------------------------------------------

 AP-922 floating-point iDCT (SSE)
 --------------------------------
 
 Intel Application Note AP-922 discusses an MMX
 (scaled integer) implementation of the MPEG fDCT/iDCT operations.
 "AP922float" is a floating-point adaptation of the AP922 algorithm.
 
 The SSE implementation of AP922float follows the AP-922 document
 very closely.  However, since all math operations use 32-bit floats 
 instead of 16/32-bit integers, most of the actual assembly code 
 was rewritten from scratch.
 
 Several coding-issues arose during the development of this file.
 They are summarized here :

  #1) 16-bit integer -> 32-bit float conversion
  --------------------------------------------
     The usual MMX/SSE sequence for converting int->float is as follows:

     a) mov mm0, [int16_data];
        // extract element 0 only

     b) pshufw mm1, mm0, 00000000b; // mm1 = [d0 d0 d0 d0] <16-bit int>

     c) psrad mm1, 16;              // mm1 = [   d0    d0] <32-bit int>

     d) cvtpi2ps xmm0, mm1;         // xmm0= [   d0    d0] <32b float>

        //done

     This sequence has several problems.  pshufw, psrad, cvtpi2ps all issue
     from "port1", meaning the instructions will not pair on Pentium-III.
     Without further optimization, the int->float unpack step wastes many
     instruction slots.

     Although we cannot eliminate decoder contention, we can reduce it by
     using an instruction other than psrad.  For example, we substitute
     "pand" for psrad.

     a) mov mm0, [int16_data];
        // extract element 0 only

     b) pshufw mm1, mm0, 00000000b;// mm1 = [d0 d0 d0 d0] <16-bit int>

     c) pand mm1, mmAndmask;       // mm1 = [d0    d0   ] <32-bit int>*PSCALE
        // PSCALE = 65536, a row_prescale factor

     d) cvtpi2ps xmm0, mm1;        // xmm0= [   xd0   xd0] <32b float>
        // xd0 = d0 * PSCALE

     pand can issue from either port0 or port1.  As several int->float unpack
     operations are overlapped, pand can pair with any other MMX/SSE 
     instruction.  Some instruction slots are still empty, but overall
     less decoder bandwidth is wasted.

  #2) AP922 iDCT_row macro, PMADDWD/PADDD vs MULPS/ADDPS
  -----------------------------------------------------
     AP922 w[] coefficient table is ordered for PMADDWD+PADDD.

     <4 16bit integers>            <4 16bit integers>
     [X4  X0  X4  X0] mm0           [X6  X2  X6  X2] mm2
     [W6  W4  W2  W0] mm1           [W7  W5  W3  W1] mm3
            
            |                              |

       PMADDWD mm0,mm1                PMADDWD mm2,mm3

     [X0*W4 | X0*W0 ]               [X2*W5 | X2*W1 ]
     [  +   |   +   ]               [  +   |   +   ]
     [X4*W6 | X4*W2 ]               [X6*W7 | X6*W3 ]
      2 32-bit integers              2 32-bit integers

                     PADDD mm0, mm2

                      mm0 <= [A1 A0]

         ; a0 =x[0]*w[0]+x[2]*w[1]+x[4]*w[2]+x[6]*w[3];
         ; a1 =x[0]*w[4]+x[2]*w[5]+x[4]*w[6]+x[6]*w[7];

     The key observation is PMADDWD's *horizontal* add operation.

     Intel-SSE has no equivalent for MMX's PMADDWD, so the w[] table must
     be reordered for optimal execution with MULPS + ADDPS.

     // XMM0 = [x0 x0 x0 x0]     
     // XMM1 = [x2 x2 x2 x2]     
     // XMM2 = [x4 x4 x4 x4]     
     // XMM3 = [x6 x6 x6 x6]     

      MULPS XMM0, [TABLE+00];    // XMM0= [x0*w12 x0*w8  x0*w4 x0*w0]
      MULPS XMM1, [TABLE+16];    // XMM0= [x2*w13 x2*w9  x2*w5 x2*w1]
      MULPS XMM2, [TABLE+32];    // XMM0= [x4*w14 x4*w10 x4*w6 x4*w2]
      MULPS XMM3, [TABLE+48];    // XMM0= [x6*w15 x6*w11 x6*w7 x6*w3]
      ADDPS XMM1, XMM0;
      ADDPS XMM2, XMM1;
      ADDPS XMM3, XMM2;          // XMM3 =[  A3     A2     A1    A0 ]


  #3) AP922 iDCT_col macro, descale/shift/round (float->int16 converion)
  ----------------------------------------------------------------------
     The final operation in the AP922float iDCT requires float->int16
     conversion.  This potentially introduces numerical error which
     we've worked so hard to minimize.

     cvttps2pi mm0, xmm0;  // xmm0[y1 y0] -> truncate -> mm0[y1 y0]

     The cvttps2pi stores 32-bit integer from truncatation.  Truncation
     is equivalent to "negative_round_down / positive_round_up."  This
     means the 32-bit integer's magnitude is reduced by 0.5, on average.
     The usual method of adding a rounding_offset does not work, 
     because the offset's sign (+/-) depends on the input value's sign.

     Fortunately, Intel-SSE offers cvtps2pi.  This instruction uses the
     current rounding mode (stored in register mxcsr.)  MXCSR is usually set
     to "round-nearest."  Precise rounding is achieved by pre-adding a
     "correction_offset", then issuing cvtps2pi.

        ADDPS XMM0, [round_float_offset]; // [y3+. y2+. y1+. y0+.]
        CVTPS2PI MM0, XMM0;   // mm0 = [y1 y0]       <int32>
        MOVHLPS XMM0, XMM0;   // XMM0 =[y3 y2 y3 y2] <float>
        CVTPS2PI MM1, XMM0;   // mm1 = [y3 y2]       <int32>

        Execution wise, PADDD(MMX) would be faster than ADDPS, but as
        its name suggests, "round_float_offset" must be added as a 
        floating-point instruction, because numbers <1.0 cannot be 
        represented in integer!


     Recall from optimization (#1), the iDCT column output will produce
     outputs that are scaled up by 65536 (row_prescale.)
     The prescaling can be canceled by compensating the w[] table, but
     we defer discussion until item #4.

     **The above code assumes the rounding-mode bits ("mxcsr register") 
       are set for "round-nearest" mode!  If a different round-mode is
       in use (always round down, or always round up), the constant
       round_float_offset must be changed accordingly.

  #4) iDCT Output range clipping {-256,+255}
  ----------------------------------------------------------------------
     The iDCT's output range must be clipped to {-256,+255}  This range
     equals exactly a 9-bit signed integer, allowing us to leverage the
     packssdw instruction.

     cvtps2pi mm0, xmm0;  // mm0 = [y1 y0]*PRESCALE <int32>
     movhlps xmm0, xmm0;
       cvtps2pi mm1 xmm0; // mm1 = [y3 y2]*PRESCALE <int32>

     #define DESCALE_SHIFT0  9     // 16-9 = 7 bit shift
     #define DESCALE_SHIFT   7     // 16-7 = output y[] range is 9-bit int
     // (DESCALE+SHIFT0 + DESCALE_SHIFT) must equal the PRESCALE_SHIFT, 16.
     

     //psrad  mm0, DESCALE_SHIFT0; // don't need, see below

     //psrad  mm1, DESCALE_SHIFT0; // don't need, see below


     packssdw mm0, mm1;   // mm0 = [y3 y2 y1 y0]*PRESCALE <int16>

     psraw mm0, DESCALE_SHIFT; // mm0 = [y3 y2 y1 y0]
     
     movq [data], mm0;    // data is clipped to {-XXX,+YYY}

     "packssdw" applies floor/ceiling (arithmetic saturation) to -32768,+32767
     This range equals a 16-bit signed integer.  Our goal is to clip the
     final output y[] to a 9-bit signed integer.  Therefore, we choose
     DESCALE_SHIFT to be 16-9=7 bits.

     If we choose DESCALE_SHIFT = 7, the descale-factor is 2^7=128.  
     But from optimization (#1), the PRESCALE factor = 65536 (16-bits.)
     Somewhere between the iDCT-input and iDCT-output, the difference of
     these two (16-9=7bits) must be descaled. 
     
     One possible solution is to insert psrad instructions between the
     packssdw and cvtps2pi.  The psrad descales by "DESCALE_SHIFT0."
     But this solution is wasteful, because the descale_shift0 can be
     implicitly performed by adjusting the w[] coefficient table.

     Therefore, we choose to scae the coefficient table.  Since the table
     is stored in floating-point format, the stored constants do not lose
     any precision.  (The scale/descale factors are integer powers of 2.)
     
  #5) Compiler issues, data alignment to 16-byte address offset
  -------------------------------------------------------------

     I haven't figured out how to force alignment to 16-bytes, using 
     Visual C++.

 Numerical precision
 -------------------
  On Intel x86 CPUs, X87-FPU add/multiply operations are always calculated 
  with 80-bit internal precision, regardless of the input type.  In contrast,
  Pentium-III SSE calculations are performed with lower precision.  Therefore,
  all other things being equal, the X87 DCT will be more accurate than 
  Intel-SSE DCT.


 ------------------------------------------
 LEGAL MUMBO-JUMBO : Disclaimer of Warranty
 ------------------------------------------
  This software is available to the user without any license fee or
  royalty on an "as is" basis.  The author disclaims any and all warranties,
  whether express, implied, or statuary, including any implied warranties or 
  merchantability or of fitness for a particular purpose.  In no event shall 
  the copyright-holder be liable for any incidental, punitive, or 
  consequential damages of any kind whatsoever arising from the use of these
  programs.

  This disclaimer of warranty extends to the user of these programs and user's
  customers, employees, agents, transferees, successors, and assigns.

 Revision history
 ----------------
 09/03/2000 initial release of 'AP922float_SSE' iDCT
            Based on AP922, this iDCT uses SSE-float instructions for enhanced
            accuracy.  To the best of my knowledge, AP922float is fully
            IEEE 1180/1990 compliant iDCT, including output-range clipping
            to {-256,+255}

            Requires Visual C++ 6.0 Pro, Visual Studio Service Pack 4,
             and Visual C++ 6 Processor Pack Beta.  SP4 and PPBeta are both
             available at http://www.microsoft.com

            IMPORTANT, Visual C++ does not seem to align data to 16-byte
            address offset.  You will need to manually 
            
 09/03/2000 liaor@iname.com   http://members.tripod.com/~liaor
*/

#define NCOS1_16 (0.980785280403230449126182236134239) // cosine(  Pi/16 )
#define NCOS2_16 (0.923879532511286756128183189396788) // cosine( 2Pi/16 )
#define NCOS3_16 (0.831469612302545237078788377617906) // cosine( 3Pi/16 )
#define NCOS4_16 (0.707106781186547524400844362104849) // cosine( 4Pi/16 )
#define NCOS5_16 (0.555570233019602224742830813948533) // cosine( 5Pi/16 )
#define NCOS6_16 (0.382683432365089771728459984030399) // cosine( 6Pi/16 )
#define NCOS7_16 (0.195090322016128267848284868477022) // cosine( 7Pi/16 )

#define TANG1_16 ( NCOS7_16 / NCOS1_16) // tangent( Pi/16)
#define TANG2_16 ( NCOS6_16 / NCOS2_16) // tangent(2Pi/16)
#define TANG3_16 ( NCOS5_16 / NCOS3_16) // tangent(3Pi/16)

// Scaled floating-point coefficient table
//    Original (MMX) iDCT         SSE-iDCT 
//     table-ordering      ->    table-ordering
//    ------------------       ------------------
//    w00, w02, w04, w06,  ->  w00, w04, w08, w12,
//    w01, w03, w05, w07,  ->  w01, w05, w09, w13,
//    w08, w10, w12, w14,  ->  w02, w06, w10, w14,
//    w09, w11, w13, w15,  ->  w03, w07, w11, w15,
//    w16, w18, w20, w22,  ->  w16, w20, w24, w28,
//    w17, w19, w21, w23,  ->  w17, w21, w25, w29,
//    w24, w26, w28, w30,  ->  w18, w22, w26, w30,
//    w25, w27, w29, w31   ->  w19, w23, w27, w31
//
// Compared to the integer-based table, the float table is downscaled by a
// multiplication factor of 0.5

#define FIXED_DOWNSCALE (0.25/(1<<9))
 // 'fixed_downscale' includes a descale shift factor, discussed in the
 // text at the beginning of this file.

#define RS0 (NCOS4_16*FIXED_DOWNSCALE)  // iDCT row#0 scalar
#define RS1 (NCOS1_16*FIXED_DOWNSCALE)  // iDCT row#1 scalar
#define RS2 (NCOS2_16*FIXED_DOWNSCALE)  // iDCT row#2 scalar
#define RS3 (NCOS3_16*FIXED_DOWNSCALE)  // iDCT row#3 scalar
#define RS4 (NCOS4_16*FIXED_DOWNSCALE)  // iDCT row#4 scalar
#define RS5 (NCOS3_16*FIXED_DOWNSCALE)  // iDCT row#5 scalar
#define RS6 (NCOS2_16*FIXED_DOWNSCALE)  // iDCT row#6 scalar
#define RS7 (NCOS1_16*FIXED_DOWNSCALE)  // iDCT row#7 scalar

#define ROW_PRESCALE_SHIFT (16) // do not change!
#define DESCALE_SHIFT2 (7)

// compute the round_float_offset, for numerically accurate rounding
#define RND_INV_ROWF (1 << (DESCALE_SHIFT2-1) )
#define RND_FLOAT_OFFSET ((RND_INV_ROWF) - 0.5)

//////////////////////////////////////////////////////////////////////
//
//  If the function crashes after compiling, try inserting dummy
//  variables to shift around the sse-table's address.
//  Otherwise, you should recode the function to dynamically allocate
//  RAM for the table.  Then generate the table on the fly.
//

 const static __int64 useless_padding[1];

const float tab_i_01234567ssef[] =
{ 
 // Row #0
  NCOS4_16*RS0,  NCOS4_16*RS0,  NCOS4_16*RS0,  NCOS4_16*RS0, 
  NCOS2_16*RS0,  NCOS6_16*RS0, -NCOS6_16*RS0, -NCOS2_16*RS0, 
  NCOS4_16*RS0, -NCOS4_16*RS0, -NCOS4_16*RS0,  NCOS4_16*RS0, 
  NCOS6_16*RS0, -NCOS2_16*RS0,  NCOS2_16*RS0, -NCOS6_16*RS0, 
  NCOS1_16*RS0,  NCOS3_16*RS0,  NCOS5_16*RS0,  NCOS7_16*RS0, 
  NCOS3_16*RS0, -NCOS7_16*RS0, -NCOS1_16*RS0, -NCOS5_16*RS0, 
  NCOS5_16*RS0, -NCOS1_16*RS0,  NCOS7_16*RS0,  NCOS3_16*RS0, 
  NCOS7_16*RS0, -NCOS5_16*RS0,  NCOS3_16*RS0, -NCOS1_16*RS0, 

 // Row #1
  NCOS4_16*RS1,  NCOS4_16*RS1,  NCOS4_16*RS1,  NCOS4_16*RS1, 
  NCOS2_16*RS1,  NCOS6_16*RS1, -NCOS6_16*RS1, -NCOS2_16*RS1, 
  NCOS4_16*RS1, -NCOS4_16*RS1, -NCOS4_16*RS1,  NCOS4_16*RS1, 
  NCOS6_16*RS1, -NCOS2_16*RS1,  NCOS2_16*RS1, -NCOS6_16*RS1, 
  NCOS1_16*RS1,  NCOS3_16*RS1,  NCOS5_16*RS1,  NCOS7_16*RS1, 
  NCOS3_16*RS1, -NCOS7_16*RS1, -NCOS1_16*RS1, -NCOS5_16*RS1, 
  NCOS5_16*RS1, -NCOS1_16*RS1,  NCOS7_16*RS1,  NCOS3_16*RS1, 
  NCOS7_16*RS1, -NCOS5_16*RS1,  NCOS3_16*RS1, -NCOS1_16*RS1, 

 // Row #2
  NCOS4_16*RS2,  NCOS4_16*RS2,  NCOS4_16*RS2,  NCOS4_16*RS2, 
  NCOS2_16*RS2,  NCOS6_16*RS2, -NCOS6_16*RS2, -NCOS2_16*RS2, 
  NCOS4_16*RS2, -NCOS4_16*RS2, -NCOS4_16*RS2,  NCOS4_16*RS2, 
  NCOS6_16*RS2, -NCOS2_16*RS2,  NCOS2_16*RS2, -NCOS6_16*RS2, 
  NCOS1_16*RS2,  NCOS3_16*RS2,  NCOS5_16*RS2,  NCOS7_16*RS2, 
  NCOS3_16*RS2, -NCOS7_16*RS2, -NCOS1_16*RS2, -NCOS5_16*RS2, 
  NCOS5_16*RS2, -NCOS1_16*RS2,  NCOS7_16*RS2,  NCOS3_16*RS2, 
  NCOS7_16*RS2, -NCOS5_16*RS2,  NCOS3_16*RS2, -NCOS1_16*RS2, 

 // Row #3
  NCOS4_16*RS3,  NCOS4_16*RS3,  NCOS4_16*RS3,  NCOS4_16*RS3, 
  NCOS2_16*RS3,  NCOS6_16*RS3, -NCOS6_16*RS3, -NCOS2_16*RS3, 
  NCOS4_16*RS3, -NCOS4_16*RS3, -NCOS4_16*RS3,  NCOS4_16*RS3, 
  NCOS6_16*RS3, -NCOS2_16*RS3,  NCOS2_16*RS3, -NCOS6_16*RS3, 
  NCOS1_16*RS3,  NCOS3_16*RS3,  NCOS5_16*RS3,  NCOS7_16*RS3, 
  NCOS3_16*RS3, -NCOS7_16*RS3, -NCOS1_16*RS3, -NCOS5_16*RS3, 
  NCOS5_16*RS3, -NCOS1_16*RS3,  NCOS7_16*RS3,  NCOS3_16*RS3, 
  NCOS7_16*RS3, -NCOS5_16*RS3,  NCOS3_16*RS3, -NCOS1_16*RS3, 

 // Row #4
  NCOS4_16*RS4,  NCOS4_16*RS4,  NCOS4_16*RS4,  NCOS4_16*RS4, 
  NCOS2_16*RS4,  NCOS6_16*RS4, -NCOS6_16*RS4, -NCOS2_16*RS4, 
  NCOS4_16*RS4, -NCOS4_16*RS4, -NCOS4_16*RS4,  NCOS4_16*RS4, 
  NCOS6_16*RS4, -NCOS2_16*RS4,  NCOS2_16*RS4, -NCOS6_16*RS4, 
  NCOS1_16*RS4,  NCOS3_16*RS4,  NCOS5_16*RS4,  NCOS7_16*RS4, 
  NCOS3_16*RS4, -NCOS7_16*RS4, -NCOS1_16*RS4, -NCOS5_16*RS4, 
  NCOS5_16*RS4, -NCOS1_16*RS4,  NCOS7_16*RS4,  NCOS3_16*RS4, 
  NCOS7_16*RS4, -NCOS5_16*RS4,  NCOS3_16*RS4, -NCOS1_16*RS4, 

 // Row #5
  NCOS4_16*RS5,  NCOS4_16*RS5,  NCOS4_16*RS5,  NCOS4_16*RS5, 
  NCOS2_16*RS5,  NCOS6_16*RS5, -NCOS6_16*RS5, -NCOS2_16*RS5, 
  NCOS4_16*RS5, -NCOS4_16*RS5, -NCOS4_16*RS5,  NCOS4_16*RS5, 
  NCOS6_16*RS5, -NCOS2_16*RS5,  NCOS2_16*RS5, -NCOS6_16*RS5, 
  NCOS1_16*RS5,  NCOS3_16*RS5,  NCOS5_16*RS5,  NCOS7_16*RS5, 
  NCOS3_16*RS5, -NCOS7_16*RS5, -NCOS1_16*RS5, -NCOS5_16*RS5, 
  NCOS5_16*RS5, -NCOS1_16*RS5,  NCOS7_16*RS5,  NCOS3_16*RS5, 
  NCOS7_16*RS5, -NCOS5_16*RS5,  NCOS3_16*RS5, -NCOS1_16*RS5, 

 // Row #6
  NCOS4_16*RS6,  NCOS4_16*RS6,  NCOS4_16*RS6,  NCOS4_16*RS6, 
  NCOS2_16*RS6,  NCOS6_16*RS6, -NCOS6_16*RS6, -NCOS2_16*RS6, 
  NCOS4_16*RS6, -NCOS4_16*RS6, -NCOS4_16*RS6,  NCOS4_16*RS6, 
  NCOS6_16*RS6, -NCOS2_16*RS6,  NCOS2_16*RS6, -NCOS6_16*RS6, 
  NCOS1_16*RS6,  NCOS3_16*RS6,  NCOS5_16*RS6,  NCOS7_16*RS6, 
  NCOS3_16*RS6, -NCOS7_16*RS6, -NCOS1_16*RS6, -NCOS5_16*RS6, 
  NCOS5_16*RS6, -NCOS1_16*RS6,  NCOS7_16*RS6,  NCOS3_16*RS6, 
  NCOS7_16*RS6, -NCOS5_16*RS6,  NCOS3_16*RS6, -NCOS1_16*RS6, 

 // Row #7
  NCOS4_16*RS7,  NCOS4_16*RS7,  NCOS4_16*RS7,  NCOS4_16*RS7, 
  NCOS2_16*RS7,  NCOS6_16*RS7, -NCOS6_16*RS7, -NCOS2_16*RS7, 
  NCOS4_16*RS7, -NCOS4_16*RS7, -NCOS4_16*RS7,  NCOS4_16*RS7, 
  NCOS6_16*RS7, -NCOS2_16*RS7,  NCOS2_16*RS7, -NCOS6_16*RS7, 
  NCOS1_16*RS7,  NCOS3_16*RS7,  NCOS5_16*RS7,  NCOS7_16*RS7, 
  NCOS3_16*RS7, -NCOS7_16*RS7, -NCOS1_16*RS7, -NCOS5_16*RS7, 
  NCOS5_16*RS7, -NCOS1_16*RS7,  NCOS7_16*RS7,  NCOS3_16*RS7, 
  NCOS7_16*RS7, -NCOS5_16*RS7,  NCOS3_16*RS7, -NCOS1_16*RS7, 

 // iDCT column coefficients
  NCOS4_16, NCOS4_16, NCOS4_16, NCOS4_16,
  TANG1_16, TANG1_16, TANG1_16, TANG1_16,
  TANG2_16, TANG2_16, TANG2_16, TANG2_16,
  TANG3_16, TANG3_16, TANG3_16, TANG3_16,
  RND_FLOAT_OFFSET, RND_FLOAT_OFFSET, RND_FLOAT_OFFSET, RND_FLOAT_OFFSET,
  0       , 0       , 0       , 0,        // not used, padding
  0       , 0       , 0       , 0,        // not used, padding
  0       , 0       , 0       , 0         // not used, padding
};

/*
const static float cos_4_16f= NCOS4_16;
const static float tg_1_16f = TANG1_16;
const static float tg_2_16f = TANG2_16;
const static float tg_3_16f = TANG3_16;

; The original, unscaled idct coefficient table
; static const short w[32] = {
; FIX(cos_4_16), FIX(cos_2_16), FIX(cos_4_16), FIX(cos_6_16),
; FIX(cos_4_16), FIX(cos_6_16), -FIX(cos_4_16), -FIX(cos_2_16),
; FIX(cos_4_16), -FIX(cos_6_16), -FIX(cos_4_16), FIX(cos_2_16),
; FIX(cos_4_16), -FIX(cos_2_16), FIX(cos_4_16), -FIX(cos_6_16),
; FIX(cos_1_16), FIX(cos_3_16), FIX(cos_5_16), FIX(cos_7_16),
; FIX(cos_3_16), -FIX(cos_7_16), -FIX(cos_1_16), -FIX(cos_5_16),
; FIX(cos_5_16), -FIX(cos_1_16), FIX(cos_7_16), FIX(cos_3_16),
; FIX(cos_7_16), -FIX(cos_5_16), FIX(cos_3_16), -FIX(cos_1_16) };
;
; #define DCT_8_INV_ROW(x, y)



#define BITS_INV_ACC	4	//; 4 or 5 for IEEE
	// 5 yields higher accuracy, but lessens dynamic range on the input matrix
#define SHIFT_INV_ROW	(16 - BITS_INV_ACC)
#define SHIFT_INV_COL	(1 + BITS_INV_ACC )

#define RND_INV_ROW		(1 << (SHIFT_INV_ROW-1))
#define RND_INV_COL		(1 << (SHIFT_INV_COL-1)) 
#define RND_INV_CORR	(RND_INV_COL - 1)		//; correction -1.0 and round
//#define RND_INV_ROW		(1024 * (6 - BITS_INV_ACC)) //; 1 << (SHIFT_INV_ROW-1)
//#define RND_INV_COL		(16 * (BITS_INV_ACC - 3)) //; 1 << (SHIFT_INV_COL-1)
#define SHIFT_ROUND_COL( x ) ( ( x + RND_INV_COL ) >> (SHIFT_INV_COL) )
#define SHIFT_ROUND_ROW( x ) ( ( x + RND_INV_ROW ) >> (SHIFT_INV_ROW) )

//.data
//Align 16
//const static long r_inv_row[2] = { RND_INV_ROW, RND_INV_ROW};

const static short tg_1_16[4] = {13036, 13036, 13036, 13036 }; //tg * (2<<16) + 0.5
const static short tg_2_16[4] = {27146, 27146, 27146, 27146 }; //tg * (2<<16) + 0.5
const static short tg_3_16[4] = {-21746, -21746, -21746, -21746 }; //tg * (2<<16) + 0.5
//const static short cos_4_16[4] = {-19195, -19195, -19195, -19195 }; //cos * (2<<16) + 0.5
const static short ocos_4_16[4] = {23170, 23170, 23170, 23170 }; //cos * (2<<15) + 0.5



//#define SHIFT_ROUND_COLF( x ) ( floor( (x)*0.125*0.5 + 0.5 ) )
#define SHIFT_ROUND_COLF( x ) ( floor( (x)*0.5 + 0.5 ) )

#define SHIFT_ROUND_ROWF( x ) ( x )
*/

// externally allocated array of 64 floats, for temporary storage
extern void *fTempArray;


void
idct_ap922float_sse( short *data )
{
  static const __int64 mmAndmask = 0xFFFF0000FFFF0000; // [ XX __ XX __ ]

  /*
  static float a[8], b[8], e[8];
  static float dp[8], tmp[2];

 
  short *x; // pointer to input
  short *y; // pointer to final output
  float *yr, *xc; // pointer to output row
  float *ti; // pointer to IDCT row coefficient table

  int i;
  */

  // AP-922 inverse_DCT row transform (floating point version)
  //
  // Row IDCT operator :A_T*M_T*P_T
  // Let Y=[output column data, 8 elements] 32-bit IEEE-754 float
  //     X=[input column data, 8 elements] 16-bit short integer
  //
  //     Y= [ A_T*M_T*P_T ] * X
  //
  //   (Y and X are both column vectors)

#define INPR eax
#define TABLE ecx
#define TABLEP1 esi  // starting address of (TABLE +32 elements)
#define OUTR edx

__asm {
    mov OUTR, dword ptr [fTempArray];
     mov edi, 0x00; // i = 0

    mov INPR, dword ptr [data];   ;// row 0
     sub OUTR, 32;    // precompensate OUT
 
    lea TABLEP1, dword ptr [tab_i_01234567ssef]; // row 0

    movq mm7, [mmAndmask];      // mm7 <= [ FFFF 0000 FFFF 0000]
     mov TABLE, TABLEP1;

    add TABLEP1, 128;   // initialize 
;//   mov OUT, INP;  // algorithm writes data in-place  -> row 0

  ;// for ( x = 0; x < 8; ++x )  // transform one row per iteration
ALIGN 16 ;// align jump address to 16 byte offset
  acc_idct_rowloop1:

;//
;// begin (0)- Convert short -> float data
;// --------------------------------------
;//  Step 0a) : pshufw, generate 64-bit vectors [x0 x0 x0 x0],[x1 x1 x1 x1], ...
;//  Step 0b) : pand, mask out lower-word of each DWORD,
;//             example [x0 x0 x0 x0] -> [x0 __ x0 __]
;//             This effectively prescales all x[] inputs by 65536
;//               (denoted ROW_PRESCALE_SHIFT)
;//  Step 0c)   Convert DWORD vectors [X X] into ssefloat vectors [ X X X X]
;//             (use cvtpi2ps and movlhps)
;//
    movq mm2, [INPR] ;           // (0) mm2 <= x3 x2 x1 x0
     ;//

    movq mm4, [INPR+8] ;         // (0) mm4 <= x7 x6 x5 x4
     pshufw mm0, mm2, 00000000b ;// (0a) mm0 <= x0 x0 x0 x0

    pshufw mm1, mm2, 10101010b  ;// (0a) mm1 <= x2 x2 x2 x2
     pand mm0, mm7;             ;// (0b) mm0 = [ x0 x0 ]*SCALEUP (dword)

    cvtpi2ps xmm0, mm0;         ;// (0c) xmm0 <= [__ __ x0 x0] float
;// mm0 free
     pand mm1, mm7;             ;// (0b) mm1 = [ x2 x2 ]*SCALEUP (dword)

    prefetcht0 [TABLEP1 + 0*32] ;// prefetch coeff +1row, [w7..0]
     pshufw mm0, mm4, 00000000b ;// (0a)  mm0 <= x4 x4 x4 x4

    cvtpi2ps xmm1, mm1;         ;// (0c) xmm1 <= [__ __ x2 x2] float
;// mm1 free
     pand mm0, mm7;             ;// (0b) mm4 <= [x4 x4]*SCALEUP (dword)

    movlhps xmm0, xmm0;         ;// (0c) xmm0 <= [x0 x0 x0 x0] float
     cvtpi2ps xmm2, mm0;        ;// (0c) xmm2 <= [__ __ x4 x4] float
;// mm0 free

;// 1) Apply M_T*P_T operator (P_T is implicit with table/element order)
;//    (table order has been changed for optimum SSE execution)
;//       1a) mulps to generate a_v0, a_v1, a_v2, a_v3, b_v0, b_v1, b_v2, b_v3
;//       1b) addps to generate a_v[] = a_v0 + a_v1 + a_v2 + a_v3
;//                             b_v[] = b_v0 + b_v1 + b_v2 + b_v3

    mulps xmm0, [TABLE + 0*16]  ;// (1a) xmm0 <= a_v0 = [w3..w0] * x0
     pshufw mm1, mm4, 10101010b; // (0a) mm1 <= [x6 x6 x6 x6]

    pand mm1, mm7;              ;// (0b) mm1 <= [x6 x6]*SCALEUP (dword)
     movlhps xmm1, xmm1;        ;// (0c) xmm1 <=[ x2 x2 x2 x2] float

    mulps xmm1, [TABLE + 1*16]  ;// (1a) xmm1 <= a_v1 = [w7..w4] * x2
     cvtpi2ps xmm3, mm1;        ;// (0c) xmm3 <=[ __ __ x6 x6] float
;// mm1 free

    movlhps xmm2, xmm2;         ;// (0c) xmm2 <= [x4 x4 x4 x4] float
     pshufw mm0, mm2, 01010101b; // (0a) mm1 <=[x1 x1 x1 x1]

    mulps xmm2, [TABLE + 2*16];  // (1a) xmm2 <= a_v2 = [w11..w8] * x4
     pshufw mm1, mm2, 11111111b; // (0a) mm1 <=[x3 x3 x3 x3]
;// mm2 free

    pand mm0, mm7;               // (0b) mm0 <=[x1 x1]*SCALEUP (dword)
     movlhps xmm3, xmm3;        ;// (0) xmm3 <= [x6 x6 x6 x6]

    mulps xmm3, [TABLE + 3*16] ; // (1a) xmm3 <= a_v3 = [w15..w12] * x6
     cvtpi2ps xmm4, mm0;         // (0c) xmm4 <=[__ __ x1 x1] float
;// mm0 free

    addps xmm1, xmm0;           ;// (1b) xmm1 <= a_v0 + a_v1
;// xmm0 free
     pand mm1, mm7;              // (0b) mm1 <=[x3 x3]*SCALEUP (dword)

    prefetcht0 [TABLEP1 + 1*32] ;// prefetch coeff +1row [w15..8]
     pshufw mm0, mm4, 01010101b; // (0a) mm0 <= [x5 x5 x5 x5]

    cvtpi2ps xmm5, mm1;          // (0c) xmm5 <=[__ __ x3 x3] float
     pand mm0, mm7;              // (0b) mm0 <= [x5 x5]*SCALEUP (dword)
;// mm1 free

    cvtpi2ps xmm6, mm0;          // (0c) xmm6 <=[__ __ x5 x5] float
;// mm0 free
     movlhps xmm4, xmm4;         // (0c) xmm4 <= [x1 x1 x1 x1]

    mulps xmm4, [TABLE + 4*16]  ;// (1a) xmm4 <= b_v0
     pshufw mm1, mm4, 11111111b; // (0a) mm1 <= [x7 x7 x7 x7]
;// mm4 free

    addps xmm2, xmm1;           ;// (1b) xmm2 + a_v2 <= (a_v0+a_v1) + a_v2
;// xmm1 free
     movlhps xmm5, xmm5;        ;// (0c) xmm5 <=[x3 x3 x3 x3] float

    mulps xmm5, [TABLE + 5*16]  ;// (1a) xmm5 <= b_v1
     pand mm1, mm7;             ;// (0b) mm1 <= [x7 x7]*SCALEUP (dword)

    cvtpi2ps xmm7, mm1;         ;// (0c) xmm7 <= [__ __ x7 x7]
;// mm1 free
     movlhps xmm6, xmm6;         // (0c) xmm6 <= [x5 x5 x5 x5] float

    mulps xmm6, [TABLE + 6*16]  ;// (1a) xmm6 <= b_v2
     add OUTR, 32;              ;// OUT <= OUT + 32 (increment row) 

    addps xmm3, xmm2;           ;// (1b) xmm3 + a_v3 <=(a_v0+a_v1+a_v2) + a_v3
;// xmm3 = complete a-vector ( a_v[] )
;// xmm2 free
     add INPR, 16;               // increment INP(psrc) + 1row 

    prefetcht0 [TABLEP1 + 2*32] ;// prefetch coeff +1row [w23..15]
     movlhps xmm7, xmm7;

    mulps xmm7, [TABLE + 7*16]  ;// (1a) xmm7 <= b_v3
     addps xmm5, xmm4;          ;// (1b) xmm5 <= b_v0 + b_v1
;// xmm4 free

    add TABLE, 128;              // move TABLE to w[32]
     add edi, 1;                 // i = i + 1 (loop index)

    prefetcht0 [TABLEP1 + 3*32] ;// prefetch coeff +1row [w31..24]
     add TABLEP1, 128;           // move TABLEP1 to w[32]

    movaps xmm2, xmm3;           // (1b) xmm2 <= a_v[3..0] 



    addps xmm6, xmm5;            // (1b) xmm6 + b_v2 <=(b_v0+b_v1) + b_v2
;// xmm5 free

    addps xmm7, xmm6;            // (1b) xmm7 + b_v3 <=(b_v0+b_v1+b_v2) + b_v3
;// xmm6 free
;// xmm7 = complete b-vector ( b_v[] )

;// 2) Apply A_T operator to generate outputs y[7..0]
;//     2a) y[3..0] = a_v[] + b_v[]
;//     2b) *y[4..7] = a_v[] - b_v[]   (* requires shufps to invert order)
;//     2c) store y[] to temporary memory


;// output *y[4..7] = a[7..4] + b[7..4], output order must be reversed!
    subps xmm3, xmm7;      ;// 2b) xmm3 <= a_v[3..0]-b_v[3..0] (y4 y5 y6 y7)

;// output y[3..0] = a[3..0] + b[3..0]
    addps xmm2, xmm7;      ;// 2a) xmm2 <= a_v[3..0]+b_v[3..0] (y3 y2 y1 y0)
;// xmm7 free

;// fix output order, y[4..7] -> y[7..4]
    shufps xmm3, xmm3, 00011011b;// 2b) xmm3  [y7 y6 y5 y4] <- [y4 y5 y6 y7]

;// !*@&#@ Visual C++ refuses to align data to 16-byte offset, SIGH

;//    movups [OUT + 0*16], xmm2; ;// 2c) store [y3 y2 y1 y0]
  movaps [OUTR + 0*16], xmm2; ;// 2c) store [y3 y2 y1 y0]
;// xmm2 free

;//    movups [OUT + 1*16], xmm3;  ;// 2c) store [y7 y6 y5 y4]
  movaps [OUTR + 1*16], xmm3;  ;// 2c) store [y7 y6 y5 y4]
;// xmm3 free

    cmp edi, 0x08;

       ;// end for ( x = 0; x < 8; ++x )  // transform one row per iteration
    jl  acc_idct_rowloop1;

;/////////////////////////////////////////////////////////////////////
;//
;//
;// iDCT column transform
;//
;//
;/////////////////////////////////////////////////////////////////////

    ;// 1) Apply (D_T * P_T) - the cos() coefficients of D_T are implicit
    ;//    in the idct_row operation.  But we still need to apply the
    ;//    shuffling operation of D_T.
    ;//
    ;//    1  0  0  0   0  0  0  0
    ;//    0  0  0  0   1  0  0  0
    ;//    0  0  1  0   0  0  0  0
    ;//    0  0  0  0   0  0  1  0
    ;//    0  1  0  0   0  0  0  0
    ;//    0  0  0  0   0  0  0  1
    ;//    0  0  0  1   0  0  0  0
    ;//    0  0  0  0   0  1  0  0

    ;// dp[0] = xc[ 0 *8];
    ;// dp[1] = xc[ 4 *8];
    ;// dp[2] = xc[ 2 *8];
    ;// dp[3] = xc[ 6 *8];

    ;// dp[4] = xc[ 1 *8];
    ;// dp[5] = xc[ 7 *8];
    ;// dp[6] = xc[ 3 *8];
    ;// dp[7] = xc[ 5 *8];

#define SCRATCH esi
#define INPC edx
#define OUTC eax
    mov OUTC, dword ptr [data];   ;// row 0
     mov edi, 0x00; // i = 0


    mov INPC, dword ptr [fTempArray];
 //    sub OUT, 16;    // precompensate OUT


#define _PCOS4        TABLE   
#define _PTAN1        TABLE+16
#define _PTAN2        TABLE+32
#define _PTAN3        TABLE+48
#define _PRND_INV_ROW TABLE+64

    lea TABLE, [tab_i_01234567ssef];
    mov SCRATCH, INPC; // after we read the first input column, use as scratch

    add TABLE, 1024; // increment TABLE to the COS4 entry
//
ALIGN 16
acc_idct_colloop1:
    movaps xmm0, [INPC + 1*32] ; // (1) xmm0 <= dp[4]
     add edi, 1;

    ;

    movaps xmm2, [INPC + 3*32];  // (1) xmm2 <= dp[6]

    ;

    movaps xmm1, [INPC + 7*32];  // (1) xmm1 <= dp[5]
     movaps xmm4, xmm0;          // (1) xmm4 <= dp[4]

    ;

    movaps xmm3, [INPC + 5*32] ; // (1) xmm3 <= dp[7]
     movaps xmm6, xmm2;         ;// (2a) xmm6 <= dp[6]

    // 2) Apply B_T
    //
    //    b[4] =   dp[4] + ( dp[5]*tg_1_16f );
    //    b[5] = - dp[5] + ( dp[4]*tg_1_16f );
    //    b[6] =   dp[6] + ( dp[7]*tg_3_16f );
    //    b[7] = - dp[7] + ( dp[6]*tg_3_16f );

    ;

    mulps xmm4, [_PTAN1];    ;// (2a) xmm4 <= (dp[4]*tg_1_16f)
     movaps xmm5, xmm1       ;// (2a) xmm5 <= dp[5] 

    ;

    mulps xmm5, [_PTAN1];    ;// (2a) xmm5 <= (dp[5]*tg_1_16f)

    ;

    mulps xmm6, [_PTAN3];    ;// (2a) xmm6 <= (dp[6]*tg_3_16f)
     movaps xmm7, xmm3;      ;// (2a) xmm7 <= dp[7]

    subps xmm4, xmm1;        ;// (2b) xmm4 <= b[5] = (dp[4]*tg_1_16f) - dp[5]
// xmm1 free
     ;

     
    mulps xmm7, [_PTAN3];    ;// (2a) xmm7 <= (dp[7]*tg_3_16f)
     ;

    addps xmm5, xmm0;        ;// (2b) xmm5 <= b[4] = (dp[5]*tg_1_16f) + dp[4]
// xmm0 free

    ;


    subps xmm6, xmm3;        ;// (2b) xmm6 <= b[7] = (dp[6]*tg_1_16f) - dp[7]
// xmm3 free


// xmm1 used
     movaps xmm1, xmm4;      ;// (2b) xmm1 <= b[5]

// xmm3 used
    movaps xmm3, [INPC + 6*32]   ;// (2a) xmm3 <= dp[3]
     ;

    addps xmm7, xmm2;         // (2b) xmm7 <= b[6] = (dp[7]*tg_1_16f) + dp[6]
// xmm2 free
// xmm0 used
     movaps xmm0, xmm5;      ;// (2b) xmm0 <= b[4]

    ;

    // 3) Apply E_T
    //
    //     e[4] = b[4] + b[6];
    //     e[5] = b[4] - b[6];
    //     e[6] = b[5] + b[7];
    //     e[7] = b[5] - b[7];

// xmm2 used
    movaps xmm2, xmm3;          ;// (2a) xmm2 <= dp[3]
     addps xmm1, xmm6;          ;// (3) xmm1 <= e[6] = b[5] + b[7]

    mulps xmm3, [_PTAN2];       ;// (2a) xmm3 <= (dp[3]*tg_2_16f)
     ;

    subps xmm5, xmm7;           ;// (3) xmm5 <= e[5] = b[4] - b[6]

    ;

    subps xmm4, xmm6;           ;// (3) xmm4 <= e[7] = b[5] - b[7]
// xmm6 free
// xmm6 used
     movaps xmm6, [INPC + 2*32] ;// (2a) xmm6 <= dp[2]

    ;

    addps xmm0, xmm7;           ;// (3) xmm0 <= e[4] = b[4] + b[6]
// xmm7 free
// xmm7 used
     movaps xmm7, xmm5;         ;// (3) xmm7 <= e[5]

    // 4) Apply F_T
    //
    //    f[4] = e[4]
    //    f[5] = (e[5] + e[6]) * cos_4_16f;
    //    f[6] = (e[5] - e[6]) * cos_4_16f;
    //    f[7] = e[7]

// xmm0 <= f[4] = e[4]
// xmm4 <= f[7] = e[7]

    ;

    addps xmm5, xmm1;           ;// (4a) xmm5 <= (e[5]+e[6])

    ;

    subps xmm7, xmm1;           ;// (4a) xmm7 <= (e[5]-e[6])
// xmm1 free
// xmm1 used
     movaps xmm1, [INPC + 0*32] ;// (1) xmm1 <= dp[0]

movaps [SCRATCH], xmm0;         ;// store f[4] in scratch space
// xmm0 free (in scratch)

    mulps xmm5, [_PCOS4]     ;// (4b) xmm5 <= f[5] = (e[5] + e[6]) * cos_4_16f;
     addps xmm3, xmm6;       ;// (2b) xmm3 <= b[2] = (dp[3]*tg_2_16f) + dp[2]

// xmm4 f[7] free (in scratch)

    ;

    mulps xmm6, [_PTAN2];       ;// (2a) xmm6 <= (dp[2]*tg_2_16f)

    ;

    mulps xmm7, [_PCOS4]        ;// (4b) xmm7 <= f[6] = (e[5] - e[6]) * cos_4_16f;
    
    ;

    subps xmm6, xmm2;           ;// (2b) xmm6 <= b[3] = (dp[2]*tg_2_16f) - dp[3]
// xmm2 free
// xmm2 used
     movaps xmm2, [INPC+4*32];  ;// (1) xmm2 <= dp[1]

// xmm0 used
    movaps xmm0, xmm1;          ;// (1) xmm0 <= dp[0]
     ;

    addps xmm1, xmm2;           ;// (2b) xmm1 <= b[0] = dp[0] + dp[1]

    ;

    subps xmm0, xmm2;           ;// (2b) xmm0 <= b[1] = dp[0] - dp[1]
// xmm2 free

//    e[0] = b[0] + b[2];
//    e[1] = b[1] + b[3];
//    e[2] = b[1] - b[3];
//    e[3] = b[0] - b[2];

    ;

// xmm2 used 
    movaps xmm2, xmm1;          ;// (2b) xmm1 <= b[0]
     addps xmm1, xmm3;          ;// (3) xmm1 <= e[0] = b[0] + b[2];

    ;

    subps xmm2, xmm3;           ;// (3) xmm3 <= e[3] = b[0] - b[2];
// xmm3 free
// xmm3 used
     movaps xmm3, xmm0;         ;// (2b) xmm3 <= b[1]

    ;

    addps xmm0, xmm6;           ;// (3) xmm0 <= e[1] = b[1] + b[3];

    ;

    subps xmm3, xmm6;           ;// (3) xmm3 <= e[2] = b[1] - b[3];
// xmm6 free
// xmm6 used
     movaps xmm6, xmm2;         ;// (4) xmm6 <= f[3]

// xmm1 <= f[0]
// xmm0 <= f[1]
// xmm3 <= f[2]
// xmm2 <= f[3]
/*

    y[0*8] = SHIFT_ROUND_COLF( _F0 + _F4 );
    y[1*8] = SHIFT_ROUND_COLF( _F1 + _F5 );
    y[2*8] = SHIFT_ROUND_COLF( _F2 + _F6 );
    y[3*8] = SHIFT_ROUND_COLF( _F3 + _F7 );

    y[4*8] = SHIFT_ROUND_COLF( _F3 - _F7 );
    y[5*8] = SHIFT_ROUND_COLF( _F2 - _F6 );
    y[6*8] = SHIFT_ROUND_COLF( _F1 - _F5 );
    y[7*8] = SHIFT_ROUND_COLF( _F0 - _F4 );

    6a) round compensation, adjust float value to correct round alignment
        addps (ROUND_COMP - 0.5)

    6b) float -> int32 conversion (cvtps2pi)
       cvtps2pi mm0, xmm0;  // mm0 <= int32 versions of y[1 0]
        movhlps xmm0, xmm0; 
       cvtps2pi mm1, xmm0;  // mm1 <= int32 versions of y[3 2]

    6c) int32 -> int16 conversion (packssdw)
        packssdw mm0, mm1; // mm0 <= int16 versions of y[3 2 1 0]

    6d) int16 decsale (psraw)
       psraw mm0, DESCALE_SHIFT2; // mm0 <= final descaled output, {-256,+255}

    ** The descale shift is chosen to implement a {-256,+255} output range.

*/ 
    ;

    addps xmm2, xmm4;           ;// (5) xmm2 <= y[3] = f[3] + f[7]

    ;

    subps xmm6, xmm4;           ;// (5) xmm6 <= y[4] = f[3] - f[7]
// xmm4 free
// xmm4 used
     movaps xmm4, xmm0;         ;// xmm0 <= f[1]

    ;

    addps xmm2, [_PRND_INV_ROW]; // (6a) ROUND compensation

    ;

    addps xmm0, xmm5;           ;// (5) xmm0 <= y[1] = f[1] + f[5]

    ;

   cvtps2pi mm0, xmm2;          ;// (6b) mm0 <= y[3] low
    movhlps xmm2, xmm2;

   cvtps2pi mm1, xmm2;          ;// (6b) mm1 <= y[3] high
    ;

// xmm2 free
   addps xmm0, [_PRND_INV_ROW]; ;// (6a) ROUND compensation

   ;

   subps xmm4, xmm5;            ;// (5) xmm4 <= y[6] = f[1] - f[5]
// xmm5 free
// xmm2 used
    movaps xmm2, [SCRATCH];     ;// xmm2 <= f[4]

   packssdw mm0, mm1;           ;// (6c) mm0 <= y[3] 
    ;

   cvtps2pi mm2, xmm0;          ;// (6b) mm2 <= y[1] low
    movhlps xmm0, xmm0;

   cvtps2pi mm3, xmm0;          ;// (6b) mm3 <= y[1] high
// xmm0 free
// xmm5 used
    movaps xmm5, xmm3;          ;// xmm5 <= f[2]

   addps xmm3, xmm7;            ;// (5) xmm3 <= y[2] = f[2] + f[6]
    ;

   psraw mm0, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   addps xmm6, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    ;

   packssdw mm2, mm3;           ;// (6c) mm2 <= y[1]
    ;

   subps xmm5, xmm7;            ;// (5) xmm5 <= y[5] = f[2] - f[6]
// xmm7 free
    movq [OUTC + 3*16], mm0;    ;// mm0 <= y[3] stored

   psraw mm2, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm4, xmm6;          ;// (6b) mm4 <= y[4] low
    movhlps xmm6, xmm6;

   cvtps2pi mm5, xmm6;          ;// (6b) mm5 <= y[4] high
// xmm6 free
    ;
  
   addps xmm3, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    movq [OUTC + 1*16], mm2;    ;// y[1] stored

// xmm7 used
   movaps xmm7, xmm1;           ;// xmm7 <= f[0]
    ;

   addps xmm1, xmm2;            ;// (5) xmm2 <= y[0] = f[0] + f[4]
    ;

   packssdw mm4, mm5;           ;// (6c) mm4 <= y[4]
    ;

   cvtps2pi mm6, xmm3;          ;// (6b) mm6 <= y[2] low
    movhlps xmm3, xmm3;

   cvtps2pi mm7, xmm3;          ;// (6b) mm7 <= y[2] high
// xmm3 free
    ;

   subps xmm7, xmm2;            ;// (5) xmm7 <= y[7] = f[0] - f[4]
// xmm2 free

   psraw mm4, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   addps xmm5, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    movq [OUTC + 4*16], mm4;    ;// y[4] stored

   packssdw mm6, mm7;           ;// (6c) mm6 <= y[2]
    ;

   addps xmm4, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    add INPC, 16;

   psraw mm6, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm0, xmm5;          ;// (6b) mm0 <= y[5] low
    movhlps xmm5, xmm5;

   cvtps2pi mm1, xmm5;          ;// (6b) mm1 <= y[5] high
// xmm5 free
    movq [OUTC + 2*16], mm6;    ;// y[2] stored

   cvtps2pi mm2, xmm4;          ;// (6b) mm2 <= y[6] low
    movhlps xmm4, xmm4;

   cvtps2pi mm3, xmm4;          ;// (6b) mm3 <= y[6] high
// xmm4 free
    ;

   addps xmm1, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    ;

   packssdw mm0, mm1;           ;// (6c) mm0 <= y[5]
    ;

   addps xmm7, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    ;

   psraw mm0, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm4, xmm1;          ;// (6b) mm4 <= y[0] low
    movhlps xmm1, xmm1;

   packssdw mm2, mm3;           ;// (6c) mm2 <= y[6]
    ;

   cvtps2pi mm5, xmm1;          ;// (6b) mm5 <= y[0] high
// xmm1 free
    movq [OUTC + 5*16], mm0;

   psraw mm2, DESCALE_SHIFT2;   // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm6, xmm7;         ;// (6b) mm6 <= y[7] low
    movhlps xmm7, xmm7;

   cvtps2pi mm7, xmm7;         ;// (6b) mm7 <= y[7] high
// xmm7 free
    ;

   packssdw mm4, mm5;          ;// (6c) mm4 <= y[0]
    ;

   psraw mm4, DESCALE_SHIFT2;   // (6d) clip y[] to {-256,+255}
    ;

   packssdw mm6, mm7;          ;// (6c) mm6 <= y[7]
    movq [OUTC + 6*16], mm2;

   psraw mm6, DESCALE_SHIFT2;   // (6d) clip y[] to {-256,+255}
    movq [OUTC + 0*16], mm4;

   movq [OUTC + 7*16], mm6;

    add OUTC, 8;
 
   cmp edi, 2;
    jl acc_idct_colloop1;

//   ldmxcsr [_mxcsr_backup]; // restore state
   emms;
  }
    return;
}