/*
 ---------------------------------------------------------
 idct_ap922tdn.c - AP922float iDCT, 3DNow implementation
 ---------------------------------------------------------

 AP-922float iDCT (with AMD 3D-Now! acceleration)
 ------------------------------------------------
 
    Intel Application Note AP-922 discusses an MMX implementation of 
  the MPEG fDCT/iDCT operations.  "AP922float" is a floating-point
  adaptation of the AP922 algorithm.
 
    The 3DNow implementation requires a 3D-Now capable CPU, such as
  the AMD K6-2 or Athlon.  The provided assembly code works with Visual C++
  6.0 Pro's inline-assembler (with the help of 'amd3dx.h')  The
  Microsoft Visual C++ Processor Pack Beta isn't needed to compile this
  file.

    In addition to the 3D-Now assembly listing, this file includes an 
  "equivalent C-code listing."  The "equivalent C-Code" is functionally
  identical to ap922float_x87.  However the two versions differ in the
  ordering of iDCT coefficient table w[].  The 3D-Now ordering is more
  'natural' for standard x87 execution, and may run sligtly faster than
  the other X87 listing located in ap922x87.c

    For reasons documented below, the C-code and 3D-Now assembly code in 
  this file potentially generate different output:

 Numerical Issue #1 : float->int conversion
 ------------------------------------------

    Currently, the 3D-Now instruction set only supports truncation
  for float->int conversion.  Therefore, accurate_rounding is
  disabled for the X87 implementation.  

    To minimize the truncation error, this version of AP922 utilizes the
  same strategy discussed in idct_ap922x87.c  Please refer to that file.

    (For int->float conversion, AMD's microarchitecture does not have the 
  same instruction-pairing limitations as Intel's Pentium3.  The
  standard 3D-Now instruction set lacks "pshufw", so the coefficient
  table w[] has been reordered to minimize the amount of input x[]
  data-shuffling.)

 Numerical Issue #2 : 3D-Now precision vs X87 FPU
 ------------------------------------------------
    On 80x86 CPUs, FPU add/multiply operations are always  calculated with 
  80-bit internal precision.  Therefore, all other things being equal, the 
  X87 DCT will be more accurate than AMD 3D-Now, even when the X87 DCT 
  stores intermediate results in 32-bit float.

 
 ------------------------------------------
 LEGAL MUMBO-JUMBO : Disclaimer of Warranty
 ------------------------------------------

  This software is available to the user without any license fee or
  royalty on an "as is" basis.  The author disclaims any and all warranties,
  whether express, implied, or statuary, including any implied warranties or 
  merchantability or of fitness for a particular purpose.  In no event shall 
  the copyright-holder be liable for any incidental, punitive, or 
  consequential damages of any kind whatsoever arising from the use of these
  programs.

  This disclaimer of warranty extends to the user of these programs and user's
  customers, employees, agents, transferees, successors, and assigns.

 Revision history
 ----------------
 09/05/2000 initial release of 'AP922float' 3D-Now iDCT
            Based on AP922, this iDCT uses the original scaled-int (MMX) row
            iDCT operator, paired with an all new floating-point column iDCT.
            To the best of my knowledge, AP922hybr is fully IEEE 1180/1990 
            compliant iDCT, including output-range clipping to {-256,+255}

            This file contains inline-assembly code developed under 
            Visual C++ 6.0 Pro.
            
 09/03/2000 liaor@iname.com   http://members.tripod.com/~liaor
 liaor@iname.com   http://members.tripod.com/~liaor
*/


#include <math.h>   // standard C-library function "floor()"
#include "amd3dx.h" // inline-assembly macros for AMD 3D-Now

typedef float tdnfloat;  // 3D-Now float = IEEE-754 32-bit float
#define USE_FAST_FLOAT2INT 1

#define PSCALE_SHIFT 16    // iDCT row input "prescale", 16-bit left shift
#define DESCALE_SHIFT PSCALE_SHIFT // Total descale shift amount (#bits)
#define DESCALE_SHIFT2 7 // 2nd stage descale shift: '7' produces 9-bit output range
#define DESCALE_SHIFT1 (DESCALE_SHIFT-DESCALE_SHIFT2) // 1st stage descale
#define ROW_PRESCALE( x ) (((int)(x))<<PSCALE_SHIFT)
#define DESCALE_ROUND_COMPENSATION (1<<(DESCALE_SHIFT-1))

#ifdef USE_FAST_FLOAT2INT
  // approximate rounding

  #define SHIFT_ROUND_COLF( x )  ( ( ( (int)(x) ) + DESCALE_ROUND_COMPENSATION) >> DESCALE_SHIFT)

#else

  #define SHIFT_ROUND_COLF( x )  (  ( ((int)floor(x) ) + DESCALE_ROUND_COMPENSATION )>>DESCALE_SHIFT  ) 

#endif // 

#define NCOS1_16 (0.980785280403230449126182236134239) // cosine(  Pi/16 )
#define NCOS2_16 (0.923879532511286756128183189396788) // cosine( 2Pi/16 )
#define NCOS3_16 (0.831469612302545237078788377617906) // cosine( 3Pi/16 )
#define NCOS4_16 (0.707106781186547524400844362104849) // cosine( 4Pi/16 )
#define NCOS5_16 (0.555570233019602224742830813948533) // cosine( 5Pi/16 )
#define NCOS6_16 (0.382683432365089771728459984030399) // cosine( 6Pi/16 )
#define NCOS7_16 (0.195090322016128267848284868477022) // cosine( 7Pi/16 )

#define TANG1_16 ( NCOS7_16 / NCOS1_16) // tangent( Pi/16)
#define TANG2_16 ( NCOS6_16 / NCOS2_16) // tangent(2Pi/16)
#define TANG3_16 ( NCOS5_16 / NCOS3_16) // tangent(3Pi/16)

// Scaled floating-point coefficient table (ordering for AMD-3DNow ap922float)
// w00 w01 w02 w03
// w04 w05 w06 w07
// w08 w09 w10 w11
// w12 w13 w14 w15    ( all elements in float table are multiplied by 0.5)
// w16 w17 w18 w19
// w20 w21 w22 w23
// w24 w25 w26 w27
// w28 w29 w30 w31
//
// Compared to the integer-based table, the float table is downscaled by a
// multiplication factor of 0.5
// Scaled floating-point coefficient table
//      Original iDCT               3DNow-iDCT 
//     table-ordering      ->    table-ordering
//    ------------------       ------------------
//    w00, w02, w04, w06,  ->  w00, w01, w02, w03,
//    w01, w03, w05, w07,  ->  w04, w05, w06, w07,
//    w08, w10, w12, w14,  ->  w08, w09, w10, w11,
//    w09, w11, w13, w15,  ->  w12, w13, w14, w15,
//    w16, w18, w20, w22,  ->  w16, w17, w18, w19,
//    w17, w19, w21, w23,  ->  w20, w21, w22, w23,
//    w24, w26, w28, w30,  ->  w24, w25, w26, w27,
//    w25, w27, w29, w31   ->  w28, w29, w30, w31
//
// Compared to the integer-based table, the float table is downscaled by a
// multiplication factor of 0.25*"DESCALE_SHIFT0"

#define IDCT_CONSTANT (0.25)
 // all table entries multiplied by (0.5*0.5)=0.25


#define RS0 (NCOS4_16*IDCT_CONSTANT)  // iDCT row#0 scalar
#define RS1 (NCOS1_16*IDCT_CONSTANT)  // iDCT row#1 scalar
#define RS2 (NCOS2_16*IDCT_CONSTANT)  // iDCT row#2 scalar
#define RS3 (NCOS3_16*IDCT_CONSTANT)  // iDCT row#3 scalar
#define RS4 (NCOS4_16*IDCT_CONSTANT)  // iDCT row#4 scalar
#define RS5 (NCOS3_16*IDCT_CONSTANT)  // iDCT row#5 scalar
#define RS6 (NCOS2_16*IDCT_CONSTANT)  // iDCT row#6 scalar
#define RS7 (NCOS1_16*IDCT_CONSTANT)  // iDCT row#7 scalar

const static tdnfloat tab_i_01234567tdn[] =
{ 
 // Row #0
  NCOS4_16*RS0,  NCOS2_16*RS0,  NCOS4_16*RS0,  NCOS6_16*RS0, 
  NCOS4_16*RS0,  NCOS6_16*RS0, -NCOS4_16*RS0, -NCOS2_16*RS0, 
  NCOS4_16*RS0, -NCOS6_16*RS0, -NCOS4_16*RS0,  NCOS2_16*RS0, 
  NCOS4_16*RS0, -NCOS2_16*RS0,  NCOS4_16*RS0, -NCOS6_16*RS0, 
  NCOS1_16*RS0,  NCOS3_16*RS0,  NCOS5_16*RS0,  NCOS7_16*RS0, 
  NCOS3_16*RS0, -NCOS7_16*RS0, -NCOS1_16*RS0, -NCOS5_16*RS0, 
  NCOS5_16*RS0, -NCOS1_16*RS0,  NCOS7_16*RS0,  NCOS3_16*RS0, 
  NCOS7_16*RS0, -NCOS5_16*RS0,  NCOS3_16*RS0, -NCOS1_16*RS0, 

 // Row #1
  NCOS4_16*RS1,  NCOS2_16*RS1,  NCOS4_16*RS1,  NCOS6_16*RS1, 
  NCOS4_16*RS1,  NCOS6_16*RS1, -NCOS4_16*RS1, -NCOS2_16*RS1, 
  NCOS4_16*RS1, -NCOS6_16*RS1, -NCOS4_16*RS1,  NCOS2_16*RS1, 
  NCOS4_16*RS1, -NCOS2_16*RS1,  NCOS4_16*RS1, -NCOS6_16*RS1, 
  NCOS1_16*RS1,  NCOS3_16*RS1,  NCOS5_16*RS1,  NCOS7_16*RS1, 
  NCOS3_16*RS1, -NCOS7_16*RS1, -NCOS1_16*RS1, -NCOS5_16*RS1, 
  NCOS5_16*RS1, -NCOS1_16*RS1,  NCOS7_16*RS1,  NCOS3_16*RS1, 
  NCOS7_16*RS1, -NCOS5_16*RS1,  NCOS3_16*RS1, -NCOS1_16*RS1, 

 // Row #2
  NCOS4_16*RS2,  NCOS2_16*RS2,  NCOS4_16*RS2,  NCOS6_16*RS2, 
  NCOS4_16*RS2,  NCOS6_16*RS2, -NCOS4_16*RS2, -NCOS2_16*RS2, 
  NCOS4_16*RS2, -NCOS6_16*RS2, -NCOS4_16*RS2,  NCOS2_16*RS2, 
  NCOS4_16*RS2, -NCOS2_16*RS2,  NCOS4_16*RS2, -NCOS6_16*RS2, 
  NCOS1_16*RS2,  NCOS3_16*RS2,  NCOS5_16*RS2,  NCOS7_16*RS2, 
  NCOS3_16*RS2, -NCOS7_16*RS2, -NCOS1_16*RS2, -NCOS5_16*RS2, 
  NCOS5_16*RS2, -NCOS1_16*RS2,  NCOS7_16*RS2,  NCOS3_16*RS2, 
  NCOS7_16*RS2, -NCOS5_16*RS2,  NCOS3_16*RS2, -NCOS1_16*RS2, 

 // Row #3
  NCOS4_16*RS3,  NCOS2_16*RS3,  NCOS4_16*RS3,  NCOS6_16*RS3, 
  NCOS4_16*RS3,  NCOS6_16*RS3, -NCOS4_16*RS3, -NCOS2_16*RS3, 
  NCOS4_16*RS3, -NCOS6_16*RS3, -NCOS4_16*RS3,  NCOS2_16*RS3, 
  NCOS4_16*RS3, -NCOS2_16*RS3,  NCOS4_16*RS3, -NCOS6_16*RS3, 
  NCOS1_16*RS3,  NCOS3_16*RS3,  NCOS5_16*RS3,  NCOS7_16*RS3, 
  NCOS3_16*RS3, -NCOS7_16*RS3, -NCOS1_16*RS3, -NCOS5_16*RS3, 
  NCOS5_16*RS3, -NCOS1_16*RS3,  NCOS7_16*RS3,  NCOS3_16*RS3, 
  NCOS7_16*RS3, -NCOS5_16*RS3,  NCOS3_16*RS3, -NCOS1_16*RS3, 

 // Row #4
  NCOS4_16*RS4,  NCOS2_16*RS4,  NCOS4_16*RS4,  NCOS6_16*RS4, 
  NCOS4_16*RS4,  NCOS6_16*RS4, -NCOS4_16*RS4, -NCOS2_16*RS4, 
  NCOS4_16*RS4, -NCOS6_16*RS4, -NCOS4_16*RS4,  NCOS2_16*RS4, 
  NCOS4_16*RS4, -NCOS2_16*RS4,  NCOS4_16*RS4, -NCOS6_16*RS4, 
  NCOS1_16*RS4,  NCOS3_16*RS4,  NCOS5_16*RS4,  NCOS7_16*RS4, 
  NCOS3_16*RS4, -NCOS7_16*RS4, -NCOS1_16*RS4, -NCOS5_16*RS4, 
  NCOS5_16*RS4, -NCOS1_16*RS4,  NCOS7_16*RS4,  NCOS3_16*RS4, 
  NCOS7_16*RS4, -NCOS5_16*RS4,  NCOS3_16*RS4, -NCOS1_16*RS4, 

 // Row #5
  NCOS4_16*RS5,  NCOS2_16*RS5,  NCOS4_16*RS5,  NCOS6_16*RS5, 
  NCOS4_16*RS5,  NCOS6_16*RS5, -NCOS4_16*RS5, -NCOS2_16*RS5, 
  NCOS4_16*RS5, -NCOS6_16*RS5, -NCOS4_16*RS5,  NCOS2_16*RS5, 
  NCOS4_16*RS5, -NCOS2_16*RS5,  NCOS4_16*RS5, -NCOS6_16*RS5, 
  NCOS1_16*RS5,  NCOS3_16*RS5,  NCOS5_16*RS5,  NCOS7_16*RS5, 
  NCOS3_16*RS5, -NCOS7_16*RS5, -NCOS1_16*RS5, -NCOS5_16*RS5, 
  NCOS5_16*RS5, -NCOS1_16*RS5,  NCOS7_16*RS5,  NCOS3_16*RS5, 
  NCOS7_16*RS5, -NCOS5_16*RS5,  NCOS3_16*RS5, -NCOS1_16*RS5, 

 // Row #6
  NCOS4_16*RS6,  NCOS2_16*RS6,  NCOS4_16*RS6,  NCOS6_16*RS6, 
  NCOS4_16*RS6,  NCOS6_16*RS6, -NCOS4_16*RS6, -NCOS2_16*RS6, 
  NCOS4_16*RS6, -NCOS6_16*RS6, -NCOS4_16*RS6,  NCOS2_16*RS6, 
  NCOS4_16*RS6, -NCOS2_16*RS6,  NCOS4_16*RS6, -NCOS6_16*RS6, 
  NCOS1_16*RS6,  NCOS3_16*RS6,  NCOS5_16*RS6,  NCOS7_16*RS6, 
  NCOS3_16*RS6, -NCOS7_16*RS6, -NCOS1_16*RS6, -NCOS5_16*RS6, 
  NCOS5_16*RS6, -NCOS1_16*RS6,  NCOS7_16*RS6,  NCOS3_16*RS6, 
  NCOS7_16*RS6, -NCOS5_16*RS6,  NCOS3_16*RS6, -NCOS1_16*RS6, 

 // Row #7
  NCOS4_16*RS7,  NCOS2_16*RS7,  NCOS4_16*RS7,  NCOS6_16*RS7, 
  NCOS4_16*RS7,  NCOS6_16*RS7, -NCOS4_16*RS7, -NCOS2_16*RS7, 
  NCOS4_16*RS7, -NCOS6_16*RS7, -NCOS4_16*RS7,  NCOS2_16*RS7, 
  NCOS4_16*RS7, -NCOS2_16*RS7,  NCOS4_16*RS7, -NCOS6_16*RS7, 
  NCOS1_16*RS7,  NCOS3_16*RS7,  NCOS5_16*RS7,  NCOS7_16*RS7, 
  NCOS3_16*RS7, -NCOS7_16*RS7, -NCOS1_16*RS7, -NCOS5_16*RS7, 
  NCOS5_16*RS7, -NCOS1_16*RS7,  NCOS7_16*RS7,  NCOS3_16*RS7, 
  NCOS7_16*RS7, -NCOS5_16*RS7,  NCOS3_16*RS7, -NCOS1_16*RS7,

 // iDCT column coefficients
  NCOS4_16,  NCOS4_16,   // cosine  (4Pi/16)
  TANG1_16,  TANG1_16,   // tangent ( Pi/16)
  TANG2_16,  TANG2_16,   // tangent (2Pi/16)
  TANG3_16,  TANG3_16    // tangent (3Pi/16)
}; 

// externally allocated array of 64 floats, for temporary storage
extern void *fTempArray;

const static tdnfloat tg_1_16f = TANG1_16; // scalar versions of tab_i_col[]
const static tdnfloat tg_2_16f = TANG2_16;
const static tdnfloat tg_3_16f = TANG3_16;
const static tdnfloat cos_4_16f= NCOS4_16;

// Rounding compensation for final output (in 3DNow column_iDCT operator)
const static int rnd_compensation[2] = { DESCALE_ROUND_COMPENSATION,DESCALE_ROUND_COMPENSATION};



////////////////////////////////////////////////////////////////////////
//
//
// AP922float_3dn(), AMD 3D-Now implementation of AP922float
//
//
////////////////////////////////////////////////////////////////////////
void
idct_ap922float_3dn( short *data )
{
    const static __int64 mmAndMask = 0xFFFF0000FFFF0000;
  __asm {

#define INPR  eax
#define OUTR  edx
#define TABLE ecx
  mov OUTR, dword ptr [fTempArray];
   mov edi, 0x00; //x = 0

  mov INPR, dword ptr [data];   ;// row 0
//   mov OUT, INP;  // algorithm writes data in-place  -> row 0

  lea TABLE, dword ptr [tab_i_01234567tdn]; // row 0
   sub OUTR, 32;  // precompensate output address

  // for ( x = 0; x < 8; ++x )  // transform one row per iteration
ALIGN 16
acc_idct_rowloop1:
  movq mm0, qword ptr [INPR] ;   // 0 ; x3 x2 x1 x0
   add edi, 0x01;         // x=x+1, increment row counter

  movq mm3, qword ptr [INPR+8] ; // 1 ; x7 x6 x5 x4
   movq mm2, mm0 ;        // 2 ; x3 x2 x1 x0

// 1) convert mm0 into [x2 x0] <dword>
  pslld mm0, PSCALE_SHIFT         // 1a) 0 ; [x2 x0]*PSCALE <dword>
   movq mm1, mm3;                 // 1a) 3 ; x7 x6 x5 x4

  PI2FD (mm0, mm0);               // 1b) mm0 <=[x2 x0]*PSCALE <float>
   pslld mm1, PSCALE_SHIFT;       // 1a) 2 ; [x3 x1]*PSCALE <dword>

  pand mm2, [mmAndMask];         // 1a) mm2 <=[x3 x1]*PSCALE <float>
   PI2FD (mm1, mm1);               // 1b) mm1 <=[x6 x4]*PSCALE <float>
 
  pand mm3, [mmAndMask];         // 1a) mm2 <=[x7 x5]*PSCALE <float>
   movq mm4, mm0;                  // 2a) mm4 <=[x2 x0] <float>   (a0)

  PFMULM( mm4, TABLE, 0);         // 2a) mm4 <=[x2*w1 x0*w0]     (a0a)
   movq mm5, mm1;                  // 2a) mm5 <=[x6 x4] <float>

  PFMULM( mm5, TABLE, 8);         // 2a) mm5 <=[x6*w3 x4*w2]     (a0b)
   PI2FD (mm2, mm2);               // 1b) mm2 <=[x3 x1]*PSCALE <float>

  PI2FD (mm3, mm3);               // 1b) mm3 <=[x7 x5]*PSCALE <float>
   movq mm6, mm0;                 // 2a) mm6 <=[x2 x0] <float>   (a1a)

  PFMULM( mm6, TABLE, 16);         // 2a) mm6 <=[x2*w5 x0*w4]     (a1a)
   movq mm7, mm1;                 // 2a) mm7 <=[x6 x4] <float>   (a1b)

  PFMULM( mm7, TABLE, 24);         // 2a) mm7 <=[x6*w7 x4*w6]     (a1b)
   PFACC( mm4, mm5);                // 2b) mm4 <=[a0b a0a]       (a0)
// mm5 free

  add OUTR, 32;                    // increment output-pointer (+8 floats)
// mm5 used
   movq mm5, mm0;                  // 2a) mm5 <=[x2 x0] <float>   (a2a)

//mm7 free
  PFACC( mm6, mm7);                // 2b) mm6 <=[a1b a1a]       (a1)
   PFMULM( mm5, TABLE, 32);         // 2a) mm5 <=[x2*w9 x0*w8]     (a2a)
//mm7 used

  movq mm7, mm1;                  // 2a) mm7 <=[x6 x4] <float>   (a2b)
   ;//

  PFMULM( mm7, TABLE, 40);         // 2a) mm7 <=[x6*w11 x4*w10]     (a2b)
   PFACC( mm4, mm6 );              // 2c) mm4 <= [a1 a0] 
// mm6 free, mm4 contains [a1 a0]

  PFMULM( mm0, TABLE, 48);         // 2a) mm0 <=[x2*w13 x0*w12]     (a3a)
// mm6 used
   movq mm6, mm2;                  // 2d) mm6 <= [x3 x1]          (b0a)
  

  PFMULM( mm1, TABLE, 56);         // 2a) mm1 <=[x6*w15 x4*w14]     (a3b)
   PFACC( mm5, mm7);                // 2b) mm5 <=[a2b a2a]       (a2)
//mm7 free

  movq [ OUTR], mm4;           // store [a1 a0] to INPR-16
// mm4 free
//mm7 used
   movq mm7, mm3;                  // 2d) mm7 <= [x7 x5]          (b0b)


  PFMULM( mm6, TABLE, 64);         // 2d) mm6 <=[x3*w17 x1*w16]     (b0a)
   PFACC( mm0, mm1);                // 2b) mm5 <=[a3b a3a]       (a3)
//mm1 free

  PFMULM( mm7, TABLE, 72);         // 2d) mm7 <=[x7*w19 x5*w18]     (b0b)
// mm4 used
   movq mm4, mm2;                  // 2d) mm4 <= [x3 x1]          (b1a)
   
  PFMULM( mm4, TABLE, 80);         // 2d) mm4 <=[x3*w21 x1*w20]     (b1a)
   PFACC( mm5, mm0);                // 2c) mm5 <=[a3 a2]
//mm0 free, mm5 contains [a3 a2]

//mm1 used
  movq mm1, mm3;                  // 2d) mm1 <= [x7 x5]          (b1b)
   PFACC( mm6, mm7 );             // 2e) mm6 <=[b0b b0a]         (b0)
//mm7 free

  PFMULM( mm1, TABLE, 88);         // 2d) mm1 <=[x7*w23 x5*w22]     (b1b)
   movq mm0, mm2;                  // 2d) mm0 <= [x3 x1]          (b2a)
//mm0 used

//  movq [INPR - 8 ], mm5;            // store [a3 a2] to INPR-8
// mm5 free
  PFMULM( mm0, TABLE, 96);         // 2d) mm0 <=[x3*w25 x1*w24]     (b2a)
   ;

// mm7 used
  movq mm7, mm3;                  // 2d) mm7 <= [x7 x5]          (b3b)
   PFMULM( mm3, TABLE, 104);         // 2d) mm3 <=[x7*w27 x5*w26]     (b2b)

  PFACC( mm4, mm1);               // 2e) mm4 <= [b1b b1a]            (b1)
// mm1 free
   PFMULM( mm2, TABLE, 112);         // 2d) mm2 <=[x3*w29 x1*w28]     (b3a)

  PFMULM( mm7, TABLE, 120);         // 2d) mm7 <=[x7*w31 x5*w30]     (b3b)
   PFACC( mm0, mm3);               // 2e) mm0 <= [b2b b2a]            (b2)
// mm3 free

// mm3 used
  movq mm3, [ OUTR];           // 3a) mm3 <= [a1 a0] (fetch from memory)
   PFACC( mm6, mm4 );               // 2f) mm6 <=[b1 b0] done
// mm4 free, mm6 = [b1 b0]

  PFACC( mm2, mm7);                 // 2f) mm2 <=[b3b b3a]          (b3)
//mm7 free
//mm1 used
   movq mm1, mm5;                   // 3a) mm1 <= [a3 a2]

//mm4 used
  movq mm4, mm3;                    // 3a) mm4 <= [a1 a0]
   PFADD( mm3, mm6 );                  // 3a) mm3 <= [y1 y0] <= [a1+b1 a0+b0]

  add TABLE, 128;                  // increment table pointer (+32 floats)
   PFACC( mm0, mm2);                // 2f) mm0 <= [b3 b2]  done
// mm2 free, mm0 = [b3 b2]

  PFSUB( mm4, mm6 );                   // 3a) mm4 <= [y6 y7] <= [a1-b1 a0-b0]
// mm6 free
   add INPR, 16;                    // increment input pointer (+8 shorts)

  PFADD( mm5, mm0 );                  // 3a) mm5 <= [y3 y2] <= [a3+b3 a2+b2]
   movq [ OUTR + 0], mm3;            // 3b) store [y1 y0]
// mm3 free

  PFSUB( mm1, mm0 );                   // 3a) mm1 <= [y4 y5] <= [a3-b3 a2-b2]
// mm0 free  
// mm6 used
   punpckldq mm6, mm4;             // 3b) mm6 = [y7 XX]

  punpckhdq mm4, mm6;              // 3b) mm4 = [y7 y6] done
// mm6 free
   cmp edi, 0x08;          // evaluate { x <> 8 }

  movq [ OUTR + 8], mm5;            // 3b) store [y3 y2]
// mm5 free
   punpckldq mm6, mm1;             // 3b) mm6 = [y5 XX]

  movq [ OUTR + 24], mm4;           // 3b) store [y7 y6] done
// mm4 free
   punpckhdq mm1, mm6;              // 3b) mm1 = [y5 y4]
// mm6 free

  movq [ OUTR + 16], mm1;           // 3b) store [y5 y4] done
// mm1 free
   jl acc_idct_rowloop1;  // end for ( x=0; x < 8; x=x+1)

//  FEMMS;
//  }

  //
  // done with the iDCT row-transformation
  ////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////
;//
;//
;// AP922float-3DNow iDCT column transform
;//
;//
;/////////////////////////////////////////////////////////////////////


#define SCRATCH esi
#define INPC edx
#define OUTC eax
    mov OUTC, dword ptr [data];   ;// row 0
     mov edi, 0x00; // x = 0

    mov INPC, dword ptr [fTempArray];
     sub OUTC, 4;   // precompensate out

#define _PRND_INV_ROW TABLE

#define _OCOS4        0
#define _OTAN1        8
#define _OTAN2        16
#define _OTAN3        24
#define _ORND_INV_ROW 32

//  TABLE is already set to proper location by the previous idct_row operation
//    lea TABLE, [tab_i_col];
     mov SCRATCH, INPC; // after we read the first input column, use as scratch

     // for ( x = 0; x < 8; x=x+2 )   // process two columns per pass
ALIGN 16
acc_idct_colloop1:
    movq mm0, [INPC + 1*32] ; // (1) xmm0 <= dp[4]
     add edi, 2;              // x = x +2 (two columns per pass)

    movq mm2, [INPC + 3*32];  // (1) xmm2 <= dp[6]

    movq mm1, [INPC + 7*32];  // (1) xmm1 <= dp[5]
     movq mm4, mm0;          // (1) xmm4 <= dp[4]

    movq mm3, [INPC + 5*32] ; // (1) xmm3 <= dp[7]
     movq mm6, mm2;         ;// (2a) xmm6 <= dp[6]
    ;// 1) Apply (D_T * P_T) - the cos() coefficients of D_T are implicit
    ;//    in the idct_row operation.  But we still need to apply the
    ;//    shuffling operation of D_T.
    ;//
    ;// dp[0] = xc[ 0 *8];
    ;// dp[1] = xc[ 4 *8];
    ;// dp[2] = xc[ 2 *8];
    ;// dp[3] = xc[ 6 *8];

    ;// dp[4] = xc[ 1 *8];
    ;// dp[5] = xc[ 7 *8];
    ;// dp[6] = xc[ 3 *8];
    ;// dp[7] = xc[ 5 *8];

    // 2) Apply B_T
    //
    //    b[4] =   dp[4] + ( dp[5]*tg_1_16f );
    //    b[5] = - dp[5] + ( dp[4]*tg_1_16f );
    //    b[6] =   dp[6] + ( dp[7]*tg_3_16f );
    //    b[7] = - dp[7] + ( dp[6]*tg_3_16f );

    PFMULM( mm4, TABLE, _OTAN1);    ;// (2a) xmm4 <= (dp[4]*tg_1_16f)
     movq mm5, mm1                   ;// (2a) xmm5 <= dp[5] 

    PFMULM( mm5, TABLE, _OTAN1);    ;// (2a) xmm5 <= (dp[5]*tg_1_16f)
     movq mm7, mm3;                  ;// (2a) xmm7 <= dp[7]

    PFMULM( mm6, TABLE, _OTAN3);    ;// (2a) xmm6 <= (dp[6]*tg_3_16f)
     PFSUB( mm4, mm1);       ;// (2b) xmm4 <= b[5] = (dp[4]*tg_1_16f) - dp[5]

// xmm1 free

    PFMULM( mm7, TABLE, _OTAN3);    ;// (2a) xmm7 <= (dp[7]*tg_3_16f)
     PFADD( mm5, mm0);       ;// (2b) xmm5 <= b[4] = (dp[5]*tg_1_16f) + dp[4]
// xmm0 free

    PFSUB( mm6, mm3);        ;// (2b) xmm6 <= b[7] = (dp[6]*tg_1_16f) - dp[7]
// xmm3 free
// xmm1 used
     movq mm1, mm4;                  ;// (2b) xmm1 <= b[5]

// xmm3 used
    movq mm3, [INPC + 6*32]   ;// (2a) xmm3 <= dp[3]
     PFADD( mm7, mm2);        // (2b) xmm7 <= b[6] = (dp[7]*tg_1_16f) + dp[6]
// xmm2 free

// xmm0 used
    movq mm0, mm5;                   ;// (2b) xmm0 <= b[4]

    // 3) Apply E_T
    //
    //     e[4] = b[4] + b[6];
    //     e[5] = b[4] - b[6];
    //     e[6] = b[5] + b[7];
    //     e[7] = b[5] - b[7];
     PFADD( mm1, mm6 );              ;// (3) xmm1 <= e[6] = b[5] + b[7]

// xmm2 used
    movq mm2, mm3;                   ;// (2a) xmm2 <= dp[3]
     PFSUB( mm5, mm7);               ;// (3) xmm5 <= e[5] = b[4] - b[6]

    PFMULM( mm3, TABLE, _OTAN2);    ;// (2a) xmm3 <= (dp[3]*tg_2_16f)
     PFSUB( mm4, mm6);               ;// (3) xmm4 <= e[7] = b[5] - b[7]
// xmm6 free

// xmm6 used
    movq mm6, [INPC + 2*32];          // (2a) xmm6 <= dp[2]
     PFADD( mm0, mm7);               ;// (3) xmm0 <= e[4] = b[4] + b[6]
// xmm7 free

// xmm7 used
    movq mm7, mm5;                   ;// (3) xmm7 <= e[5]

    // 4) Apply F_T
    //
    //    f[4] = e[4]
    //    f[5] = (e[5] + e[6]) * cos_4_16f;
    //    f[6] = (e[5] - e[6]) * cos_4_16f;
    //    f[7] = e[7]

// xmm0 <= f[4] = e[4]
// xmm4 <= f[7] = e[7]
     PFADD( mm5, mm1 );              ;// (4a) xmm5 <= (e[5]+e[6])

    PFSUB( mm7, mm1 );               ;// (4a) xmm7 <= (e[5]-e[6])
// xmm1 free
// xmm1 used
     movq mm1, [INPC + 0*32];         // (1) xmm1 <= dp[0]

movq [SCRATCH], mm0;                 ;// store f[4] in scratch space
// xmm0 free (in scratch)

    PFMULM( mm5, TABLE, _OCOS4); ;// (4b) xmm5 <= f[5] = (e[5] + e[6]) * cos_4_16f;
     PFADD( mm3, mm6);       ;// (2b) xmm3 <= b[2] = (dp[3]*tg_2_16f) + dp[2]

// xmm4 f[7] free (in scratch)

    PFMULM( mm6, TABLE, _OTAN2);    ;// (2a) xmm6 <= (dp[2]*tg_2_16f)
     ;

    PFMULM( mm7, TABLE, _OCOS4); ;// (4b) xmm7 <= f[6] = (e[5] - e[6]) * cos_4_16f;
// xmm0 used
     movq mm0, mm1;               ;// (1) xmm0 <= dp[0]
    
    PFSUB( mm6, mm2 );       ;// (2b) xmm6 <= b[3] = (dp[2]*tg_2_16f) - dp[3]
// xmm2 free
// xmm2 used
     movq mm2, [INPC+4*32];       ;// (1) xmm2 <= dp[1]

    PFADD( mm1, mm2 );            ;// (2b) xmm1 <= b[0] = dp[0] + dp[1]
     add INPC, 8;

    PFSUB( mm0, mm2 );            ;// (2b) xmm0 <= b[1] = dp[0] - dp[1]
// xmm2 free
     add OUTC, 4;

//    e[0] = b[0] + b[2];
//    e[1] = b[1] + b[3];
//    e[2] = b[1] - b[3];
//    e[3] = b[0] - b[2];


// xmm2 used 
    movq mm2, mm1;                ;// (2b) xmm1 <= b[0]
     PFADD( mm1, mm3);            ;// (3) xmm1 <= e[0] = b[0] + b[2];

    PFSUB( mm2, mm3);             ;// (3) xmm3 <= e[3] = b[0] - b[2];
// xmm3 free
// xmm3 used
     movq mm3, mm0;               ;// (2b) xmm3 <= b[1]

    PFADD( mm0, mm6);             ;// (3) xmm0 <= e[1] = b[1] + b[3];
     ;

    PFSUB( mm3, mm6);             ;// (3) xmm3 <= e[2] = b[1] - b[3];
// xmm6 free
// xmm6 used
     movq mm6, mm2;               ;// (4) xmm6 <= f[3]

// xmm1 <= f[0]
// xmm0 <= f[1]
// xmm3 <= f[2]
// xmm2 <= f[3]
/*

    y[0*8] = SHIFT_ROUND_COLF( _F0 + _F4 );
    y[1*8] = SHIFT_ROUND_COLF( _F1 + _F5 );
    y[2*8] = SHIFT_ROUND_COLF( _F2 + _F6 );
    y[3*8] = SHIFT_ROUND_COLF( _F3 + _F7 );

    y[4*8] = SHIFT_ROUND_COLF( _F3 - _F7 );
    y[5*8] = SHIFT_ROUND_COLF( _F2 - _F6 );
    y[6*8] = SHIFT_ROUND_COLF( _F1 - _F5 );
    y[7*8] = SHIFT_ROUND_COLF( _F0 - _F4 );

    6a) float -> int32 conversion (cvtps2pi)
       pf2id mm0, mm0;  // mm0 <= int32 versions of y[1 0]

    6b) round compensation, adjust int32 value to correct round alignment
        paddd mm0, [rnd_compensation]

    6c) stage1 descaling shift
        psrad mm0, DESCALE_SHIFT1

    6d) int32 -> int16 conversion (packssdw)
        packssdw mm0, mm0; // mm0 <= int16 versions of y[1 0 1 0] (duplicates)

    6e) stage2 descaling shift (psraw)
       psraw mm0, DESCALE_SHIFT2; // mm0 <= final descaled output, {-256,+255}

    6f) store results
    ** The descale1/descale2 shifts are chosen to clip final result to 
       {-256,+255} (9-bit range)

*/ 

    PFADD( mm2, mm4);           ;// (5) xmm2 <= y[3] = f[3] + f[7]
     ;

    PFSUB( mm6, mm4);           ;// (5) xmm6 <= y[4] = f[3] - f[7]
// xmm4 free
// xmm4 used
     movq mm4, mm0;             ;// xmm0 <= f[1]


    PF2ID( mm2, mm2);           ;// (6a) [y3] <float> -> <int32>
     ;

    PFADD( mm0, mm5);           ;// (5) xmm0 <= y[1] = f[1] + f[5]
     ;

    paddd mm2, [ rnd_compensation ] ;// (6b) ROUND compensation
     PF2ID( mm0, mm0 )              ;// (6a) [y1] <float> -> <int32>

    psrad mm2, DESCALE_SHIFT1;
     PFSUB( mm4, mm5);            ;// (5) xmm4 <= y[6] = f[1] - f[5]
// xmm5 free

    paddd mm0, [ rnd_compensation ] ;// (6b) ROUND compensation
// xmm5 used
     movq mm5, mm3;          ;// xmm5 <= f[2]

    psrad mm0, DESCALE_SHIFT1;
     PF2ID( mm6, mm6 );      ;// (6a) [y4] <float> -> <int32>

    packssdw mm0, mm2;           ;// (6d) mm0 <= [ y[3] | y[1]]
     PFADD( mm3, mm7 );            ;// (5) xmm3 <= y[2] = f[2] + f[6]
// xmm2 free
    
    psraw mm0, DESCALE_SHIFT2;    // (6e) clip y[3,1] to {-256,+255}
     PFSUB( mm5, mm7);            ;// (5) xmm5 <= y[5] = f[2] - f[6]

// xmm7 free
    paddd mm6, [ rnd_compensation ] ;// (6b) ROUND compensation
     PF2ID( mm3, mm3 );      ;// (6a) [y2] <float> -> <int32>

    movd [OUTC + 1*16], mm0;    l// (6f) store y[1]  (store y[3] next)
     PF2ID( mm5, mm5 );      ;// (6a) [y5] <float> -> <int32>

    movq mm2, [SCRATCH];     ;// xmm2 <= f[4]
     psrad mm6, DESCALE_SHIFT1;

    punpckhdq mm0, mm0;         // (6f) mm0 = [y3c1 y3c0 y3c1 y3c0] <int16>
     movq mm7, mm1;           ;// xmm7 <= f[0]

    paddd mm3, [rnd_compensation];  ;// (6b) ROUND compensation
     PFADD( mm1, mm2);            ;// (5) xmm2 <= y[0] = f[0] + f[4]

    movd [OUTC + 3*16], mm0;    // (6f) store y[1]  (store y[3] next)
// mm0 free
     PFSUB( mm7, mm2);            ;// (5) xmm7 <= y[7] = f[0] - f[4]

    psrad mm3, DESCALE_SHIFT1;
     PF2ID( mm1, mm1)            ;// (6a) [y0] <float> -> <int32>

    packssdw mm3, mm6;           ;// (6d) mm3 <= [ y[4] y[2] ]
     PF2ID( mm7, mm7)            ;// (6a) [y7] <float> -> <int32>

    paddd mm5, [rnd_compensation];  ;// (6b) ROUND compensation
     PF2ID( mm4, mm4)            ;// (6a) [y6] <float> -> <int32>

    paddd mm1, [rnd_compensation];  ;// (6b) ROUND compensation
     psraw mm3, DESCALE_SHIFT2;    // (6e) clip y[4,2] to {-256,+255}

    paddd mm7, [rnd_compensation];  ;// (6b) ROUND compensation
     psrad mm5, DESCALE_SHIFT1;        

    paddd mm4, [rnd_compensation];  ;// (6b) ROUND compensation
     psrad mm1, DESCALE_SHIFT1;        

    movd [ OUTC + 2*16], mm3;    // (6f) store y[2]  (store y[4] next)
     psrad mm4, DESCALE_SHIFT1;

// mm1 free
    punpckhdq mm3, mm3;            ;//(6f) mm3 = [y4c1 y4c0 y4c1 y4c0] <int16>
     psrad mm7, DESCALE_SHIFT1;        

    packssdw mm5, mm1;            ;// (6d) mm5 = [y0 y5] <int16>
     packssdw mm7, mm4;            ;// (6d) mm7 = [y6 y7] <int16>

    movd [ OUTC + 4*16], mm3;    // (6f) store y[4]
     psraw mm5, DESCALE_SHIFT2;    // (6e) clip y[5,0] to {-256,+255}

    cmp edi, 8;
     psraw mm7, DESCALE_SHIFT2;    // (6e) clip y[6,7] to {-256,+255}

    movd [ OUTC + 5*16], mm5;    // (6f) store y[5]
     punpckhdq mm5, mm5;            ;// (6f) mm3 = [y0c1 y0c0 y0c1 y0c0] <int16>

    movd [ OUTC + 7*16], mm7;    // (6f) store y[7]
     punpckhdq mm7, mm7;            ;// (6f) mm3 = [y6c1 y6c0 y6c1 y6c0] <int16>

    movd [ OUTC + 0*16], mm5;    // (6f) store y[0]

    movd [ OUTC + 6*16], mm7;    // (6f) store y[6]

    jl acc_idct_colloop1;
  // end for ( x=0; x < 8; x=x+2 )

   FEMMS;
  }
}


////////////////////////////////////////////////////////////////////////
//
//
// "Equivalent C-code" for 3DNow-AP922float
//
//
////////////////////////////////////////////////////////////////////////
void
idct_ap922float_amdx87( short *data )
{
//  static tdnfloat fTempArray[64]; // intermediate (row-iDCT output) array
  static tdnfloat a[8], b[8], e[8], dp[8]; // intermediate results
  static tdnfloat tmp[2];  // temp calculation, scratch pad
  static int xi[8]; // temp 32-bit ints for initial scale-up shift

  short *x; // pointer to input
  short *y; // pointer to final output
  tdnfloat *yr, *xc; // pointer to output row
  tdnfloat *ti; // pointer to IDCT row coefficient table

  int i;

  ////////////////////////////////////////////////////////////////////////
  //
  // AP922float iDCT row transform (AMD 3DNow)
  //
  // Row IDCT operator :A_T*M_T*P_T
  // Let Y=[output column data, 8 elements] 32-bit IEEE-754 float
  //     X=[input column data, 8 elements] 16-bit short integer
  //
  //     Y= [ A_T*M_T*P_T ] * X
  //
  //   (Y and X are both column vectors)

  for ( i = 0; i < 8; ++i ) // row iDCT
  {
    ti = &tab_i_01234567tdn[ i * 32 ];
    x = &data[ 8*i ];
    yr = ((tdnfloat *) fTempArray) + 8*i; // intermediate output, select row#i

    // 0) Prescale input values by left-shifting them, convert to float
    //
    // The prescaling operation duplicates the Intel-SSE code.  
    //   The prescaling arises from an optimzation within the actual SSE 
    // code.  Our 3D-Now code uses "pslld+pand" to convert 16-bit integers
    // (short) to 32-bit integers(long.)  On AMD's 3D-Now CPUs,
    // pi2fd can be paired with any MMX instruction, hence "port-contention"
    // is not an issue here as it is with Intel's Pentium3.
    //
    // Using pand/pslld applies an equivalent prescalar multiplier of 65536
    // (left shift by 16.)
    //
    // Here, the prescaling is included in this C-Code to mimic the 3DN code.

    // 1) apply P_T operator, permute input values

    xi[0] = ROW_PRESCALE( x[0] );
    xi[1] = ROW_PRESCALE( x[1] );
    xi[2] = ROW_PRESCALE( x[2] );
    xi[3] = ROW_PRESCALE( x[3] );
    xi[4] = ROW_PRESCALE( x[4] );
    xi[5] = ROW_PRESCALE( x[5] );
    xi[6] = ROW_PRESCALE( x[6] );
    xi[7] = ROW_PRESCALE( x[7] );

//////////////////////////////////////////////////////////////////////
//
// original table order,
//    The scaled-int (MMX) iDCT implementation uses "PMADDWD", which
//    simultaneously multiplies and adds adjacent data elements.  The
//    original AP922 w[] table is ordered for optimum execution
//    with pmaddwd and paddd.
//
//    a0 = x[0] * ti[0]  + x[4] * ti[1]  + x[2] * ti[4]  + x[6] * ti[5];
//    a1 = x[0] * ti[2]  + x[4] * ti[3]  + x[2] * ti[6]  + x[6] * ti[7];
//    a2 = x[0] * ti[8]  + x[4] * ti[9]  + x[2] * ti[12] + x[6] * ti[13];
//    a3 = x[0] * ti[10] + x[4] * ti[11] + x[2] * ti[14] + x[6] * ti[15];
//
//    b0 = x[1] * ti[16] + x[5] * ti[17] + x[3] * ti[20] + x[7] * ti[21];
//    b1 = x[1] * ti[18] + x[5] * ti[19] + x[3] * ti[22] + x[7] * ti[23];
//    b2 = x[1] * ti[24] + x[5] * ti[25] + x[3] * ti[28] + x[7] * ti[29];
//    b3 = x[1] * ti[26] + x[5] * ti[27] + x[3] * ti[30] + x[7] * ti[31];    
//
//////////////////////////////////////////////////////////////////////


    // 2) Apply M_T operator 
    //    
    //    Differences in the 
    //    Although AMD 3D-Now lacks the floating-point equivalent of pmaddwd,
    //    3D-Now does include a nifty PFACC instruction, which performs a 
    //    "horizontal" pfadd.  Pairing PFACC with PFMUL yields a 32-bit float
    //    equivalent of MMX PMADDWD.  (Of course PFACC/PFMUL consumes a total
    //    of 4 MMX registers, versus 2 for PMADDWD.)
    //    The increase in MMX register consumption means that the 3D-Now
    //    implementation must spill intermediate results to memory.  
    //    The w[] coefficient table is re-ordered for optimum 3D-Now
    //    execution.  
    //
    //    The SSE-iDCT packing and 3D-Now use *DIFFERENT* vector packing order!

    //    AMD 3D-Now packing :
    //
    //       register mmX = [float1 float0]  <2 floats per MMX register>
    //    a[0] = xi[0]*ti[0] + xi[2]*ti[1]  + xi[4]*ti[2]  + xi[6]*ti[3];
    //
    //       ;// data has alreay been converted to float, and re-ordered
    //       ;// 
    //  
    //       MOVQ  mm0, [ INPUT]     ; // mm0 <= [ xi[2]   xi[0] ]
    //       MOVQ  mm1, [ INPUT+1*8] ; // mm1 <= [ xi[6]   xi[4] ]
    //
    //       ;// mm0 <= [float<xi[2]*ti[1]>  float<xi[0]*ti[0]>]
    //       ;// mm1 <= [float<xi[6]*ti[3]>  float<xi[4]*ti[2]>]
    //       PFMUL( mm0, &TABLE[0] );  // address of ti[0]
    //       PFMUL( mm1, &TABLE[2] );  // address of ti[2]
    //
    //       ;// mm0 <= [float<   a0b    >   float<    a0a   > ]
    //       PFACC( mm0, mm1 );  // horizontal accumulate
    //
    //       ;// repeat operations for a1b, a1a
    //       ;// mm2 <= [float<   a1b    >   float<    a1a   > ]
    //
    //       PFACC( mm0, mm2 ); // horizontal accumulate
    //       ;// mm2 <= [float<   a1     >   float<    a0   > ]
    //         

    a[0] = xi[0]*ti[0] + xi[2]*ti[1]  + xi[4]*ti[2]  + xi[6]*ti[3];
    a[1] = xi[0]*ti[4] + xi[2]*ti[5]  + xi[4]*ti[6]  + xi[6]*ti[7];
    a[2] = xi[0]*ti[8] + xi[2]*ti[9]  + xi[4]*ti[10] + xi[6]*ti[11];
    a[3] = xi[0]*ti[12]+ xi[2]*ti[13] + xi[4]*ti[14] + xi[6]*ti[15];

    b[0] = xi[1]*ti[16]+ xi[3]*ti[17] + xi[5]*ti[18] + xi[7]*ti[19];
    b[1] = xi[1]*ti[20]+ xi[3]*ti[21] + xi[5]*ti[22] + xi[7]*ti[23];
    b[2] = xi[1]*ti[24]+ xi[3]*ti[25] + xi[5]*ti[26] + xi[7]*ti[27];
    b[3] = xi[1]*ti[28]+ xi[3]*ti[29] + xi[5]*ti[30] + xi[7]*ti[31];

    // 3) Apply A_T operator, store outputs
    //
    //    1  0  0  0   1  0  0  0
    //    0  1  0  0   0  1  0  0
    //    0  0  1  0   0  0  1  0
    //    0  0  0  1   0  0  0  1
    //    0  0  0  1   0  0  0 -1
    //    0  0  1  0   0  0 -1  0
    //    0  1  0  0   0 -1  0  0
    //    1  0  0  0  -1  0  0  0

    //    Since the internal calculations produce floating-point results,
    //    the row-store operation does not round/scale the results.

    yr[0] = ( a[0] + b[0] );
    yr[1] = ( a[1] + b[1] );
    yr[2] = ( a[2] + b[2] );
    yr[3] = ( a[3] + b[3] );

    yr[4] = ( a[3] - b[3] );
    yr[5] = ( a[2] - b[2] );
    yr[6] = ( a[1] - b[1] );
    yr[7] = ( a[0] - b[0] );
  } // end for( i = 0; i < 8; ++i ) // end of row iDCT

  //     AP922float iDCT row transform done
  //
  ////////////////////////////////////////////////////////////////////////


  //////////////////////////////////////////////////////////////////////
  //
  // Column IDCT operator :A_T*(F_T*E_T*B_T*D_T)*P_T
  // Let Y=[output column data, 8 elements], 16-bit short integer
  //     X=[input column data, 8 elements], 32-bit IEEE-754 float
  //
  //     Y= [ A_T*(F_T*E_T*B_T*D_T)*P_T ] * X
  //
  //   (Y and X are both column vectors)

  for ( i = 0; i < 8; ++i ) // column iDCT
  {
    xc = ((tdnfloat *) fTempArray) + i; // select column #i
    y = &data[ i ];

    // 1) Apply (D_T * P_T) - the cos() coefficients of D_T are implicit
    //    in the idct_row operation.  But we still need to apply the
    //    shuffling operation of D_T.
    //
    //    1  0  0  0   0  0  0  0
    //    0  0  0  0   1  0  0  0
    //    0  0  1  0   0  0  0  0
    //    0  0  0  0   0  0  1  0
    //    0  1  0  0   0  0  0  0
    //    0  0  0  0   0  0  0  1
    //    0  0  0  1   0  0  0  0
    //    0  0  0  0   0  1  0  0

    dp[0] = xc[ 0 *8];
    dp[1] = xc[ 4 *8];
    dp[2] = xc[ 2 *8];
    dp[3] = xc[ 6 *8];

    dp[4] = xc[ 1 *8];
    dp[5] = xc[ 7 *8];
    dp[6] = xc[ 3 *8];
    dp[7] = xc[ 5 *8];
 
    // 2) Apply B_T
    //
    //    1  1  0  0
    //    1 -1  0  0
    //    0  0  1 t2
    //    0  0 t2 -1
    //                1 t1  0  0
    //               t1 -1  0  0
    //                0  0  1 t3
    //                0  0 t3 -1
 
 
    b[0] =   dp[1]            + dp[0];
    b[1] =   dp[0]            - dp[1];

    b[2] = ( dp[3]*tg_2_16f ) + dp[2];
    b[3] = ( dp[2]*tg_2_16f ) - dp[3];

    b[4] = ( dp[5]*tg_1_16f ) + dp[4];
    b[5] = ( dp[4]*tg_1_16f ) - dp[5];

    b[6] = ( dp[7]*tg_3_16f ) + dp[6];
    b[7] = ( dp[6]*tg_3_16f ) - dp[7];
  
    // 3) Apply E_T
    //
    //    1  0  1  0
    //    0  1  0  1
    //    0  1  0 -1
    //    1  0 -1  0
    //                1  0  1  0
    //                1  0 -1  0
    //                0  1  0  1
    //                0  1  0 -1
 
    e[0] = b[0] + b[2];
    e[1] = b[1] + b[3];
    e[2] = b[1] - b[3];
    e[3] = b[0] - b[2];
    e[4] = b[4] + b[6];
    e[5] = b[4] - b[6];
    e[6] = b[5] + b[7];
    e[7] = b[5] - b[7];

    // 4) Apply F_T
    //
    //    1  0  0  0
    //    0  1  0  0
    //    0  0  1  0
    //    0  0  0  1
    //                1  0  0  0
    //                0  1  0  0
    //                0  0  1  0
    //                0  0  0  1

#define _F0 e[0] 
#define _F1 e[1] 
#define _F2 e[2] 
#define _F3 e[3] 
#define _F4 e[4] 
#define _F5 e[5] 
#define _F6 e[6] 
#define _F7 e[7] 

    tmp[0] = (e[5] + e[6]) * cos_4_16f;
    tmp[1] = (e[5] - e[6]) * cos_4_16f;
    _F5 = tmp[0];
    _F6 = tmp[1];

    // 5) Apply A_T
    //
    //    1  0  0  0   1  0  0  0
    //    0  1  0  0   0  1  0  0
    //    0  0  1  0   0  0  1  0
    //    0  0  0  1   0  0  0  1
    //    0  0  0  1   0  0  0 -1
    //    0  0  1  0   0  0 -1  0
    //    0  1  0  0   0 -1  0  0
    //    1  0  0  0  -1  0  0  0
    //
    //    yfloat[0]= F0 + F4
    //    yfloat[1]= F1 + F5
    //    yfloat[2]= F2 + F6
    //    yfloat[3]= F3 + F7
    //           
    //    yfloat[4]= F3 - F7
    //    yfloat[5]= F2 - F6
    //    yfloat[6]= F1 - F5
    //    yfloat[7]= F0 - F4
    //
    //
    // 6) float -> int, shift/round to final output y[]
    //    The final shift&round operation reverses the row-input prescaling.
    //    It also applies the chosen rounding-mode (accurate or fast.)
    //
    //    Note, the C-code below differs *substantially* from the AMD_3DNOW
    //    implementation.  The 3D_Now code applies this basic sequence:
    //
    //       ;// mm0 = [y1 y0] <float32>
    //       ;// mm1 = [y3 y2] <float32>
    //
    //      PF2ID mm0, mm0;  // mm0 <= [y1 y0] <int32>
    //      PF2ID mm1, mm1;  // mm1 <= [y3 y2] <int32>
    //
    //       ;// "0.5" is a 32-bit integer constant scaled up by some bitshift
    //      paddd mm0, [rnd_compensation]; // mm0 <=[y1+"0.5" y0+"0.5"]
    //      paddd mm1, [rnd_compensation]; // mm0 <=[y3+"0.5" y2+"0.5"]
    //
    //      psrad mm0, DESCALE_SHIFT1;     // stage1 shift
    //      psrad mm1, DESCALE_SHIFT1;     // stage1 shift
    //
    //      packssdw mm0, mm1;  // mm0 <= [y3 y2 y1 y0] <int16>
    //
    //      psraw mm0, DESCALE_SHIFT2;  // clip y[] to the range {-256,+255}
    //
    //       ;// DESCALE_SHIFT1 + DESCALE_SHIFT2 = PRESCALE_SHIFT 
    //       ;//      9         +        7       =       16
    //
    //      movq [OUTC+...*8], mm0;
 

    y[0*8] = SHIFT_ROUND_COLF( _F0 + _F4 );
    y[1*8] = SHIFT_ROUND_COLF( _F1 + _F5 );
    y[2*8] = SHIFT_ROUND_COLF( _F2 + _F6 );
    y[3*8] = SHIFT_ROUND_COLF( _F3 + _F7 );

    y[4*8] = SHIFT_ROUND_COLF( _F3 - _F7 );
    y[5*8] = SHIFT_ROUND_COLF( _F2 - _F6 );
    y[6*8] = SHIFT_ROUND_COLF( _F1 - _F5 );
    y[7*8] = SHIFT_ROUND_COLF( _F0 - _F4 );

  } // end for ( i = 0; i < 8; ++i ) // end of 3DNow column iDCT

  //    AP922float iDCT column transform done 
  //
  ////////////////////////////////////////////////////////////////////////

  ////////////////////////////////////////////////////////////////////////
  //    Post transform clipping
  //
  //    In standard-C, the output clip code adds significantly to 
  //    execution time.  

#define IDCT_CLIP_LOW -256  // IDCT output range is 9-bits
#define IDCT_CLIP_HIGH 255  // IDCT output range is 9-bits

  for ( i = 0; i < 64; ++i )
  { // clip output to {-256,+255}
    if ( data[ i ] < IDCT_CLIP_LOW )
      data[ i ] = IDCT_CLIP_LOW;

    if ( data[ i ] > IDCT_CLIP_HIGH )
      data[ i ] = IDCT_CLIP_HIGH;
  }
}
