/*
 --------------------------------------------------------
 idct_ap922x87.c - AP922float iDCT, C-code implementation
 --------------------------------------------------------

 AP-922 floating-point iDCT (x87 FPU)
 ------------------------------------
 
    Intel Application Note AP-922 discusses an MMX (scaled integer) 
  implementation of the MPEG fDCT/iDCT algorithm.  
  "AP922float" is a floating-point adaptation of the AP922 algorithm.
 
    The X87 implementation (this file) should compile and run on any
  i486DX or higher CPU.  *This* C-implementation attempts to emulate the 
  SSE-version of AP922float.  The emulation is only approximately, however,
  due to 2 numerical issues :

 Numerical Issue #1 : float->int conversion
 ------------------------------------------

    All modern CPUs can accurately convert float->int, but since this
  algorithm was coded on an x86 platform (with its aging heritage),
  float conversion is not as straightforward as one would expect.

    In a C-compiler, the fastest way to convert float->int is to
  simply type-cast.  Unfortunately, typecasting results in 
  truncation.  Truncation is *NOT* equivalent to rounding down!
  Truncated positive numbers are rounded down (toward 0), while 
  truncated negative numbers are rounded up (toward 0.)  On 
  average, the truncated number's magnitude is reduced by 0.5.

    The standard C-library function "floor()" explicitly rounds
  down the input, but calling a function to perform a simple
  rounding operation wastes CPU time, especially when each iDCT
  call requires 64 float->int conversions.
 
    The x87 FPU supports all rounding modes, but a FPU state change 
  must be executed prior to the rounding instructions, and this
  operation is not directly accessible from a high-level language.
  Therefore, if we are not willing to hand-code assembly,
  we have two choices: 

      a) maintain speed, use truncation (C type-casting)
    or
      b) maintain accuracy, use C-function floor()
 


   a) "FAST_FLOAT2INT" uses a truncation policy to convert float->int.
      "Truncation" is equivalent to rounding-down (toward 0) for positive #s,
      and rounding-up (toward 0) for negative #s.  Statistically, the 
      truncation introduces an average offset of -0.5, meaning the least
      significant bit of the converted-int is wrong.
  
      To minimize the error impact in the final iDCT output, the
      C-code iDCT prescales input data by a 16-bit left-shift. 
      The output is descaled by a complementary right-shift.
      The prescale/descale operation reduces truncation error by a factor
      of 65536.

   b) Precise float->int conversion can be achieved with the standard 
      C-library function 'floor( value + 0.5)'  This statement implements
      a standard rounding policy : "round-up if >0.5, otherwise down."
    
      The floor() function incurs a severe performance penalty: with 
      accurate_rounding, the iDCT's overall execution time doubles!
    
      Alternatively, the accuate rounding could be accelerated by hand-
      coding in x86 assembly.


  The actual-SSE assembly code implements accurate-rounding for
  float->int conversion, but without additional speed loss.  The


 SSE Numerical precision vs X87 numerical precision
 ------------------------------------------------------------
    On Intel x86 CPUs, FPU add/multiply operations are always calculated 
  with 80-bit internal precision.  Therefore, all other things being 
  equal, the X87 iDCT will be more accurate than Intel-SSE, even
  when the X87 iDCT stores intermediate results in 32-bit float.

    As a side note, the X87 iDCT can achieve the precision-level of the 
  IEEE-1180/1990 reference iDCT.  The "Reference-precision" mode is
  accessed by setting 'typdedef double ssefloat', and disabling
  "USE_FAST_FLOAT2INT" (delete the #define statement.)  

 Output clipping 
 ------------------------------------

    The iDCT's output range is limited to {-256,+255}
  The clipper is easy to implement in standard C, but slow to execute.
  In this C-implementation, usage of the clipper is left as a decision 
  to the user.  For strict IEEE-compliance, the output clipper must be used.

  (MMX offers several instructions to efficiently implement clipping.
   Thus, output clipping is done in both AP922float_SSE and 
   AP922float_3DN.)
 
 ------------------------------------------
 LEGAL MUMBO-JUMBO : Disclaimer of Warranty
 ------------------------------------------
  This software is available to the user without any license fee or
  royalty on an "as is" basis.  The author disclaims any and all warranties,
  whether express, implied, or statuary, including any implied warranties or 
  merchantability or of fitness for a particular purpose.  In no event shall 
  the copyright-holder be liable for any incidental, punitive, or 
  consequential damages of any kind whatsoever arising from the use of these
  programs.

  This disclaimer of warranty extends to the user of these programs and user's
  customers, employees, agents, transferees, successors, and assigns.

 Revision history
 ----------------
 09/03/2000 initial release of 'AP922float_x87' iDCT
            Based on AP922, this iDCT uses floating-point instructions for 
            enhanced accuracy.  To the best of my knowledge, AP922float is
            fully IEEE 1180/1990 compliant iDCT, including output-range 
            clipping to {-256,+255}
            When set to use double-precision and accurate-rounding (see
            above), AP922float is numerically equivalent to the 
            IEEE-1180/1990 reference_iDCT.
            
 09/03/2000 liaor@iname.com   http://members.tripod.com/~liaor
 liaor@iname.com   http://members.tripod.com/~liaor
*/
#include<math.h>  // for "floor()" function

typedef float ssefloat;  // ssefloat = IEEE-754 32-bit float

  #define PSCALE_SHIFT 16  // 16 bit left-shift
  #define DESCALE_SHIFT 16  // 16 bit right-shift
  #define DESCALE_ROUND_COMPENSATION (1<<(PSCALE_SHIFT-1))
  
  #define ROW_PRESCALE( x ) (((int)(x))<<PSCALE_SHIFT)

#define USE_FAST_FLOAT2INT 1
// comment out the previous line to use accurate (slower) rounding

#ifdef USE_FAST_FLOAT2INT

  // approximate rounding
  #define SHIFT_ROUND_COLF( x )  ( ( ( (int)(x) ) + DESCALE_ROUND_COMPENSATION) >> DESCALE_SHIFT)

#else

  // accurate rounding
  //#define SHIFT_ROUND_COLF( x )  (  (((int)floor( (x) + DESCALE_ROUND_COMPENSATION )))>>DESCALE_SHIFT  ) 
  #define SHIFT_ROUND_COLF( x )  (  ( ((int)floor( (x) )) + DESCALE_ROUND_COMPENSATION )>>DESCALE_SHIFT  ) 

#endif // USE_FAST_FLOAT2INT



#define NCOS1_16 (0.980785280403230449126182236134239) // cosine(  Pi/16 )
#define NCOS2_16 (0.923879532511286756128183189396788) // cosine( 2Pi/16 )
#define NCOS3_16 (0.831469612302545237078788377617906) // cosine( 3Pi/16 )
#define NCOS4_16 (0.707106781186547524400844362104849) // cosine( 4Pi/16 )
#define NCOS5_16 (0.555570233019602224742830813948533) // cosine( 5Pi/16 )
#define NCOS6_16 (0.382683432365089771728459984030399) // cosine( 6Pi/16 )
#define NCOS7_16 (0.195090322016128267848284868477022) // cosine( 7Pi/16 )

#define TANG1_16 ( NCOS7_16 / NCOS1_16) // tangent( Pi/16)
#define TANG2_16 ( NCOS6_16 / NCOS2_16) // tangent(2Pi/16)
#define TANG3_16 ( NCOS5_16 / NCOS3_16) // tangent(3Pi/16)

// Scaled floating-point coefficient table
// w00, w04, w08, w12
// w01, w05, w09, w13
// w02, w06, w10, w14
// w03, w07, w11, w15    ( all elements in float table are multiplied by 0.5)
// w16, w20, w24, w28
// w17, w21, w25, w29
// w18, w22, w26, w30
// w19, w23, w27, w31
//
// Compared to the integer-based table, the float table is downscaled by a
// multiplication factor of 0.5
// Scaled floating-point coefficient table
//      Original iDCT               SSE-iDCT 
//     table-ordering      ->    table-ordering
//    ------------------       ------------------
//    w00, w02, w04, w06,  ->  w00, w04, w08, w12,
//    w01, w03, w05, w07,  ->  w01, w05, w09, w13,
//    w08, w10, w12, w14,  ->  w02, w06, w10, w14,
//    w09, w11, w13, w15,  ->  w03, w07, w11, w15,
//    w16, w18, w20, w22,  ->  w16, w20, w24, w28,
//    w17, w19, w21, w23,  ->  w17, w21, w25, w29,
//    w24, w26, w28, w30,  ->  w18, w22, w26, w30,
//    w25, w27, w29, w31   ->  w19, w23, w27, w31
//
// Compared to the integer-based table, the float table is downscaled by a
// multiplication factor of 0.25*"DESCALE_SHIFT0"

#define IDCT_CONSTANT (0.25)
 // all table entries multiplied by (0.5*0.5)=0.25


#define RS0 (NCOS4_16*IDCT_CONSTANT)  // iDCT row#0 scalar
#define RS1 (NCOS1_16*IDCT_CONSTANT)  // iDCT row#1 scalar
#define RS2 (NCOS2_16*IDCT_CONSTANT)  // iDCT row#2 scalar
#define RS3 (NCOS3_16*IDCT_CONSTANT)  // iDCT row#3 scalar
#define RS4 (NCOS4_16*IDCT_CONSTANT)  // iDCT row#4 scalar
#define RS5 (NCOS3_16*IDCT_CONSTANT)  // iDCT row#5 scalar
#define RS6 (NCOS2_16*IDCT_CONSTANT)  // iDCT row#6 scalar
#define RS7 (NCOS1_16*IDCT_CONSTANT)  // iDCT row#7 scalar

const static ssefloat tab_i_01234567x87[] =
{ 
 // Row #0
  NCOS4_16*RS0,  NCOS4_16*RS0,  NCOS4_16*RS0,  NCOS4_16*RS0, 
  NCOS2_16*RS0,  NCOS6_16*RS0, -NCOS6_16*RS0, -NCOS2_16*RS0, 
  NCOS4_16*RS0, -NCOS4_16*RS0, -NCOS4_16*RS0,  NCOS4_16*RS0, 
  NCOS6_16*RS0, -NCOS2_16*RS0,  NCOS2_16*RS0, -NCOS6_16*RS0, 
  NCOS1_16*RS0,  NCOS3_16*RS0,  NCOS5_16*RS0,  NCOS7_16*RS0, 
  NCOS3_16*RS0, -NCOS7_16*RS0, -NCOS1_16*RS0, -NCOS5_16*RS0, 
  NCOS5_16*RS0, -NCOS1_16*RS0,  NCOS7_16*RS0,  NCOS3_16*RS0, 
  NCOS7_16*RS0, -NCOS5_16*RS0,  NCOS3_16*RS0, -NCOS1_16*RS0, 

 // Row #1
  NCOS4_16*RS1,  NCOS4_16*RS1,  NCOS4_16*RS1,  NCOS4_16*RS1, 
  NCOS2_16*RS1,  NCOS6_16*RS1, -NCOS6_16*RS1, -NCOS2_16*RS1, 
  NCOS4_16*RS1, -NCOS4_16*RS1, -NCOS4_16*RS1,  NCOS4_16*RS1, 
  NCOS6_16*RS1, -NCOS2_16*RS1,  NCOS2_16*RS1, -NCOS6_16*RS1, 
  NCOS1_16*RS1,  NCOS3_16*RS1,  NCOS5_16*RS1,  NCOS7_16*RS1, 
  NCOS3_16*RS1, -NCOS7_16*RS1, -NCOS1_16*RS1, -NCOS5_16*RS1, 
  NCOS5_16*RS1, -NCOS1_16*RS1,  NCOS7_16*RS1,  NCOS3_16*RS1, 
  NCOS7_16*RS1, -NCOS5_16*RS1,  NCOS3_16*RS1, -NCOS1_16*RS1, 

 // Row #2
  NCOS4_16*RS2,  NCOS4_16*RS2,  NCOS4_16*RS2,  NCOS4_16*RS2, 
  NCOS2_16*RS2,  NCOS6_16*RS2, -NCOS6_16*RS2, -NCOS2_16*RS2, 
  NCOS4_16*RS2, -NCOS4_16*RS2, -NCOS4_16*RS2,  NCOS4_16*RS2, 
  NCOS6_16*RS2, -NCOS2_16*RS2,  NCOS2_16*RS2, -NCOS6_16*RS2, 
  NCOS1_16*RS2,  NCOS3_16*RS2,  NCOS5_16*RS2,  NCOS7_16*RS2, 
  NCOS3_16*RS2, -NCOS7_16*RS2, -NCOS1_16*RS2, -NCOS5_16*RS2, 
  NCOS5_16*RS2, -NCOS1_16*RS2,  NCOS7_16*RS2,  NCOS3_16*RS2, 
  NCOS7_16*RS2, -NCOS5_16*RS2,  NCOS3_16*RS2, -NCOS1_16*RS2, 

 // Row #3
  NCOS4_16*RS3,  NCOS4_16*RS3,  NCOS4_16*RS3,  NCOS4_16*RS3, 
  NCOS2_16*RS3,  NCOS6_16*RS3, -NCOS6_16*RS3, -NCOS2_16*RS3, 
  NCOS4_16*RS3, -NCOS4_16*RS3, -NCOS4_16*RS3,  NCOS4_16*RS3, 
  NCOS6_16*RS3, -NCOS2_16*RS3,  NCOS2_16*RS3, -NCOS6_16*RS3, 
  NCOS1_16*RS3,  NCOS3_16*RS3,  NCOS5_16*RS3,  NCOS7_16*RS3, 
  NCOS3_16*RS3, -NCOS7_16*RS3, -NCOS1_16*RS3, -NCOS5_16*RS3, 
  NCOS5_16*RS3, -NCOS1_16*RS3,  NCOS7_16*RS3,  NCOS3_16*RS3, 
  NCOS7_16*RS3, -NCOS5_16*RS3,  NCOS3_16*RS3, -NCOS1_16*RS3, 

 // Row #4
  NCOS4_16*RS4,  NCOS4_16*RS4,  NCOS4_16*RS4,  NCOS4_16*RS4, 
  NCOS2_16*RS4,  NCOS6_16*RS4, -NCOS6_16*RS4, -NCOS2_16*RS4, 
  NCOS4_16*RS4, -NCOS4_16*RS4, -NCOS4_16*RS4,  NCOS4_16*RS4, 
  NCOS6_16*RS4, -NCOS2_16*RS4,  NCOS2_16*RS4, -NCOS6_16*RS4, 
  NCOS1_16*RS4,  NCOS3_16*RS4,  NCOS5_16*RS4,  NCOS7_16*RS4, 
  NCOS3_16*RS4, -NCOS7_16*RS4, -NCOS1_16*RS4, -NCOS5_16*RS4, 
  NCOS5_16*RS4, -NCOS1_16*RS4,  NCOS7_16*RS4,  NCOS3_16*RS4, 
  NCOS7_16*RS4, -NCOS5_16*RS4,  NCOS3_16*RS4, -NCOS1_16*RS4, 

 // Row #5
  NCOS4_16*RS5,  NCOS4_16*RS5,  NCOS4_16*RS5,  NCOS4_16*RS5, 
  NCOS2_16*RS5,  NCOS6_16*RS5, -NCOS6_16*RS5, -NCOS2_16*RS5, 
  NCOS4_16*RS5, -NCOS4_16*RS5, -NCOS4_16*RS5,  NCOS4_16*RS5, 
  NCOS6_16*RS5, -NCOS2_16*RS5,  NCOS2_16*RS5, -NCOS6_16*RS5, 
  NCOS1_16*RS5,  NCOS3_16*RS5,  NCOS5_16*RS5,  NCOS7_16*RS5, 
  NCOS3_16*RS5, -NCOS7_16*RS5, -NCOS1_16*RS5, -NCOS5_16*RS5, 
  NCOS5_16*RS5, -NCOS1_16*RS5,  NCOS7_16*RS5,  NCOS3_16*RS5, 
  NCOS7_16*RS5, -NCOS5_16*RS5,  NCOS3_16*RS5, -NCOS1_16*RS5, 

 // Row #6
  NCOS4_16*RS6,  NCOS4_16*RS6,  NCOS4_16*RS6,  NCOS4_16*RS6, 
  NCOS2_16*RS6,  NCOS6_16*RS6, -NCOS6_16*RS6, -NCOS2_16*RS6, 
  NCOS4_16*RS6, -NCOS4_16*RS6, -NCOS4_16*RS6,  NCOS4_16*RS6, 
  NCOS6_16*RS6, -NCOS2_16*RS6,  NCOS2_16*RS6, -NCOS6_16*RS6, 
  NCOS1_16*RS6,  NCOS3_16*RS6,  NCOS5_16*RS6,  NCOS7_16*RS6, 
  NCOS3_16*RS6, -NCOS7_16*RS6, -NCOS1_16*RS6, -NCOS5_16*RS6, 
  NCOS5_16*RS6, -NCOS1_16*RS6,  NCOS7_16*RS6,  NCOS3_16*RS6, 
  NCOS7_16*RS6, -NCOS5_16*RS6,  NCOS3_16*RS6, -NCOS1_16*RS6, 

 // Row #7
  NCOS4_16*RS7,  NCOS4_16*RS7,  NCOS4_16*RS7,  NCOS4_16*RS7, 
  NCOS2_16*RS7,  NCOS6_16*RS7, -NCOS6_16*RS7, -NCOS2_16*RS7, 
  NCOS4_16*RS7, -NCOS4_16*RS7, -NCOS4_16*RS7,  NCOS4_16*RS7, 
  NCOS6_16*RS7, -NCOS2_16*RS7,  NCOS2_16*RS7, -NCOS6_16*RS7, 
  NCOS1_16*RS7,  NCOS3_16*RS7,  NCOS5_16*RS7,  NCOS7_16*RS7, 
  NCOS3_16*RS7, -NCOS7_16*RS7, -NCOS1_16*RS7, -NCOS5_16*RS7, 
  NCOS5_16*RS7, -NCOS1_16*RS7,  NCOS7_16*RS7,  NCOS3_16*RS7, 
  NCOS7_16*RS7, -NCOS5_16*RS7,  NCOS3_16*RS7, -NCOS1_16*RS7, 

 // iDCT column coefficients
//  NCOS4_16, NCOS4_16, NCOS4_16, NCOS4_16,
//  TANG1_16, TANG1_16, TANG1_16, TANG1_16,
//  TANG2_16, TANG2_16, TANG2_16, TANG2_16,
//  TANG3_16, TANG3_16, TANG3_16, TANG3_16
};

// externally allocated array of 64 floats, for temporary storage
extern void *fTempArray;

const static ssefloat cos_4_16f = NCOS4_16;
const static ssefloat tg_1_16f= NCOS7_16/NCOS1_16; // tangent (Pi/16)
const static ssefloat tg_2_16f= NCOS6_16/NCOS2_16; // tangent (2Pi/16)
const static ssefloat tg_3_16f= NCOS5_16/NCOS3_16; // tangent (3Pi/16)


/*
; The original, unscaled idct coefficient table
; static const short w[32] = {
; FIX(cos_4_16), FIX(cos_2_16), FIX(cos_4_16), FIX(cos_6_16),
; FIX(cos_4_16), FIX(cos_6_16), -FIX(cos_4_16), -FIX(cos_2_16),
; FIX(cos_4_16), -FIX(cos_6_16), -FIX(cos_4_16), FIX(cos_2_16),
; FIX(cos_4_16), -FIX(cos_2_16), FIX(cos_4_16), -FIX(cos_6_16),
; FIX(cos_1_16), FIX(cos_3_16), FIX(cos_5_16), FIX(cos_7_16),
; FIX(cos_3_16), -FIX(cos_7_16), -FIX(cos_1_16), -FIX(cos_5_16),
; FIX(cos_5_16), -FIX(cos_1_16), FIX(cos_7_16), FIX(cos_3_16),
; FIX(cos_7_16), -FIX(cos_5_16), FIX(cos_3_16), -FIX(cos_1_16) };
;
; #define DCT_8_INV_ROW(x, y)


#define BITS_INV_ACC	4	//; 4 or 5 for IEEE
	// 5 yields higher accuracy, but lessens dynamic range on the input matrix
#define SHIFT_INV_ROW	(16 - BITS_INV_ACC)
#define SHIFT_INV_COL	(1 + BITS_INV_ACC )

#define RND_INV_ROW		(1 << (SHIFT_INV_ROW-1))
#define RND_INV_COL		(1 << (SHIFT_INV_COL-1)) 
#define RND_INV_CORR	(RND_INV_COL - 1)		//; correction -1.0 and round
//#define RND_INV_ROW		(1024 * (6 - BITS_INV_ACC)) //; 1 << (SHIFT_INV_ROW-1)
//#define RND_INV_COL		(16 * (BITS_INV_ACC - 3)) //; 1 << (SHIFT_INV_COL-1)


const static long r_inv_row[2] = { RND_INV_ROW, RND_INV_ROW};

//#define SHIFT_ROUND_COL( x ) ( ( (x) + RND_INV_COL ) >> (SHIFT_INV_COL) )
//#define SHIFT_ROUND_ROW( x ) ( ( (x) + RND_INV_ROW ) >> (SHIFT_INV_ROW) )


const static short tg_1_16[4] = {13036, 13036, 13036, 13036 }; //tg * (2<<16) + 0.5
const static short tg_2_16[4] = {27146, 27146, 27146, 27146 }; //tg * (2<<16) + 0.5
const static short tg_3_16[4] = {-21746, -21746, -21746, -21746 }; //tg * (2<<16) + 0.5
const static short cos_4_16[4] = {-19195, -19195, -19195, -19195 }; //cos * (2<<16) + 0.5
const static short ocos_4_16[4] = {23170, 23170, 23170, 23170 }; //cos * (2<<15) + 0.5
*/

// AP922float_x87(), X87 implementation of floating-point AP922
void
idct_ap922float_x87( short *data )
{
//  static ssefloat fTempArray[64]; // intermediate (row-iDCT output) matrix
  static ssefloat a[8], b[8], e[8], dp[8]; // intermediate results
  static ssefloat tmp[2];  // temp calculation, scratch pad

  static int xi[8]; // temp 32-bit ints for initial scale-up shift

//  static short const _one_corr=0x0001; 
 
  short *x; // pointer to input
  short *y; // pointer to final output
  ssefloat *yr, *xc; // pointer to output row
  ssefloat *ti; // pointer to IDCT row coefficient table

  int i;

  ////////////////////////////////////////////////////////////////////////
  //
  // AP922float iDCT row transform (x87 version)
  //
  // Row IDCT operator :A_T*M_T*P_T
  // Let Y=[output column data, 8 elements] 32-bit IEEE-754 float
  //     X=[input column data, 8 elements] 16-bit short integer
  //
  //     Y= [ A_T*M_T*P_T ] * X
  //
  //   (Y and X are both column vectors)

  for ( i = 0; i < 8; ++i ) // row iDCT
  {
    ti = &tab_i_01234567x87[ i * 32 ];
    x = &data[ 8*i ];
    yr = ((ssefloat *) fTempArray) + 8*i; // intermediate output, select row#i

    // *Prescale input values by left-shifting them
    //
    // The prescaling operation duplicates the Intel-SSE code.  
    //   The prescaling arises from an optimzation within the actual SSE 
    // code.  Our SSE code uses "pshufw+pand" to convert 16-bit integers
    // (short) to 32-bit integers(long.)  Using 'pand' instead of 'psrad'
    // reduces port-contention for the MMXSHIFT/PFADDER/PSHUFW resources.
    // However, the optimization's side effect is a scaleup of 65536 
    // (left shift by 16.)
    //
    // Here, the prescaling is included in this C-Code to mimic the SSE code.

    // 0) apply P_T operator, permute input values

    xi[0] = ROW_PRESCALE( x[0] );
    xi[1] = ROW_PRESCALE( x[1] );
    xi[2] = ROW_PRESCALE( x[2] );
    xi[3] = ROW_PRESCALE( x[3] );
    xi[4] = ROW_PRESCALE( x[4] );
    xi[5] = ROW_PRESCALE( x[5] );
    xi[6] = ROW_PRESCALE( x[6] );
    xi[7] = ROW_PRESCALE( x[7] );

//////////////////////////////////////////////////////////////////////
//
// original table order,
//    The scaled-int (MMX) iDCT implementation uses "PMADDWD", which
//    simultaneously multiplies and adds adjacent data elements.  The
//    original AP922 w[] table is ordered for optimum execution
//    with pmaddwd and paddd.
//
//    a0 = x[0] * ti[0]  + x[4] * ti[1]  + x[2] * ti[4]  + x[6] * ti[5];
//    a1 = x[0] * ti[2]  + x[4] * ti[3]  + x[2] * ti[6]  + x[6] * ti[7];
//    a2 = x[0] * ti[8]  + x[4] * ti[9]  + x[2] * ti[12] + x[6] * ti[13];
//    a3 = x[0] * ti[10] + x[4] * ti[11] + x[2] * ti[14] + x[6] * ti[15];
//
//    b0 = x[1] * ti[16] + x[5] * ti[17] + x[3] * ti[20] + x[7] * ti[21];
//    b1 = x[1] * ti[18] + x[5] * ti[19] + x[3] * ti[22] + x[7] * ti[23];
//    b2 = x[1] * ti[24] + x[5] * ti[25] + x[3] * ti[28] + x[7] * ti[29];
//    b3 = x[1] * ti[26] + x[5] * ti[27] + x[3] * ti[30] + x[7] * ti[31];    
//
//////////////////////////////////////////////////////////////////////


    // 1) Apply M_T operator 
    //    
    //    Since Intel SSE has no floating-point equivalent of pmaddwd,
    //    the w[] coefficient table must be changed.  The new table-order
    //    works with the available MULPS and ADDPS instructions.

    //    SSE packing :
    //
    //    let a_vX[] = ( a_vX[0] )   ,  where [0] is stored first (LSB)
    //                 ( a_vX[1] )   
    //                 ( a_vX[2] )
    //                 ( a_vX[3] )
    //
    //     ;// xmm0 <= [ xi[0] xi[0] xi[0] xi[0] ]
    //     ;// xmm1 <= [ xi[2] xi[2] xi[2] xi[2] ]
    //     ;// xmm2 <= [ xi[4] xi[4] xi[4] xi[4] ]
    //     ;// xmm3 <= [ xi[6] xi[6] xi[6] xi[6] ]
    //
    //     ;// TABLE[0 ] <= [ w12 w8  w4 w0] ;// 4 floats (16 bytes)
    //     ;// TABLE[4 ] <= [ w13 w9  w5 w1] ;// 4 floats (16 bytes)
    //     ;// TABLE[8 ] <= [ w14 w10 w6 w2] ;// 4 floats (16 bytes)
    //     ;// TABLE[12] <= [ w15 w11 w7 w3] ;// 4 floats (16 bytes)

    //   1a) Compute "vertical" vectors (a_v0[], a_v1[], ... a_v3[])
    //
    //    MULPS xmm0, [TABLE + 0*16]; // xmm0 <= a_v0[]
    //    MULPS xmm1, [TABLE + 1*16]; // xmm1 <= a_v1[]
    //    MULPS xmm2, [TABLE + 2*16]; // xmm2 <= a_v2[]
    //    MULPS xmm3, [TABLE + 3*16]; // xmm3 <= a_v3[]
    //
    //   1b) Sum the vertical vectors to produce a0, a1, a2, a3
    //
    //    ADDPS xmm1, xmm0;  // xmm1 <= a_v1[] + ( a_v0[]                   )
    //    ADDPS xmm2, xmm1; //  xmm2 <= a_v2[] + ( a_v0[] + a_v1[]          )
    //    ADDPS xmm3, xmm2; //  xmm3 <= a_v3[] + ( a_v0[] + a_v1[] + a_v2[] )
    //
    //     ;// xmm3 <= [ a3 a2 a1 a0 ]
    //
    //   Repeat (1a) and (1b) for the vectors b_v[], 
    //     ...
    //     ...
    //     ;// xmm7 <= [ b3 b2 b1 b0 ]
    //


 // a_v[]=  ( a_v0[] ) +  ( a_v1[] )  +  ( a_v2[] )  +  ( a_v3[] );
    a[0] = xi[0]*ti[0] + xi[2]*ti[4]  + xi[4]*ti[8]  + xi[6]*ti[12];
    a[1] = xi[0]*ti[1] + xi[2]*ti[5]  + xi[4]*ti[9]  + xi[6]*ti[13];
    a[2] = xi[0]*ti[2] + xi[2]*ti[6]  + xi[4]*ti[10] + xi[6]*ti[14];
    a[3] = xi[0]*ti[3] + xi[2]*ti[7]  + xi[4]*ti[11] + xi[6]*ti[15];

 // b_v[]=  ( b_v0[] ) +  ( b_v1[] )  +  ( b_v2[] )  +  ( b_v3[] );
    b[0] = xi[1]*ti[16]+ xi[3]*ti[20] + xi[5]*ti[24] + xi[7]*ti[28];
    b[1] = xi[1]*ti[17]+ xi[3]*ti[21] + xi[5]*ti[25] + xi[7]*ti[29];
    b[2] = xi[1]*ti[18]+ xi[3]*ti[22] + xi[5]*ti[26] + xi[7]*ti[30];
    b[3] = xi[1]*ti[19]+ xi[3]*ti[23] + xi[5]*ti[27] + xi[7]*ti[31];

    // 2) Apply A_T operator, store outputs
    //
    //    1  0  0  0   1  0  0  0
    //    0  1  0  0   0  1  0  0
    //    0  0  1  0   0  0  1  0
    //    0  0  0  1   0  0  0  1
    //    0  0  0  1   0  0  0 -1
    //    0  0  1  0   0  0 -1  0
    //    0  1  0  0   0 -1  0  0
    //    1  0  0  0  -1  0  0  0

    //    Since the internal calculations produce floating-point results,
    //    the row-store operation does not round/scale the results.
    //
    //       ;// xmm3 <= [ a3 a2 a1 a0 ]
    //       ;// xmm7 <= [ b3 b2 b1 b0 ]
    //
    //      MOVPS xmm4, xmm3; // xmm3 <= [a3 a2 a1 a0]
    //
    //      SUBPS xmm3, xmm7; // xmm3 <= [a3-b3 a2-b2 a1-b1 a0-b0]
    //      ADDPS xmm4, xmm7; // xmm4 <= [a3+b3 a2+b2 a1+b1 a0+b0]
    //
    //       ;// xmm4 <= [ y3 y2 y1 y0]
    //       ;// xmm3 <= [ y4 y5 y6 y7]  *** (reversed order)
    //
    //      SHUFPS xmm3, xmm3, 00011011b;// xmm4 <=[y7 y6 y5 y4]; // shuffle
    //
    //       ;// xmm3 <= [ y7 y6 y5 y4]  (correct order)
    //
    //      MOVPS [ OUTR + 0*16 ], xmm4; // store [y3 y2 y1 y0]
    //      MOVPS [ OUTR + 1*16 ], xmm3; // store [y7 y6 y5 y4]

    yr[0] = ( a[0] + b[0] );
    yr[1] = ( a[1] + b[1] );
    yr[2] = ( a[2] + b[2] );
    yr[3] = ( a[3] + b[3] );

    yr[4] = ( a[3] - b[3] );
    yr[5] = ( a[2] - b[2] );
    yr[6] = ( a[1] - b[1] );
    yr[7] = ( a[0] - b[0] );
  } // end for( i = 0; i < 8; ++i ) // end of row iDCT


  ////////////////////////////////////////////////////////////////////////
  //
  // AP922float iDCT column transform
  // Base code originally from VirtualDub (Avery Lee),
  //   http://www.concentric.net/~psilon

  // Column IDCT operator :A_T*(F_T*E_T*B_T*D_T)*P_T
  // Let Y=[output column data, 8 elements], 16-bit short integer
  //     X=[input column data, 8 elements], 32-bit IEEE-754 float
  //
  //     Y= [ A_T*(F_T*E_T*B_T*D_T)*P_T ] * X
  //
  //   (Y and X are both column vectors)

  for ( i = 0; i < 8; ++i ) // column iDCT
  {
    xc = ((ssefloat *) fTempArray) + i; // select column #i
    y = &data[ i ];

    // 1) Apply (D_T * P_T) - the cos() coefficients of D_T are implicit
    //    in the idct_row operation.  But we still need to apply the
    //    shuffling operation of D_T.
    //
    //    1  0  0  0   0  0  0  0
    //    0  0  0  0   1  0  0  0
    //    0  0  1  0   0  0  0  0
    //    0  0  0  0   0  0  1  0
    //    0  1  0  0   0  0  0  0
    //    0  0  0  0   0  0  0  1
    //    0  0  0  1   0  0  0  0
    //    0  0  0  0   0  1  0  0

    dp[0] = xc[ 0 *8];
    dp[1] = xc[ 4 *8];
    dp[2] = xc[ 2 *8];
    dp[3] = xc[ 6 *8];

    dp[4] = xc[ 1 *8];
    dp[5] = xc[ 7 *8];
    dp[6] = xc[ 3 *8];
    dp[7] = xc[ 5 *8];
 
    // 2) Apply B_T
    //
    //    1  1  0  0
    //    1 -1  0  0
    //    0  0  1 t2
    //    0  0 t2 -1
    //                1 t1  0  0
    //               t1 -1  0  0
    //                0  0  1 t3
    //                0  0 t3 -1
 
 
    b[0] =   dp[1]            + dp[0];
    b[1] =   dp[0]            - dp[1];

    b[2] = ( dp[3]*tg_2_16f ) + dp[2];
    b[3] = ( dp[2]*tg_2_16f ) - dp[3];

    b[4] = ( dp[5]*tg_1_16f ) + dp[4];
    b[5] = ( dp[4]*tg_1_16f ) - dp[5];

    b[6] = ( dp[7]*tg_3_16f ) + dp[6];
    b[7] = ( dp[6]*tg_3_16f ) - dp[7];
  
    // 3) Apply E_T
    //
    //    1  0  1  0
    //    0  1  0  1
    //    0  1  0 -1
    //    1  0 -1  0
    //                1  0  1  0
    //                1  0 -1  0
    //                0  1  0  1
    //                0  1  0 -1
 
    e[0] = b[0] + b[2];
    e[1] = b[1] + b[3];
    e[2] = b[1] - b[3];
    e[3] = b[0] - b[2];
    e[4] = b[4] + b[6];
    e[5] = b[4] - b[6];
    e[6] = b[5] + b[7];
    e[7] = b[5] - b[7];

    // 4) Apply F_T
    //
    //    1  0  0  0
    //    0  1  0  0
    //    0  0  1  0
    //    0  0  0  1
    //                1  0  0  0
    //                0  1  0  0
    //                0  0  1  0
    //                0  0  0  1

#define _F0 e[0] 
#define _F1 e[1] 
#define _F2 e[2] 
#define _F3 e[3] 
#define _F4 e[4] 
#define _F5 e[5] 
#define _F6 e[6] 
#define _F7 e[7] 

    tmp[0] = (e[5] + e[6]) * cos_4_16f;
    tmp[1] = (e[5] - e[6]) * cos_4_16f;
    _F5 = tmp[0];
    _F6 = tmp[1];

    // 5) Apply A_T
    //
    //    1  0  0  0   1  0  0  0
    //    0  1  0  0   0  1  0  0
    //    0  0  1  0   0  0  1  0
    //    0  0  0  1   0  0  0  1
    //    0  0  0  1   0  0  0 -1
    //    0  0  1  0   0  0 -1  0
    //    0  1  0  0   0 -1  0  0
    //    1  0  0  0  -1  0  0  0
    //
    //    yfloat[0]= F0 + F4
    //    yfloat[1]= F1 + F5
    //    yfloat[2]= F2 + F6
    //    yfloat[3]= F3 + F7
    //           
    //    yfloat[4]= F3 - F7
    //    yfloat[5]= F2 - F6
    //    yfloat[6]= F1 - F5
    //    yfloat[7]= F0 - F4
    //
    //
    // 6) float -> int, shift/round to final output y[]
    //    The final shift&round operation reverses the row-input prescaling.
    //    It also applies the chosen rounding-mode (accurate or fast.)
    //
    //    Note, the C-code below differs *substantially* from the Intel-SSE 
    //    implementation.  Intel_SSE uses this basic sequence:
    //
    //       ADDPS xmm0, [round_compensation for y3 y2 y1 y0]  <float32>
    //
    //       cvtps2pi mm0,xmm0;  // mm0 <= [y1 y0] <int32>
    //     
    //       movhlps xmm0, xmm0; // xmm0 <=[y3 y2 y3 y2]
    //       cvtps2pi mm1,xmm0;  // mm1 <= [y3 y2] <int32>
    //
    //       ;// psrad is incorporated in the iDCT row coeff table w[]
    // 
    //       ;//psrad mm0, DESCALE_SHIFT1;  // NOT USED in ap922float_sse
    //       ;//psrad mm1, DESCALE_SHIFT1;  // NOT USED in ap922float_sse
    //
    //
    //       packssdw mm0, mm1;  // mm0 <= [y3 y2 y1 y0] <int16>
    //
    //       psraw mm0, 7;       // mm0 clipped to {-256,+255}
    //
    //       movq [ OUTC + 0], mm0;  // store
 

    y[0*8] = SHIFT_ROUND_COLF( _F0 + _F4 );
    y[1*8] = SHIFT_ROUND_COLF( _F1 + _F5 );
    y[2*8] = SHIFT_ROUND_COLF( _F2 + _F6 );
    y[3*8] = SHIFT_ROUND_COLF( _F3 + _F7 );

    y[4*8] = SHIFT_ROUND_COLF( _F3 - _F7 );
    y[5*8] = SHIFT_ROUND_COLF( _F2 - _F6 );
    y[6*8] = SHIFT_ROUND_COLF( _F1 - _F5 );
    y[7*8] = SHIFT_ROUND_COLF( _F0 - _F4 );

  } // end for ( i = 0; i < 8; ++i ) // end of SSEfloat column iDCT


  //    In standard-C, the output clip code adds significantly to 
  //    execution time.  

#define IDCT_CLIP_LOW -256  // IDCT output range is 9-bits
#define IDCT_CLIP_HIGH 255  // IDCT output range is 9-bits

  for ( i = 0; i < 64; ++i )
  { // clip output to {-256,+255}
    if ( data[ i ] < IDCT_CLIP_LOW )
      data[ i ] = IDCT_CLIP_LOW;

    if ( data[ i ] > IDCT_CLIP_HIGH )
      data[ i ] = IDCT_CLIP_HIGH;
  }
}